using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Xml.Serialization;
using System.IO;

namespace RegularLanguages
{
    public partial class DFAModelerForm : Form, LanguageForm<char>
    {
        public DFAModelerForm()
        {
            InitializeComponent();
            this.alphabet = null;
            this.dfa = null;
            this.okButton.Enabled = false;
            this.cancelButton.Enabled = false;
        }

        public DFAModelerForm(Alphabet<char> alphabet)
        {
            InitializeComponent();
            this.alphabet = alphabet;
            this.dfa = null;
            string dfaTemplate = "   \t     ";
            foreach (char token in this.alphabet)
                dfaTemplate += String.Format("{0,5}", token.ToString());
            dfaTemplate += "\r\n ->\t    0";
            this.dfaTextBox.Text = dfaTemplate;
        }

        public DFAModelerForm(DFA<char> dfa)
        {
            InitializeComponent();
            this.alphabet = dfa.Alphabet;
            this.dfa = dfa;
            this.DataTOUI();
        }

        public ILanguage<char> Language
        {
            get
            {
                try
                {
                    return new LanguageDFA<char>(this.dfa);
                }
                catch
                {
                    return null;
                }
            }
        }

        private Alphabet<char> alphabet;
        private DFA<char> dfa;
        private bool canClose = false;

        private void LoadDFAData(IDFAData<char> dfaData)
        {
            if (this.alphabet != null)
            {
                Alphabet<char> alphabet_cmp = new Alphabet<char>(dfaData);
                if (!this.alphabet.Equals(alphabet_cmp))
                {
                    MessageBox.Show("The DFA has an inconsistent alphabet.", "Error");
                    return;
                }
            }
            else
                this.alphabet = new Alphabet<char>(dfaData);
            this.dfa = new DFA<char>(dfaData);
            this.DataTOUI();
        }

        private void DataTOUI()
        {
            this.dfaTextBox.Text = dfa.ToString();
            this.commentTextBox.Text = dfa.Comment;
        }

        private bool BuildDFA()
        {
            this.dfa = null;
            if (this.dfaTextBox.Lines.Length < 2)
            {
                MessageBox.Show("Each DFA representation should have at least two lines of text.", "Warning");
                return false;
            }
            char[] separators = { ' ', '\t' };
            string alphabetLine = this.dfaTextBox.Lines[0];
            string[] alphabetLetters = alphabetLine.Split(separators, StringSplitOptions.RemoveEmptyEntries);
            if (alphabetLetters.Length == 0)
            {
                MessageBox.Show("Alphabet must contain at least one letter.", "Warning");
                return false;
            }
            int stateCount = this.dfaTextBox.Lines.Length - 1;
            int alphabetCount = alphabetLetters.Length;
            List<char> alphabetTokens = new List<char>(alphabetCount);
            foreach (string alphabetLetter in alphabetLetters)
            {
                if (alphabetLetter.Length != 1 || !Char.IsLetterOrDigit(alphabetLetter[0]))
                {
                    MessageBox.Show("Invalid alphabet letter " + alphabetLetter + ".", "Warning");
                    return false;
                }
                alphabetTokens.Add(alphabetLetter[0]);
            }
            for (int tokenIndex = 0; tokenIndex < alphabetCount - 1; ++tokenIndex)
                if (alphabetTokens[tokenIndex] >= alphabetTokens[tokenIndex + 1])
                {
                    MessageBox.Show("Alphabet letters should be ordered upwardly.", "Warning");
                    return false;
                }
            Alphabet<char> dfaAlphabet = new Alphabet<char>(alphabetTokens);
            DFA<char> dfaTemp = new DFA<char>(dfaAlphabet, stateCount);
            if (this.alphabet != null && !dfaAlphabet.Equals(this.alphabet))
            {
                MessageBox.Show("Inconsistent alphabet.", "Warning");
                return false;
            }
            string firstLine = this.dfaTextBox.Lines[1];
            string[] firstLineWords = firstLine.Split(separators, StringSplitOptions.RemoveEmptyEntries);
            if (firstLineWords.Length != alphabetCount + 2)
            {
                MessageBox.Show("Invalid number of words in the first line of the DFA transition table.", "Warning");
                return false;
            }
            if (firstLineWords[0] != "->" &&
                firstLineWords[0] != "<->")
            {
                MessageBox.Show("The first word in the first line of the DFA transition table should be -> or <->", "Warning");
                return false;
            }
            if (firstLineWords[0] == "<->")
                dfaTemp.SetAcceptState(0);
            for (int lineIndex = 0; lineIndex < stateCount; ++lineIndex)
            {
                string line = this.dfaTextBox.Lines[lineIndex + 1];
                string[] lineWords = line.Split(separators, StringSplitOptions.RemoveEmptyEntries);
                int transitionRowStart = 0;
                if (lineWords.Length != alphabetCount + 1 &&
                    lineWords.Length != alphabetCount + 2)
                {
                    MessageBox.Show("Invalid number of words in the line " +
                        (lineIndex + 1).ToString() + " of the DFA transition table.", "Warning");
                    return false;
                }
                if (lineWords.Length == alphabetCount + 2)
                {
                    if (lineWords[0] == "<-")
                        dfaTemp.SetAcceptState(lineIndex);
                    else if (lineIndex > 0)
                    {
                        MessageBox.Show("Invalid first word in the line " +
                            (lineIndex + 1).ToString() + " of the DFA transition table. " +
                            "Did you mean '<-' for output state?", "Warning");
                        return false;
                    }
                    transitionRowStart = 1;
                }
                int fromState = 0;
                if (!Int32.TryParse(lineWords[transitionRowStart], out fromState))
                {
                    MessageBox.Show("Unable to parse from-state " +
                        lineWords[transitionRowStart] + " in the line " +
                        (lineIndex + 1).ToString() + " of the DFA transition table.", "Warning");
                    return false;
                }
                if (fromState != lineIndex)
                {
                    MessageBox.Show("The from-state in the line " +
                        (lineIndex + 1).ToString() + " of the DFA transition table should be " +
                        lineIndex.ToString() + ".", "Warning");
                    return false;
                }
                for (int tokenIndex = 0; tokenIndex < alphabetCount; ++tokenIndex)
                {
                    int toState = 0;
                    if (!Int32.TryParse(lineWords[transitionRowStart + tokenIndex + 1], out toState))
                    {
                        MessageBox.Show("Unable to parse state " +
                            lineWords[transitionRowStart + tokenIndex + 1] + " in the line " +
                            (lineIndex + 1).ToString() + " of the DFA transition table.", "Warning");
                        return false;
                    }
                    if (toState < 0 || toState >= stateCount)
                    {
                        MessageBox.Show("State " + toState.ToString() + " is out of range.", "Warning");
                        return false;
                    }
                    dfaTemp.SetArcIndex(fromState, toState, tokenIndex);
                }
            }
            dfaTemp.Comment = this.commentTextBox.Text;
            this.dfa = dfaTemp;
            return true;
        }

        private void testButton_Click(object sender, EventArgs e)
        {
            if (this.BuildDFA())
            {
                DFAPreviewForm dfaForm = new DFAPreviewForm(this.dfa);
                dfaForm.ShowDialog();
            }
        }

        private void loadButton_Click(object sender, EventArgs e)
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();
            openFileDialog.Filter = "XML files (*.xml)|*.xml|All files (*.*)|*.*";
            DialogResult openFileDialogResult = openFileDialog.ShowDialog();
            if (openFileDialogResult == DialogResult.OK)
            {
                XmlSerializer dfaSerializer =
                    new XmlSerializer(typeof(DFAXML<char>));
                FileStream fileLoadStream =
                    new FileStream(openFileDialog.FileName, FileMode.Open);
                try
                {
                    DFAXML<char> dfaXML = dfaSerializer.Deserialize(fileLoadStream) as DFAXML<char>;
                    this.LoadDFAData(dfaXML);
                }
                catch
                {
                    MessageBox.Show("Unable to deserialize the DFA.", "Error");
                    return;
                }
                finally
                {
                    fileLoadStream.Close();
                }
            }
        }

        private void pasteButton_Click(object sender, EventArgs e)
        {
            if (Clipboard.ContainsText())
            {
                XmlSerializer dfaSerializer =
                    new XmlSerializer(typeof(DFAXML<char>));
                StringReader stringReader = new StringReader(Clipboard.GetText());
                try
                {
                    DFAXML<char> dfaXML = dfaSerializer.Deserialize(stringReader) as DFAXML<char>;
                    if (MessageBox.Show("Do you really want to replace the current DFA?",
                        "Confirmation", MessageBoxButtons.YesNo) == DialogResult.No)
                        return;
                    this.LoadDFAData(dfaXML);
                }
                catch
                {
                    MessageBox.Show("Unable to deserialize the DFA.", "Error");
                    return;
                }
                finally
                {
                    stringReader.Close();
                }
            }

        }

        private void okButton_Click(object sender, EventArgs e)
        {
            this.canClose = this.BuildDFA();
        }

        private void cancelButton_Click(object sender, EventArgs e)
        {
            this.canClose = true;
        }

        private void DFAModelerForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (e.CloseReason == CloseReason.None)
                e.Cancel = !this.canClose;
            else e.Cancel = false;
        }
    }
}