using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Xml.Serialization;
using System.IO;
using SimpleTextParser;

namespace RegularLanguages
{
    public partial class LStarForm : Form, LanguageForm<char>
    {
        public LStarForm()
        {
            InitializeComponent();
            this.lockAlphabet = false;
            this.alphabet = null;
            this.lstar = null;
            this.UpdateUI();
        }

        public LStarForm(Alphabet<char> alphabet)
        {
            InitializeComponent();
            this.lockAlphabet = true;
            this.alphabet = alphabet;
            this.lstar = new LStar<char>(alphabet);
            this.DataToUI();
            this.UpdateUI();
        }

        public LStarForm(LStar<char> lstar)
        {
            InitializeComponent();
            this.lockAlphabet = true;
            this.alphabet = lstar.Alphabet;
            this.lstar = lstar;
            this.CommandListViewRebuild();
            this.DataToUI();
            this.UpdateUI();
        }

        public ILanguage<char> Language
        {
            get
            {
                this.UIToData();
                if (this.lstar.CanConjecture)
                    return new LanguageDFALStar<char>(this.lstar);
                else return null;
            }
        }

        private string RegularToLStar(string regularWord)
        {
            if (regularWord == "") return "''";
            else return regularWord;
        }

        private string LStarToRegular(string lstarWord)
        {
            if (lstarWord == "''") return "";
            else return lstarWord;
        }

        private string AcceptToCommand(bool accept)
        {
            if (accept) return "accept";
            else return "reject";
        }

        private bool CommandToAccept(string command)
        {
            return command.Trim().ToLower() == "accept";
        }

        private void ClearUI()
        {
            this.commandListView.Items.Clear();
            this.unknownListView.Items.Clear();
            this.commentTextBox.Clear();
        }

        private void UpdateUI()
        {
            if (this.lockAlphabet)
                this.alphabetButton.Enabled = false;
            else
            {
                this.alphabetButton.Enabled = true;
                this.okButton.Enabled = false;
                this.cancelButton.Enabled = false;
            }
            if (this.alphabet == null)
            {
                this.conjectureButton.Enabled = false;
                this.rebuildButton.Enabled = false;
                this.clearButton.Enabled = false;
                this.sortButton.Enabled = false;
                this.deleteButton.Enabled = false;
                this.saveButton.Enabled = false;
                this.loadButton.Enabled = true;
                this.clearToolStripMenuItem.Enabled = false;
                this.sortToolStripMenuItem.Enabled = false;
                this.deleteToolStripMenuItem.Enabled = false;
                this.saveToolStripMenuItem.Enabled = false;
                this.loadToolStripMenuItem.Enabled = true;
                this.copyButton.Enabled = false;
                this.pasteButton.Enabled = true;
                this.viewButton.Enabled = false;
                this.copyXMLToolStripMenuItem.Enabled = false;
                this.pasteXMLToolStripMenuItem.Enabled = true;
                this.viewXMLToolStripMenuItem.Enabled = false;
                this.commandTextBox.Enabled = false;
                this.commentTextBox.Enabled = false;
                this.commandListView.Items.Clear();
                this.unknownListView.Items.Clear();
                this.commentTextBox.Clear();
            }
            else
            {
                this.conjectureButton.Enabled = this.lstar.CanConjecture;
                if (this.lockAlphabet)
                    this.okButton.Enabled = this.lstar.CanConjecture;
                this.rebuildButton.Enabled = false;
                this.clearButton.Enabled = true;
                this.sortButton.Enabled = true;
                this.deleteButton.Enabled = true;
                this.saveButton.Enabled = true;
                this.loadButton.Enabled = true;
                this.clearToolStripMenuItem.Enabled = true;
                this.sortToolStripMenuItem.Enabled = true;
                this.deleteToolStripMenuItem.Enabled = true;
                this.saveToolStripMenuItem.Enabled = true;
                this.loadToolStripMenuItem.Enabled = true;
                this.copyButton.Enabled = true;
                this.pasteButton.Enabled = true;
                this.viewButton.Enabled = true;
                this.copyXMLToolStripMenuItem.Enabled = true;
                this.pasteXMLToolStripMenuItem.Enabled = true;
                this.viewXMLToolStripMenuItem.Enabled = true;
                this.commandTextBox.Enabled = true;
                this.commentTextBox.Enabled = true;
                this.UnknownWordsRebuild();
            }
        }

        private void LockUIForRebuild()
        {
            if (this.lockAlphabet)
                this.okButton.Enabled = false;
            this.conjectureButton.Enabled = false;
            this.rebuildButton.Enabled = true;
            this.clearButton.Enabled = true;
            this.sortButton.Enabled = false;
            this.deleteButton.Enabled = true;
            this.saveButton.Enabled = false;
            this.clearToolStripMenuItem.Enabled = true;
            this.sortToolStripMenuItem.Enabled = false;
            this.deleteToolStripMenuItem.Enabled = true;
            this.saveToolStripMenuItem.Enabled = false;
            this.commandTextBox.Enabled = false;
            this.unknownListView.Items.Clear();
        }

        private void KnowledgeBaseRebuild()
        {
            KnowledgeBase<char> knowledgeBase = new KnowledgeBase<char>(this.alphabet);
            for (int index = 0; index < this.commandListView.Items.Count; ++index)
            {
                string command = this.commandListView.Items[index].SubItems[0].Text;
                string lstarWord = this.commandListView.Items[index].SubItems[1].Text;
                string regularWord = this.LStarToRegular(lstarWord);
                bool accept = this.CommandToAccept(command);
                knowledgeBase.Add(new GenericString<char>(regularWord), accept);
            }
            this.lstar.KnowledgeBase = knowledgeBase;
        }

        private void CommandListViewAdd(string regularWord, bool accept)
        {
            string lstarWord = this.RegularToLStar(regularWord);
            string command = this.AcceptToCommand(accept);
            ListViewItem listViewItem = this.commandListView.Items.Add(command);
            listViewItem.SubItems.Add(lstarWord);
        }

        private void CommandListViewSet(string regularWord, bool accept)
        {
            string lstarWord = this.RegularToLStar(regularWord);
            string command = this.AcceptToCommand(accept);
            for (int index = 0; index < this.commandListView.Items.Count; ++index)
                if (this.commandListView.Items[index].SubItems[1].Text == lstarWord)
                {
                    this.commandListView.Items[index].SubItems[0].Text = command;
                    break;
                }
        }

        private void CommandListViewDeleteSelected()
        {
            if (this.commandListView.SelectedIndices.Count == 1)
            {
                int selectedIndex = this.commandListView.SelectedIndices[0];
                this.commandListView.Items.Remove(
                    this.commandListView.Items[selectedIndex]);
                this.LockUIForRebuild();
            }
        }

        private void CommandListViewRebuild()
        {
            this.commandListView.SuspendLayout();
            this.commandListView.Items.Clear();
            foreach (WordAccept<char> wordAccept in this.lstar.KnowledgeBase)
                CommandListViewAdd(
                    wordAccept.word.ToString(),
                    wordAccept.accept);
            this.commandListView.ResumeLayout(true);
        }

        private void UnknownWordsRebuild()
        {
            if (this.lstar == null) return;
            this.unknownListView.SuspendLayout();
            this.unknownListView.Items.Clear();
            foreach (GenericString<char> unknownWord in this.lstar.UnknownWords)
            {
                string unknownWordString = unknownWord.ToString();
                this.unknownListView.Items.Add(
                    this.RegularToLStar(unknownWordString));
            }
            this.unknownListView.ResumeLayout(true);
        }

        private void LoadKnowledgeBaseData(IKnowledgeBaseData<char> knowledgeBaseData)
        {
            if (this.lockAlphabet)
            {
                Alphabet<char> alphabet_cmp = new Alphabet<char>(knowledgeBaseData);
                if (!this.alphabet.Equals(alphabet_cmp))
                {
                    MessageBox.Show("The knowledge base has an inconsistent alphabet.", "Error");
                    return;
                }
            }
            else
                this.alphabet = new Alphabet<char>(knowledgeBaseData);
            this.lstar = new LStar<char>(this.alphabet);
            this.lstar.KnowledgeBase = new KnowledgeBase<char>(knowledgeBaseData);
            this.CommandListViewRebuild();
            this.DataToUI();
            this.UpdateUI();
        }

        private void DataToUI()
        {
            this.commentTextBox.Text = this.lstar.KnowledgeBase.Comment;
        }

        private void UIToData()
        {
            this.lstar.KnowledgeBase.Comment = this.commentTextBox.Text;
        }

        private bool lockAlphabet;
        private Alphabet<char> alphabet;
        private LStar<char> lstar;
        private TextParser textParser = new TextParser();
        private bool canClose = false;

        private void alphabetButton_Click(object sender, EventArgs e)
        {
            AlphabetForm alphabetForm = new AlphabetForm();
            DialogResult dialogResult = alphabetForm.ShowDialog();
            if (dialogResult == DialogResult.OK)
            {
                if (alphabetForm.Alphabet == "")
                {
                    this.alphabet = null;
                    this.lstar = null;
                }
                else
                {
                    this.alphabet = new Alphabet<char>(alphabetForm.Alphabet);
                    this.lstar = new LStar<char>(this.alphabet);
                }
                this.ClearUI();
                this.UpdateUI();
            }
        }

        private void conjectureButton_Click(object sender, EventArgs e)
        {
            this.UIToData();
            DFAPreviewForm dfaForm = new DFAPreviewForm(this.lstar.Conjecture);
            dfaForm.ShowDialog();
        }

        private void rebuildButton_Click(object sender, EventArgs e)
        {
            this.KnowledgeBaseRebuild();
            this.UpdateUI();
        }

        private void commandListView_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Delete)
                this.CommandListViewDeleteSelected();
        }

        private void commandListView_DragDrop(object sender, DragEventArgs e)
        {
            this.LockUIForRebuild();
        }

        private void commandTextBox_KeyDown(object sender, KeyEventArgs e)
        {
            string text = this.commandTextBox.Text;
            if (e.KeyCode == Keys.Enter)
            {
                if (text.Trim() == "") return;
                this.commandTextBox.Clear();
                TextParserStruct[] textParserStructs =
                    this.textParser.Parse(text);
                if (textParserStructs != null &&
                    textParserStructs.Length == 2)
                {
                    string regularWord = null;
                    bool accept = false;
                    regularWord = textParserStructs[1].word;
                    if ((textParserStructs[0].word == "acc" ||
                         textParserStructs[0].word == "accept") &&
                        (textParserStructs[0].quoted == false))
                        accept = true;
                    else if ((textParserStructs[0].word == "rej" ||
                              textParserStructs[0].word == "reject") &&
                             (textParserStructs[0].quoted == false))
                        accept = false;
                    else
                    {
                        MessageBox.Show("Allowed commands: acc, accept, rej, reject.", "Parse error");
                        return;
                    }
                    if (!this.alphabet.Contains(regularWord))
                    {
                        MessageBox.Show("All the letters of the word must be from the alphabet.", "Warning");
                        return;
                    }
                    KnowledgeBase<char>.AddResult addResult =
                        this.lstar.KnowledgeBase.Add(
                            new GenericString<char>(regularWord), accept);
                    switch (addResult)
                    {
                        case KnowledgeBase<char>.AddResult.ItemAdded:
                            this.CommandListViewAdd(regularWord, accept);
                            break;
                        case KnowledgeBase<char>.AddResult.ItemChanged:
                            this.CommandListViewSet(regularWord, accept);
                            break;
                        case KnowledgeBase<char>.AddResult.NothingChanged:
                            break;
                        default:
                            break;
                    }
                    this.UpdateUI();
                }
                else
                {
                    MessageBox.Show("Hint: Try something like accept abba or reject ''", "Parse error");
                    return;
                }
            }
        }

        private void commandTextBox_KeyUp(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
                this.commandTextBox.Clear();
        }

        private void clearButton_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show("Do you really want to clear the knowledge base?",
                "Confirmation", MessageBoxButtons.YesNo) == DialogResult.No)
                return;
            this.lstar.KnowledgeBase.Clear();
            this.commandListView.Items.Clear();
            this.commentTextBox.Clear();
            this.UpdateUI();
        }

        private void sortButton_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show("Do you really want to sort the knowledge base?",
                "Confirmation", MessageBoxButtons.YesNo) == DialogResult.No)
                return;
            this.lstar.KnowledgeBase.Sort();
            this.CommandListViewRebuild();
            this.UpdateUI();
        }

        private void deleteButton_Click(object sender, EventArgs e)
        {
            this.CommandListViewDeleteSelected();
        }

        private void saveButton_Click(object sender, EventArgs e)
        {
            this.UIToData();
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            saveFileDialog.Filter = "XML files (*.xml)|*.xml|All files (*.*)|*.*";
            DialogResult saveFileDialogResult = saveFileDialog.ShowDialog();
            if (saveFileDialogResult == DialogResult.OK)
            {
                XmlSerializer knowledgeBaseSerializer =
                    new XmlSerializer(typeof(KnowledgeBaseXML<char>));
                FileStream fileSaveStream =
                    new FileStream(saveFileDialog.FileName, FileMode.Create);
                KnowledgeBaseXML<char> knowledgeBaseXML =
                    new KnowledgeBaseXML<char>(this.lstar.KnowledgeBase);
                knowledgeBaseSerializer.Serialize(fileSaveStream, knowledgeBaseXML);
                fileSaveStream.Close();
            }
        }

        private void loadButton_Click(object sender, EventArgs e)
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();
            openFileDialog.Filter = "XML files (*.xml)|*.xml|All files (*.*)|*.*";
            DialogResult openFileDialogResult = openFileDialog.ShowDialog();
            if (openFileDialogResult == DialogResult.OK)
            {
                // Knowledge base
                FileStream fileLoadStream = new FileStream(openFileDialog.FileName, FileMode.Open);
                try
                {
                    XmlSerializer knowledgeBaseSerializer =
                        new XmlSerializer(typeof(KnowledgeBaseXML<char>));
                    KnowledgeBaseXML<char> knowledgeBaseXML =
                        knowledgeBaseSerializer.Deserialize(fileLoadStream) as KnowledgeBaseXML<char>;
                    this.LoadKnowledgeBaseData(knowledgeBaseXML);
                    return;
                }
                catch { }
                finally { fileLoadStream.Close(); }
                MessageBox.Show("Unable to deserialize the positive and negative samples.", "Error");
            }
        }

        private void copyButton_Click(object sender, EventArgs e)
        {
            this.UIToData();
            XmlSerializer knowledgeBaseSerializer =
                new XmlSerializer(typeof(KnowledgeBaseXML<char>));
            StringWriter stringWriter = new StringWriter();
            KnowledgeBaseXML<char> knowledgeBaseXML =
                new KnowledgeBaseXML<char>(this.lstar.KnowledgeBase);
            knowledgeBaseSerializer.Serialize(stringWriter, knowledgeBaseXML);
            Clipboard.SetText(stringWriter.ToString());
            stringWriter.Close();
        }

        private void pasteButton_Click(object sender, EventArgs e)
        {
            if (Clipboard.ContainsText())
            {
                // Knowledge base
                StringReader stringReader = new StringReader(Clipboard.GetText());
                try
                {
                    XmlSerializer knowledgeBaseSerializer =
                        new XmlSerializer(typeof(KnowledgeBaseXML<char>));
                    KnowledgeBaseXML<char> knowledgeBaseXML =
                        knowledgeBaseSerializer.Deserialize(stringReader) as KnowledgeBaseXML<char>;
                    if (MessageBox.Show("Do you really want to replace the current knowledge base?",
                        "Confirmation", MessageBoxButtons.YesNo) == DialogResult.No)
                        return;
                    this.LoadKnowledgeBaseData(knowledgeBaseXML);
                    return;
                }
                catch { }
                finally { stringReader.Close(); }
                MessageBox.Show("Unable to deserialize the positive and negative samples.", "Error");
            }
        }

        private void viewButton_Click(object sender, EventArgs e)
        {
            this.UIToData();
            XmlSerializer knowledgeBaseSerializer =
                new XmlSerializer(typeof(KnowledgeBaseXML<char>));
            StringWriter stringWriter = new StringWriter();
            KnowledgeBaseXML<char> knowledgeBaseXML =
                new KnowledgeBaseXML<char>(this.lstar.KnowledgeBase);
            knowledgeBaseSerializer.Serialize(stringWriter, knowledgeBaseXML);
            XMLPreviewForm xmlPreviewForm = new XMLPreviewForm(stringWriter.ToString());
            stringWriter.Close();
            xmlPreviewForm.ShowDialog();
        }

        private void okButton_Click(object sender, EventArgs e)
        {
            this.UIToData();
            this.canClose = this.lstar.CanConjecture;
        }

        private void cancelButton_Click(object sender, EventArgs e)
        {
            this.canClose = true;
        }

        private void LStarForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (e.CloseReason == CloseReason.None)
                e.Cancel = !this.canClose;
            else e.Cancel = false;
        }
    }
}