using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Xml.Serialization;
using System.IO;
using SimpleTextParser;

namespace RegularLanguages
{
    public partial class RPNIForm : Form, LanguageForm<char>
    {
        public RPNIForm()
        {
            InitializeComponent();
            this.lockAlphabet = false;
            this.alphabet = null;
            this.rpni = null;
            this.UpdateUI();
        }

        public RPNIForm(Alphabet<char> alphabet)
        {
            InitializeComponent();
            this.lockAlphabet = true;
            this.alphabet = alphabet;
            this.rpni = new RPNI<char>(alphabet);
            this.UpdateUI();
        }

        public RPNIForm(RPNI<char> rpni)
        {
            InitializeComponent();
            this.lockAlphabet = true;
            this.alphabet = rpni.Alphabet;
            this.rpni = rpni;
            this.DataToUI();
            this.UpdateUI();
        }

        public ILanguage<char> Language
        {
            get
            {
                if (this.UIToData())
                    return new LanguageDFARPNI<char>(this.rpni);
                else return null;
            }
        }

        private string RegularToRPNI(string regularWord)
        {
            if (regularWord == "") return "''";
            else return regularWord;
        }

        private void ClearUI()
        {
            this.acceptTextBox.Clear();
            this.rejectTextBox.Clear();
            this.commentTextBox.Clear();
        }

        private void UpdateUI()
        {
            if (this.lockAlphabet)
                this.alphabetButton.Enabled = false;
            else
            {
                this.alphabetButton.Enabled = true;
                this.okButton.Enabled = false;
                this.cancelButton.Enabled = false;
            }
            if (alphabet == null)
            {
                this.ClearUI();
                this.clearButton.Enabled = false;
                this.conjectureButton.Enabled = false;
                this.saveButton.Enabled = false;
                this.acceptTextBox.Enabled = false;
                this.rejectTextBox.Enabled = false;
                this.commentTextBox.Enabled = false;
                this.copyButton.Enabled = false;
                this.pasteButton.Enabled = true;
                this.viewButton.Enabled = false;
            }
            else
            {
                this.clearButton.Enabled = true;
                this.conjectureButton.Enabled = true;
                this.saveButton.Enabled = true;
                this.acceptTextBox.Enabled = true;
                this.rejectTextBox.Enabled = true;
                this.commentTextBox.Enabled = true;
                this.copyButton.Enabled = true;
                this.pasteButton.Enabled = true;
                this.viewButton.Enabled = true;
            }
        }

        private bool lockAlphabet;
        private Alphabet<char> alphabet;
        private RPNI<char> rpni;
        private TextParser textParser = new TextParser();
        private bool canClose = false;

        private bool LoadAlphabet(IAlphabetData<char> alphabetData)
        {
            if (this.alphabet != null)
            {
                Alphabet<char> alphabet_cmp =
                    new Alphabet<char>(alphabetData);
                if (!this.alphabet.Equals(alphabet_cmp))
                {
                    MessageBox.Show("The positive and negative samples have an inconsistent alphabet.", "Error");
                    return false;
                }
            }
            else
            {
                this.alphabet = new Alphabet<char>(alphabetData);
            }
            return true;
        }

        private void LoadPositiveNegativeSamplesData(
            IPositiveNegativeSamplesData<char> positiveNegativeSamplesData)
        {
            if (!LoadAlphabet(positiveNegativeSamplesData)) return;
            if (this.rpni == null) this.rpni = new RPNI<char>(this.alphabet);
            this.rpni.Samples = new PositiveNegativeSamples<char>(
                positiveNegativeSamplesData, PositiveNegativeSamplesOptions.AllSamplesEnabled);
            this.DataToUI();
            this.UpdateUI();
        }

        private void LoadKnowledgeBaseData(
            IKnowledgeBaseData<char> knowledgeBaseData)
        {
            if (!LoadAlphabet(knowledgeBaseData)) return;
            if (this.rpni == null) this.rpni = new RPNI<char>(this.alphabet);
            this.rpni.Samples = new PositiveNegativeSamples<char>(
                knowledgeBaseData, PositiveNegativeSamplesOptions.AllSamplesEnabled);
            this.DataToUI();
            this.UpdateUI();
        }

        private void LoadSLTData(ISLTData<char> sltData)
        {
            if (!LoadAlphabet(sltData)) return;
            if (this.rpni == null) this.rpni = new RPNI<char>(this.alphabet);
            this.rpni.Samples = new PositiveNegativeSamples<char>(
                sltData, PositiveNegativeSamplesOptions.AllSamplesEnabled);
            this.DataToUI();
            this.UpdateUI();
        }

        private void DataToUI()
        {
            this.acceptTextBox.Clear();
            this.rejectTextBox.Clear();
            List<string> positiveSampleList = new List<string>();
            foreach (GenericString<char> positiveSample in this.rpni.Samples.PositiveSamples)
                positiveSampleList.Add(RegularToRPNI(positiveSample.ToString()));
            List<string> negativeSampleList = new List<string>();
            foreach (GenericString<char> negativeSample in this.rpni.Samples.NegativeSamples)
                negativeSampleList.Add(RegularToRPNI(negativeSample.ToString()));
            this.acceptTextBox.Text = String.Join(" ", positiveSampleList.ToArray());
            this.rejectTextBox.Text = String.Join(" ", negativeSampleList.ToArray());
            this.commentTextBox.Text = this.rpni.Samples.Comment;
        }

        private bool UIToData()
        {
            this.rpni.Samples.Clear();
            TextParserStruct[] acceptStructs =
                this.textParser.Parse(acceptTextBox.Text);
            TextParserStruct[] rejectStructs =
                this.textParser.Parse(rejectTextBox.Text);
            int acceptCount = 0;
            int rejectCount = 0;
            if (acceptStructs != null)
                acceptCount = acceptStructs.Length;
            if (rejectStructs != null)
                rejectCount = rejectStructs.Length;
            try
            {
                if (acceptStructs != null)
                    foreach (TextParserStruct acceptStruct in acceptStructs)
                    {
                        if (!this.rpni.Alphabet.Contains(acceptStruct.word))
                        {
                            MessageBox.Show("All the letters of the word must be from the alphabet.", "Warning");
                            return false;
                        }
                        this.rpni.Samples.AddPositiveSample(new GenericString<char>(acceptStruct.word));
                    }
                if (rejectStructs != null)
                    foreach (TextParserStruct rejectStruct in rejectStructs)
                    {
                        if (!this.rpni.Alphabet.Contains(rejectStruct.word))
                        {
                            MessageBox.Show("All the letters of the word must be from the alphabet.", "Warning");
                            return false;
                        }
                        this.rpni.Samples.AddNegativeSample(new GenericString<char>(rejectStruct.word));
                    }
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.Message, "Warning");
                return false;
            }
            this.rpni.Samples.Comment = this.commentTextBox.Text;
            return true;
        }

        private void alphabetButton_Click(object sender, EventArgs e)
        {
            AlphabetForm alphabetForm = new AlphabetForm();
            DialogResult dialogResult = alphabetForm.ShowDialog();
            if (dialogResult == DialogResult.OK)
            {
                if (alphabetForm.Alphabet == "")
                {
                    this.alphabet = null;
                    this.rpni = null;
                }
                else
                {
                    this.alphabet = new Alphabet<char>(alphabetForm.Alphabet);
                    this.rpni = new RPNI<char>(alphabet);
                }
                this.ClearUI();
                this.UpdateUI();
            }
        }

        private void clearButton_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show("Do you really want to clear the positive and negative samples?",
                "Confirmation", MessageBoxButtons.YesNo) == DialogResult.No)
                return;
            this.ClearUI();
            this.UpdateUI();
        }

        private void conjectureButton_Click(object sender, EventArgs e)
        {
            if (this.UIToData())
            {
                DFAPreviewForm dfaForm = new DFAPreviewForm(this.rpni.Conjecture);
                dfaForm.ShowDialog();
            }
        }

        private void saveButton_Click(object sender, EventArgs e)
        {
            if (this.UIToData())
            {
                SaveFileDialog saveFileDialog = new SaveFileDialog();
                saveFileDialog.Filter = "XML files (*.xml)|*.xml|All files (*.*)|*.*";
                DialogResult saveFileDialogResult = saveFileDialog.ShowDialog();
                if (saveFileDialogResult == DialogResult.OK)
                {
                    XmlSerializer positiveNegativeSamplesSerializer =
                        new XmlSerializer(typeof(PositiveNegativeSamplesXML<char>));
                    FileStream fileSaveStream =
                        new FileStream(saveFileDialog.FileName, FileMode.Create);
                    PositiveNegativeSamplesXML<char> positiveNegativeSamplesXML =
                        new PositiveNegativeSamplesXML<char>(this.rpni.Samples);
                    positiveNegativeSamplesSerializer.Serialize(
                        fileSaveStream, positiveNegativeSamplesXML);
                    fileSaveStream.Close();
                }
            }
        }

        private void loadButton_Click(object sender, EventArgs e)
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();
            openFileDialog.Filter = "XML files (*.xml)|*.xml|All files (*.*)|*.*";
            DialogResult openFileDialogResult = openFileDialog.ShowDialog();
            if (openFileDialogResult == DialogResult.OK)
            {
                // Positive and negative samples
                FileStream fileLoadStream = new FileStream(openFileDialog.FileName, FileMode.Open);
                try
                {
                    XmlSerializer positiveNegativeSamplesSerializer =
                        new XmlSerializer(typeof(PositiveNegativeSamplesXML<char>));
                    PositiveNegativeSamplesXML<char> positiveNegativeSamplesXML =
                        positiveNegativeSamplesSerializer.Deserialize(fileLoadStream)
                            as PositiveNegativeSamplesXML<char>;
                    this.LoadPositiveNegativeSamplesData(positiveNegativeSamplesXML);
                    return;
                }
                catch { }
                finally { fileLoadStream.Close(); }
                // Knowledge base
                fileLoadStream = new FileStream(openFileDialog.FileName, FileMode.Open);
                try
                {
                    XmlSerializer knowledgeBaseSerializer =
                        new XmlSerializer(typeof(KnowledgeBaseXML<char>));
                    KnowledgeBaseXML<char> knowledgeBaseXML =
                        knowledgeBaseSerializer.Deserialize(fileLoadStream) as KnowledgeBaseXML<char>;
                    this.LoadKnowledgeBaseData(knowledgeBaseXML);
                    return;
                }
                catch { }
                finally { fileLoadStream.Close(); }
                // SLT
                fileLoadStream = new FileStream(openFileDialog.FileName, FileMode.Open);
                try
                {
                    XmlSerializer sltSerializer =
                        new XmlSerializer(typeof(SLTXML<char>));
                    SLTXML<char> sltXML =
                        sltSerializer.Deserialize(fileLoadStream) as SLTXML<char>;
                    this.LoadSLTData(sltXML);
                    return;
                }
                catch { }
                finally { fileLoadStream.Close(); }
                MessageBox.Show("Unable to deserialize the positive and negative samples.", "Error");
            }
        }

        private void copyButton_Click(object sender, EventArgs e)
        {
            if (this.UIToData())
            {
                XmlSerializer positiveNegativeSamplesSerializer =
                    new XmlSerializer(typeof(PositiveNegativeSamplesXML<char>));
                StringWriter stringWriter = new StringWriter();
                PositiveNegativeSamplesXML<char> positiveNegativeSamplesXML =
                    new PositiveNegativeSamplesXML<char>(this.rpni.Samples);
                positiveNegativeSamplesSerializer.Serialize(stringWriter,
                    positiveNegativeSamplesXML);
                Clipboard.SetText(stringWriter.ToString());
                stringWriter.Close();
            }
        }

        private void pasteButton_Click(object sender, EventArgs e)
        {
            if (Clipboard.ContainsText())
            {
                string confirmation = "Do you really want to replace the current positive and negative samples?";
                // Positive and negative samples
                StringReader stringReader = new StringReader(Clipboard.GetText());
                try
                {
                    XmlSerializer positiveNegativeSamplesSerializer =
                        new XmlSerializer(typeof(PositiveNegativeSamplesXML<char>));
                    PositiveNegativeSamplesXML<char> positiveNegativeSamplesXML =
                        positiveNegativeSamplesSerializer.Deserialize(stringReader)
                            as PositiveNegativeSamplesXML<char>;
                    if (MessageBox.Show(confirmation, "Confirmation", 
                            MessageBoxButtons.YesNo) == DialogResult.No)
                        return;
                    this.LoadPositiveNegativeSamplesData(positiveNegativeSamplesXML);
                    return;
                }
                catch { }
                finally { stringReader.Close(); }
                // Knowledge base
                stringReader = new StringReader(Clipboard.GetText());
                try
                {
                    XmlSerializer knowledgeBaseSerializer =
                        new XmlSerializer(typeof(KnowledgeBaseXML<char>));
                    KnowledgeBaseXML<char> knowledgeBaseXML =
                        knowledgeBaseSerializer.Deserialize(stringReader) as KnowledgeBaseXML<char>;
                    if (MessageBox.Show(confirmation, "Confirmation",
                            MessageBoxButtons.YesNo) == DialogResult.No)
                        return;
                    this.LoadKnowledgeBaseData(knowledgeBaseXML);
                    return;
                }
                catch { }
                finally { stringReader.Close(); }
                // SLT
                stringReader = new StringReader(Clipboard.GetText());
                try
                {
                    XmlSerializer sltSerializer =
                        new XmlSerializer(typeof(SLTXML<char>));
                    SLTXML<char> sltXML =
                        sltSerializer.Deserialize(stringReader) as SLTXML<char>;
                    if (MessageBox.Show(confirmation, "Confirmation",
                            MessageBoxButtons.YesNo) == DialogResult.No)
                        return;
                    this.LoadSLTData(sltXML);
                    return;
                }
                catch { }
                finally { stringReader.Close(); }
                MessageBox.Show("Unable to deserialize the positive and negative samples.", "Error");
            }
        }

        private void viewButton_Click(object sender, EventArgs e)
        {
            XmlSerializer positiveNegativeSamplesSerializer =
                new XmlSerializer(typeof(PositiveNegativeSamplesXML<char>));
            StringWriter stringWriter = new StringWriter();
            PositiveNegativeSamplesXML<char> positiveNegativeSamplesXML =
                new PositiveNegativeSamplesXML<char>(this.rpni.Samples);
            positiveNegativeSamplesSerializer.Serialize(stringWriter,
                positiveNegativeSamplesXML);
            XMLPreviewForm xmlPreviewForm = new XMLPreviewForm(stringWriter.ToString());
            stringWriter.Close();
            xmlPreviewForm.ShowDialog();
        }

        private void okButton_Click(object sender, EventArgs e)
        {
            this.canClose = this.UIToData();
        }

        private void cancelButton_Click(object sender, EventArgs e)
        {
            this.canClose = true;
        }

        private void RPNIForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (e.CloseReason == CloseReason.None)
                e.Cancel = !this.canClose;
            else e.Cancel = false;
        }
    }
}