using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Text;
using System.Text.RegularExpressions;

namespace RegularLanguages
{
    /// <summary>
    /// Simple static class designed to remove all duplicities from a list.
    /// </summary>
    /// <typeparam name="T">Any type T that implements IEquatable&lt;T&gt; interface.</typeparam>
    public static class Unique<T>
        where T : IEquatable<T>
    {
        /// <summary>
        /// Static method that removes all duplicities from the input list.
        /// </summary>
        /// <example>
        /// <code>
        /// List&lt;int&gt; list = new List&lt;int&gt;();
        /// list.Add(4);
        /// list.Add(3);
        /// list.Add(4);
        /// list.Add(2);
        /// list.Add(1);
        /// list.Sort();
        /// List&lt;int&gt; listWithoutDuplicities =
        ///     Unique&lt;int&gt;.UniqueList(list);
        /// </code>
        /// </example>
        /// <param name="list">[Sorted] list.</param>
        /// <returns>List without duplicities.</returns>
        public static List<T> UniqueList(List<T> list)
        {
            List<T> listUnique = new List<T>(list.Count);
            T currentItem = default(T);
            bool firstItem = true;
            foreach (T item in list)
                if (firstItem || !item.Equals(currentItem))
                {
                    listUnique.Add(item);
                    firstItem = false;
                    currentItem = item;
                }
            return listUnique;
        }
    }

    /// <summary>
    /// Immutable class that represents a generic string that consists of tokens of type T.
    /// Warning: GenericString concatenation is expensive. Each concatenation creates a new
    /// GenericString object. This can cause object-allocation overhead.
    /// Use GenericStringBuilder instead.
    /// </summary>
    /// <example>
    /// <code>
    /// GenericString&lt;char&gt; genStr1 = new GenericString&lt;char&gt;("Hello ");
    /// GenericString&lt;char&gt; genStr2 = new GenericString&lt;char&gt;("world.");
    /// GenericString&lt;char&gt; genStrConcat = genStr1 + genStr2;
    /// </code>
    /// </example>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class GenericString<T> :
        ICloneable,
        IComparable<GenericString<T>>,
        IEnumerable<T>,
        IEquatable<GenericString<T>>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        private GenericString()
        {
            this.data = null;
        }

        /// <summary>
        /// Initializes a new instance of the GenericString&lt;T&gt; class
        /// that encapsulates a copy of an input generic string.
        /// </summary>
        /// <param name="tokens">Input generic string.</param>
        public GenericString(IEnumerable<T> tokens)
        {
            if (tokens == null)
                this.data = new T[0];
            else
            {
                const int minimalCapacity = 32;
                List<T> tokenList = new List<T>(minimalCapacity);
                tokenList.AddRange(tokens);
                this.data = tokenList.ToArray();
            }
        }
        #endregion

        #region Static Constructors (Wrapper)
        /// <summary>
        /// Initializes a new instance of the GenericString&lt;T&gt; class 
        /// that holds exactly one input token.
        /// </summary>
        /// <example>
        /// <code>
        /// GenericString&lt;char&gt; genStr =
        ///     GenericString&lt;char&gt;.Wrap('a');
        /// </code>
        /// </example>
        /// <param name="token">Input token.</param>
        public static GenericString<T> Wrap(T token)
        {
            T[] tokens = new T[1];
            tokens[0] = token;
            return GenericString<T>.Wrap(tokens);
        }

        /// <summary>
        /// Initializes a new instance of the GenericString&lt;T&gt; class
        /// that wraps an input source array.
        /// Warning: It is very important not to change this array outside
        ///     when wrapped by GenericString&lt;T&gt;.
        /// </summary>
        /// <example>
        /// <code>
        /// string str = "Hello world.";
        /// GenericString&lt;char&gt; genStr =
        ///     GenericString&lt;char&gt;.Wrap(str.ToArray());
        /// </code>
        /// </example>
        /// <param name="source">Input source array.</param>
        public static GenericString<T> Wrap(T[] source)
        {
            GenericString<T> genericString = new GenericString<T>();
            if (source != null)
                genericString.data = source;
            else genericString.data = new T[0];
            return genericString;
        }
        #endregion

        #region Static Members
        /// <summary>
        /// EmptyWord represents an empty generic string.
        /// </summary>
        public static readonly GenericString<T> EmptyWord = GenericString<T>.Wrap(new T[0]);
        #endregion

        #region ICloneable Interface
        /// <summary>
        /// Returns a reference to this instance of GenericString&lt;T&gt;.
        /// </summary>
        public object Clone()
        {
            return this;
        }
        #endregion

        #region IComparable Interface
        /// <summary>
        /// Standard enumeration of strings.
        /// </summary>
        /// <example>
        /// Standard enumeration of strings over alphabet {a, b} is:
        /// "", "a", "b", "aa", "ab", "ba", "bb", "aaa", ...
        /// </example>
        public int CompareTo(GenericString<T> genericString)
        {
            if (this.data.Length < genericString.data.Length) return -1;
            else if (this.data.Length > genericString.data.Length) return +1;
            else
            {
                for (int index = 0; index < this.data.Length; ++index)
                {
                    int cmp = this.data[index].CompareTo(genericString.data[index]);
                    if (cmp < 0) return -1;
                    else if (cmp > 0) return +1;
                }
            }
            return 0;
        }
        #endregion

        #region IEnumerable Interface
        /// <summary>
        /// Returns an enumerator that iterates through the GenericString&lt;T&gt;. 
        /// </summary>
        IEnumerator<T> IEnumerable<T>.GetEnumerator()
        {
            foreach (T token in this.data)
                yield return token;
        }

        System.Collections.IEnumerator
            System.Collections.IEnumerable.GetEnumerator()
        {
            throw new NotImplementedException();
        }
        #endregion

        #region IEquatable Interface
        /// <summary>
        /// Overloaded. Determines whether two Object instances are equal. (Inherited from Object.)
        /// </summary>
        public override bool Equals(object obj)
        {
            if (obj is GenericString<T>)
                return this.Equals(obj as GenericString<T>);
            else
                return false;
        }

        /// <summary>
        /// Determines whether two GenericString&lt;T&gt; instances are equal.
        /// </summary>
        public bool Equals(GenericString<T> genericString)
        {
            if (this.data.Length != genericString.data.Length) return false;
            for (int index = 0; index < this.data.Length; ++index)
                if (!this.data[index].Equals(genericString.data[index]))
                    return false;
            return true;
        }

        /// <summary>
        /// Determines whether a substring of this instance is equal to the given generic string.
        /// </summary>
        public bool EqualsFrom(GenericString<T> genericString, int start)
        {
            if (start > this.data.Length - genericString.data.Length)
            {
                return false;
            }
            else
            {
                for (int index = 0; index < genericString.data.Length; ++index)
                    if (!this.data[start + index].Equals(genericString.data[index]))
                        return false;
                return true;
            }
        }

        /// <summary>
        /// Serves as a hash function for a particular type. GetHashCode is suitable for use in hashing algorithms
        /// and data structures like a hash table. (Inherited from Object.)
        /// </summary>
        public override int GetHashCode()
        {
            int hashCode = 0;
            for (int index = 0; index < this.data.Length; ++index)
                hashCode = 16 * hashCode + this.data[index].GetHashCode() + index;
            return hashCode;
        }
        #endregion

        #region Indexer
        /// <summary>
        /// Gets an element at the specified index.
        /// </summary>
        public T this[int index]
        {
            get
            {
                return this.data[index];
            }
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets the number of tokens actually contained in the GenericString&lt;T&gt;.
        /// </summary>
        public int Length
        {
            get
            {
                return this.data.Length;
            }
        }
        #endregion

        #region Operators
        /// <summary>
        /// Creates a new GenericString&lt;T&gt; instance that is a concatenation of two generic strings.
        /// Warning: GenericString concatenation is expensive. Each concatenation creates a new
        /// GenericString object. This can cause object-allocation overhead.
        /// Use GenericStringBuilder instead.
        /// </summary>
        /// <param name="lhs">Left hand side.</param>
        /// <param name="rhs">Right hand side.</param>
        /// <returns>Concatenation of the operands.</returns>
        public static GenericString<T> operator +(GenericString<T> lhs, GenericString<T> rhs)
        {
            return lhs.Concat(rhs);
        }
        #endregion

        #region Methods
        /// <summary>
        /// Creates a new GenericString&lt;T&gt; instance that is a concatenation of two generic strings.
        /// Warning: GenericString concatenation is expensive. Each concatenation creates a new
        /// GenericString object. This can cause object-allocation overhead.
        /// Use GenericStringBuilder instead.
        /// </summary>
        /// <param name="genericString">Generic string that will be added to this generic string.</param>
        /// <returns>Concatenation of this and genericString.</returns>
        public GenericString<T> Concat(GenericString<T> genericString)
        {
            T[] concatData = new T[this.data.Length + genericString.data.Length];
            for (int index = 0; index < this.data.Length; ++index)
                concatData[index] = this.data[index];
            for (int index = 0; index < genericString.data.Length; ++index)
                concatData[this.data.Length + index] = genericString.data[index];
            return GenericString<T>.Wrap(concatData);
        }

        /// <summary>
        /// Creates a new GenericString&lt;T&gt; instance that is a substring 
        /// of this generic string starting in the beginning. It is a prefix.
        /// </summary>
        /// <param name="length">Length of the prefix.</param>
        /// <returns>Prefix of this generic string.</returns>
        public GenericString<T> SubString(int length)
        {
            return this.SubString(0, length);
        }

        /// <summary>
        /// Creates a new GenericString&lt;T&gt; instance that is a substring
        /// of this generic string starting in the startIndex.
        /// </summary>
        /// <param name="startIndex">Starting index of the substring (first token has index 0).</param>
        /// <param name="length">Length of the substring.</param>
        /// <returns>Substring of this GenericString&lt;T&gt;.</returns>
        public GenericString<T> SubString(int startIndex, int length)
        {
            T[] subString = new T[length];
            for (int index = 0; index < length; ++index)
                subString[index] = this.data[startIndex + index];
            return GenericString<T>.Wrap(subString);
        }

        /// <summary>
        /// Creates a new GenericString&lt;T&gt; instance that is a prefix
        /// of this generic string.
        /// </summary>
        /// <param name="length">Length of the prefix.</param>
        /// <returns>Prefix of this generic string.</returns>
        public GenericString<T> Prefix(int length)
        {
            return this.SubString(0, length);
        }

        /// <summary>
        /// Creates a new GenericString&lt;T&gt; instance that is a postfix
        /// of this generic string.
        /// </summary>
        /// <param name="length">Length of the postfix.</param>
        /// <returns>Postfix of this generic string.</returns>
        public GenericString<T> Postfix(int length)
        {
            return this.SubString(data.Length - length, length);
        }

        /// <summary>
        /// Copies the elements of the GenericString&lt;T&gt; to a new array. 
        /// </summary>
        public T[] ToArray()
        {
            return this.data.Clone() as T[];
        }

        /// <summary>
        /// Returns a String that represents the current Object. (Inherited from Object.)
        /// </summary>
        public override string ToString()
        {
            return this.ToString(null);
        }

        /// <summary>
        /// ToString method with given separator of tokens.
        /// </summary>
        public string ToString(string separator)
        {
            if (separator == null)
            {
                if (typeof(T) == typeof(char)) separator = "";
                else separator = " ";
            }
            List<string> stringList = new List<string>(this.data.Length);
            foreach (T token in this.data)
                stringList.Add(token.ToString());
            return String.Join(separator, stringList.ToArray());
        }
        #endregion

        #region Private Data Members
        private T[] data;
        #endregion
    }

    /// <summary>
    /// Immutable class that represents a generic alphabet that consists of tokens of type T.
    /// </summary>
    /// <example>
    /// <code>
    /// Alphabet&lt;char&gt; alphabet = new Alphabet&lt;char&gt;("abcd");
    /// int indexOfCharC = alphabet.IndexOf('c');
    /// if (indexOfCharC >= 0)
    ///     Console.WriteLine("The alphabet contains the character 'c'.");
    /// </code>
    /// </example>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class Alphabet<T> : 
        IAlphabetData<T>,
        ICloneable,
        IEnumerable<T>,
        IEquatable<Alphabet<T>>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of the Alphabet&lt;T&gt; class
        /// that contains all the tokens that are found in the input string.
        /// </summary>
        /// <param name="tokens">Input generic string.</param>
        public Alphabet(IEnumerable<T> tokens)
        {
            const int minimalCapacity = 32;
            List<T> tokenList = new List<T>(minimalCapacity);
            tokenList.AddRange(tokens);
            tokenList.Sort();
            tokenList = Unique<T>.UniqueList(tokenList);
            this.data = tokenList.ToArray();
            this.CheckConsistency();
        }

        /// <summary>
        /// Initializes a new instance of the Alphabet&lt;T&gt; class
        /// based on the alphabet data provider which implements
        /// IAlphabetData&lt;T&gt; interface.
        /// </summary>
        /// <param name="alphabetData">Alphabet data provider.</param>
        public Alphabet(IAlphabetData<T> alphabetData)
        {
            this.data = alphabetData.GetAlphabetData();
            this.CheckConsistency();
        }
        #endregion

        #region IAlphabetData Interface
        /// <summary>
        /// Returns data describing this alphabet.
        /// </summary>
        public T[] GetAlphabetData()
        {
            return this.ToArray();
        }
        #endregion

        #region ICloneable Interface
        /// <summary>
        /// Returns a reference to this instance of Alphabet&lt;T&gt;.
        /// </summary>
        public object Clone()
        {
            return this;
        }
        #endregion

        #region IEnumerable Interface
        /// <summary>
        /// Returns an enumerator that iterates through the Alphabet&lt;T&gt;. 
        /// </summary>
        IEnumerator<T> IEnumerable<T>.GetEnumerator()
        {
            foreach (T token in this.data)
                yield return token;
        }

        System.Collections.IEnumerator
            System.Collections.IEnumerable.GetEnumerator()
        {
            throw new NotImplementedException();
        }
        #endregion

        #region IEquatable Interface
        /// <summary>
        /// Overloaded. Determines whether two Object instances are equal. (Inherited from Object.)
        /// </summary>
        public override bool Equals(object obj)
        {
            if (obj is Alphabet<T>)
                return this.Equals(obj as Alphabet<T>);
            else
                return false;
        }

        /// <summary>
        /// Determines whether two Alphabet&lt;T&gt; instances are equal.
        /// </summary>
        public bool Equals(Alphabet<T> alphabet)
        {
            if (this.data.Length != alphabet.data.Length) return false;
            for (int index = 0; index < this.data.Length; ++index)
                if (!this.data[index].Equals(alphabet.data[index]))
                    return false;
            return true;
        }

        /// <summary>
        /// Serves as a hash function for a particular type. GetHashCode is suitable for use in hashing algorithms 
        /// and data structures like a hash table. (Inherited from Object.)
        /// </summary>
        public override int GetHashCode()
        {
            int hashCode = 0;
            for (int index = 0; index < this.data.Length; ++index)
                hashCode = 15 * hashCode + this.data[index].GetHashCode() + index;
            return hashCode;
        }
        #endregion

        #region Indexer
        /// <summary>
        /// Gets an element at the specified index.
        /// </summary>
        public T this[int index]
        {
            get
            {
                return this.data[index];
            }
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets the number of different tokens actually contained in the Alphabet&lt;T&gt;.
        /// </summary>
        public int Count
        {
            get
            {
                return this.data.Length;
            }
        }
        #endregion

        #region Methods
        /// <summary>
        /// Decides whether the given input token is in the Alphabet&lt;T&gt;.
        /// </summary>
        /// <param name="token">Input token.</param>
        /// <returns>True if the input token is in the Alphabet&lt;T&gt;.</returns>
        public bool Contains(T token)
        {
            return (this.IndexOf(token) >= 0);
        }

        /// <summary>
        /// Decides whether the given generic string has all the tokens from the Alphabet&lt;T&gt;.
        /// </summary>
        /// <param name="tokens">Input generic string.</param>
        /// <returns>True if the input generic string has all the tokens from the Alphabet&lt;T&gt;.</returns>
        public bool Contains(IEnumerable<T> tokens)
        {
            foreach (T token in tokens)
                if (!this.Contains(token))
                    return false;
            return true;
        }

        /// <summary>
        /// Checks whether the given letter token belongs to the Alphabet&lt;T&gt;.
        /// </summary>
        /// <param name="token">Input token.</param>
        public void CheckWord(T token)
        {
            if (!this.Contains(token))
                throw new ArgumentException("The letter must be from the alphabet.");
        }

        /// <summary>
        /// Checks whether the given generic string has all the tokens from the Alphabet&lt;T&gt;.
        /// </summary>
        /// <param name="tokens">Input generic string.</param>
        public void CheckWord(IEnumerable<T> tokens)
        {
            if (!this.Contains(tokens))
                throw new ArgumentException("The letters of the word must be all from the alphabet.");
        }

        /// <summary>
        /// Checks whether the given alphabet is equal to this alphabet.
        /// </summary>
        /// <param name="alphabet">Input alphabet to compare with.</param>
        public void CheckAlphabet(Alphabet<T> alphabet)
        {
            if (!this.Equals(alphabet))
                throw new ArgumentException("The alphabets are different.");
        }

        /// <summary>
        /// Finds an index of the input token.
        /// </summary>
        /// <param name="token">Input token.</param>
        /// <returns>The index of the input token. Negative if the input token is not in the Alphabet&lt;T&gt;.</returns>
        public int IndexOf(T token)
        {
            return Array.BinarySearch<T>(this.data, token);
        }

        /// <summary>
        /// Copies the tokens of the Alphabet&lt;T&gt; to a new array. 
        /// </summary>
        public T[] ToArray()
        {
            return data.Clone() as T[];
        }

        /// <summary>
        /// Returns a String that represents the current Object. (Inherited from Object.)
        /// </summary>
        public override string ToString()
        {
            List<string> stringList = new List<string>(this.data.Length);
            foreach (T token in this.data)
                stringList.Add(token.ToString());
            return String.Join(", ", stringList.ToArray());
        }
        #endregion

        #region Private Methods
        /// <summary>
        /// Checks the consistency of the private data members.
        /// </summary>
        private void CheckConsistency()
        {
            if (this.data.Length == 0)
                throw new InvalidProgramException("The alphabet must contain at least one token.");
            for (int tokenIndex = 0; tokenIndex < this.data.Length - 1; ++tokenIndex)
                if (this.data[tokenIndex].CompareTo(this.data[tokenIndex + 1]) >= 0)
                    throw new InvalidProgramException("The tokens of the alphabet must be ordered upwardly.");
        }
        #endregion

        #region Private Data Members
        private T[] data;
        #endregion
    }

    /// <summary>
    /// Class that represents a deterministic finite-state automaton.
    /// </summary>
    /// <example>
    /// <code>
    /// Alphabet&lt;char&gt; alphabet = new Alphabet&lt;char&gt;("ab");
    /// DFA&lt;char&gt; dfa = new DFA&lt;char&gt;(alphabet, 3);
    /// dfa.SetArc(0, 1, 'a'); dfa.SetArc(0, 0, 'b');
    /// dfa.SetArc(1, 2, 'a'); dfa.SetArc(1, 1, 'b');
    /// dfa.SetArc(2, 0, 'a'); dfa.SetArc(2, 2, 'b');
    /// dfa.SetAcceptState(0);
    /// </code>
    /// </example>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class DFA<T> :
        IDFAData<T>,
        ICloneable,
        IEquatable<DFA<T>>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of the DFA&lt;T&gt; class
        /// with given alphabet and number of states.
        /// </summary>
        /// <param name="alphabet">Alphabet of the dfa.</param>
        /// <param name="stateCount">Number of states of the dfa.</param>
        public DFA(Alphabet<T> alphabet, int stateCount)
        {
            this.alphabet = alphabet;
            this.alphabetCount = alphabet.Count;
            this.stateCount = stateCount;
            this.acceptStates = new bool[this.stateCount];
            this.transitionTable = new int[this.stateCount, this.alphabetCount];
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
            {
                this.acceptStates[stateIndex] = false;
                for (int tokenIndex = 0; tokenIndex < this.alphabetCount; ++tokenIndex)
                    this.transitionTable[stateIndex, tokenIndex] = 0;
            }
            this.comment = "";
            this.CheckConsistency();
        }

        /// <summary>
        /// Initializes a new instance of the DFA&lt;T&gt; class
        /// based on the DFA data provider which implements
        /// IDFAData&lt;T&gt; interface.
        /// </summary>
        /// <param name="dfaData">DFA data provider.</param>
        public DFA(IDFAData<T> dfaData)
        {
            this.alphabet = new Alphabet<T>(dfaData);
            this.alphabetCount = this.alphabet.Count;
            this.stateCount = dfaData.GetDFAStateCountData();
            this.acceptStates = new bool[this.stateCount];
            this.transitionTable = new int[this.stateCount, this.alphabetCount];
            char[] separators = { ' ', '\t', '\r', '\n' };
            string[] acceptStateList =
                dfaData.GetDFAAcceptStatesData().Split(separators, StringSplitOptions.RemoveEmptyEntries);
            string[] transitionTableList =
                dfaData.GetDFATransitionTableData().Split(separators, StringSplitOptions.RemoveEmptyEntries);
            if (transitionTableList.Length != this.alphabetCount * this.stateCount)
                throw new InvalidProgramException("Invalid transition table.");
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
            {
                this.acceptStates[stateIndex] = false;
                for (int tokenIndex = 0; tokenIndex < this.alphabetCount; ++tokenIndex)
                    this.transitionTable[stateIndex, tokenIndex] =
                        Int32.Parse(transitionTableList[
                            this.alphabetCount * stateIndex + tokenIndex]);
            }
            for (int acceptIndex = 0; acceptIndex < acceptStateList.Length; ++acceptIndex)
                this.acceptStates[Int32.Parse(acceptStateList[acceptIndex])] = true;
            this.comment = dfaData.GetDFACommentData();
            this.CheckConsistency();
        }
        #endregion

        #region Static Constructors
        /// <summary>
        /// Returns a new instance of the DFA&lt;T&gt; class
        /// with given alphabet that accepts all the words from
        /// the given alphabet.
        /// </summary>
        public static DFA<T> CreateAcceptingDFA(Alphabet<T> alphabet)
        {
            DFA<T> dfa = new DFA<T>(alphabet, 1);
            dfa.acceptStates[0] = true;
            return dfa;
        }

        /// <summary>
        /// Returns a new instance of the DFA&lt;T&gt; class
        /// with given alphabet that rejects all the words from
        /// the given alphabet.
        /// </summary>
        public static DFA<T> CreateRejectingDFA(Alphabet<T> alphabet)
        {
            DFA<T> dfa = new DFA<T>(alphabet, 1);
            dfa.acceptStates[0] = false;
            return dfa;
        }
        #endregion

        #region IDFAData Interface
        /// <summary>
        /// Returns data describing the alphabet.
        /// </summary>
        public T[] GetAlphabetData()
        {
            return this.alphabet.GetAlphabetData();
        }

        /// <summary>
        /// Returns the number of states of this DFA.
        /// </summary>
        public int GetDFAStateCountData()
        {
            return this.stateCount;
        }

        /// <summary>
        /// Returns the string that consist of the indices of the accepted states
        /// separated by a single space.
        /// </summary>
        public string GetDFAAcceptStatesData()
        {
            List<string> acceptStateList = new List<string>(this.stateCount);
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                if (this.acceptStates[stateIndex])
                    acceptStateList.Add(stateIndex.ToString());
            return String.Join(" ", acceptStateList.ToArray());
        }

        /// <summary>
        /// Returns the content of the transition table as a string of numbers
        /// separated by a single space.
        /// </summary>
        public string GetDFATransitionTableData()
        {
            List<string> transitionTableList = 
                new List<string>(this.stateCount * this.alphabetCount);
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                for (int alphabetIndex = 0; alphabetIndex < this.alphabetCount; ++alphabetIndex)
                    transitionTableList.Add(
                        this.transitionTable[stateIndex, alphabetIndex].ToString());
            return String.Join(" ", transitionTableList.ToArray());
        }

        /// <summary>
        /// Returns a comment (a short description) of the dfa.
        /// </summary>
        public string GetDFACommentData()
        {
            return this.comment;
        }
        #endregion

        #region ICloneable Interface
        /// <summary>
        /// Returns a new instance of the DFA&lt;T&gt; class that is a deep copy of this instance.
        /// </summary>
        public object Clone()
        {
            DFA<T> dfa = new DFA<T>(this.alphabet, this.stateCount);
            for (int stateIndex = 0; stateIndex < dfa.stateCount; ++stateIndex)
            {
                dfa.acceptStates[stateIndex] = this.acceptStates[stateIndex];
                for (int tokenIndex = 0; tokenIndex < dfa.alphabetCount; ++tokenIndex)
                    dfa.transitionTable[stateIndex, tokenIndex] = 
                        this.transitionTable[stateIndex, tokenIndex];
            }
            dfa.comment = this.comment;
            return dfa;
        }
        #endregion

        #region IEquatable Interface
        /// <summary>
        /// Overloaded. Determines whether two Object instances are equal. (Inherited from Object.)
        /// </summary>
        public override bool Equals(object obj)
        {
            if (obj is DFA<T>)
                return this.Equals(obj as DFA<T>);
            else
                return false;
        }

        /// <summary>
        /// Decides whether this instance of DFA&lt;T&gt; is exactly the same as given dfa.
        /// This simple comparision is even stronger than isomorphism (it is an 1 : 1 comparision).
        /// </summary>
        /// <param name="dfa">Input dfa to compare.</param>
        /// <returns>True if this automaton is exactly the same as given dfa.</returns>
        public bool Equals(DFA<T> dfa)
        {
            if (!this.alphabet.Equals(dfa.alphabet)) return false;
            if (this.stateCount != dfa.stateCount) return false;
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                if (this.acceptStates[stateIndex] != dfa.acceptStates[stateIndex])
                    return false;
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                for (int tokenIndex = 0; tokenIndex < this.alphabetCount; ++tokenIndex)
                    if (this.transitionTable[stateIndex, tokenIndex] !=
                        dfa.transitionTable[stateIndex, tokenIndex])
                        return false;
            return true;
        }

        /// <summary>
        /// Serves as a hash function for a particular type. GetHashCode is suitable for use in hashing algorithms
        /// and data structures like a hash table. (Inherited from Object.)
        /// </summary>
        public override int GetHashCode()
        {
            int hashCode = this.alphabet.GetHashCode();
            hashCode = hashCode + this.alphabetCount * this.stateCount;
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                if (this.acceptStates[stateIndex])
                    hashCode = 3 * hashCode + stateIndex;
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                for (int tokenIndex = 0; tokenIndex < this.alphabetCount; ++tokenIndex)
                    hashCode = stateIndex * hashCode + 
                        tokenIndex * this.transitionTable[stateIndex, tokenIndex];
            return hashCode;
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets the alphabet of the DFA&lt;T&gt; class.
        /// </summary>
        public Alphabet<T> Alphabet
        {
            get
            {
                return this.alphabet;
            }
        }

        /// <summary>
        /// Gets the number of elements actually contained in the alphabet of the DFA&lt;T&gt;.
        /// </summary>
        public int AlphabetCount
        {
            get
            {
                return this.alphabetCount;
            }
        }

        /// <summary>
        /// Gets the number of states actually contained in the DFA&lt;T&gt;.
        /// </summary>
        public int StateCount
        {
            get
            {
                return this.stateCount;
            }
        }

        /// <summary>
        /// Gets or sets the comment (a short description) of the dfa.
        /// </summary>
        public string Comment
        {
            get
            {
                return this.comment;
            }
            set
            {
                this.comment = value;
            }
        }
        #endregion

        #region Methods
        /// <summary>
        /// Sets the state with an index stateIndex to be accepted.
        /// </summary>
        /// <param name="stateIndex">The index of the state.</param>
        public void SetAcceptState(int stateIndex)
        {
            this.acceptStates[stateIndex] = true;
        }

        /// <summary>
        /// Sets the state with an index stateIndex to be accepted or rejected.
        /// </summary>
        /// <param name="stateIndex">The index of the state.</param>
        /// <param name="stateAccept">Determines whether the state is accepted or rejected.</param>
        public void SetAcceptState(int stateIndex, bool stateAccept)
        {
            this.acceptStates[stateIndex] = stateAccept;
        }

        /// <summary>
        /// Sets the set of accepted states to be exactly the set of states
        /// with indexes in the acceptStateList.
        /// </summary>
        /// <param name="acceptStateList">The list of accepted states.</param>
        public void SetAcceptStates(IEnumerable<int> acceptStateList)
        {
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                this.acceptStates[stateIndex] = false;
            foreach (int state in acceptStateList)
                this.acceptStates[state] = true;
        }

        /// <summary>
        /// Returns an array of indexes of accepted states.
        /// </summary>
        public int[] GetAcceptStates()
        {
            List<int> acceptList = new List<int>(this.stateCount);
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                if (this.acceptStates[stateIndex])
                    acceptList.Add(stateIndex);
            return acceptList.ToArray();
        }

        /// <summary>
        /// Returns true if and only if the state with the index stateIndex is accepted.
        /// </summary>
        /// <param name="stateIndex">The index of the state.</param>
        public bool IsAcceptState(int stateIndex)
        {
            return this.acceptStates[stateIndex];
        }

        /// <summary>
        /// Sets the transition from one state to the other state through the given token.
        /// </summary>
        public void SetArc(int fromState, int toState, T token)
        {
            int tokenIndex = this.alphabet.IndexOf(token);
            this.transitionTable[fromState, tokenIndex] = toState;
        }

        /// <summary>
        /// Sets the transition from one state to the other state through the given index of a token.
        /// </summary>
        public void SetArcIndex(int fromState, int toState, int tokenIndex)
        {
            this.transitionTable[fromState, tokenIndex] = toState;
        }

        /// <summary>
        /// Returns the index of the state where you get if you go from the given state
        /// through the given token.
        /// </summary>
        public int GetArc(int fromState, T token)
        {
            int tokenIndex = this.alphabet.IndexOf(token);
            return this.transitionTable[fromState, tokenIndex];
        }

        /// <summary>
        /// Returns the index of the state where you get if you go from the given state
        /// through the given index of a token.
        /// </summary>
        public int GetArcIndex(int fromState, int tokenIndex)
        {
            return this.transitionTable[fromState, tokenIndex];
        }

        /// <summary>
        /// Returns the index of the state where you get if you go from starting state 0
        /// and reads the given generic string.
        /// </summary>
        /// <param name="tokens">Input generic string.</param>
        public int Transition(IEnumerable<T> tokens)
        {
            int stateIndex = 0;
            foreach (T token in tokens)
                stateIndex = this.GetArc(stateIndex, token);
            return stateIndex;
        }

        /// <summary>
        /// Returns true if and only if the input generic string is accepted by the DFA&lt;T&gt;.
        /// </summary>
        /// <param name="tokens">Input generic string.</param>
        public bool Decide(IEnumerable<T> tokens)
        {
            return this.acceptStates[this.Transition(tokens)];
        }

        /// <summary>
        /// Creates a new DFA&lt;T&gt; from this automaton that is normalized
        /// and without unreachable states.
        /// </summary>
        /// <returns>Normalized dfa without unreachable states.</returns>
        public DFA<T> Normalize()
        {
            int actualSourceState = 0;
            int actualDestState = 0;
            int totalDestAdded = 0;
            List<int> sourceToDest = new List<int>(this.stateCount);
            List<int> destToSource = new List<int>(this.stateCount);
            List<int> destAccept = new List<int>(this.stateCount);
            int[,] destStates = new int[this.stateCount, this.alphabetCount];
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                sourceToDest.Add(0);
            destToSource.Add(0);
            while (actualDestState <= totalDestAdded)
            {
                actualSourceState = destToSource[actualDestState];
                if (this.acceptStates[actualSourceState])
                    destAccept.Add(actualDestState);
                for (int tokenIndex = 0; tokenIndex < this.alphabetCount; ++tokenIndex)
                {
                    if (this.transitionTable[actualSourceState, tokenIndex] == 0)
                        destStates[actualDestState, tokenIndex] = 0;
                    else
                    {
                        if (sourceToDest[this.transitionTable[actualSourceState, tokenIndex]] == 0)
                        {
                            sourceToDest[this.transitionTable[actualSourceState, tokenIndex]] = ++totalDestAdded;
                            destToSource.Add(this.transitionTable[actualSourceState, tokenIndex]);
                        }
                        destStates[actualDestState, tokenIndex] =
                            sourceToDest[this.transitionTable[actualSourceState, tokenIndex]];
                    }
                }
                ++actualDestState;
            }
            DFA<T> destDFA = new DFA<T>(this.alphabet, actualDestState);
            foreach (int acceptState in destAccept)
                destDFA.acceptStates[acceptState] = true;
            for (int destStateIndex = 0; destStateIndex < actualDestState; ++destStateIndex)
                for (int destTokenIndex = 0; destTokenIndex < this.alphabetCount; ++destTokenIndex)
                    destDFA.transitionTable[destStateIndex, destTokenIndex] =
                        destStates[destStateIndex, destTokenIndex];
            return destDFA;
        }

        /// <summary>
        /// Creates a new DFA&lt;T&gt; from this automaton that does not contain equivalent states.
        /// </summary>
        /// <returns>Dfa without equivalent states.</returns>
        public DFA<T> EraseEquivalentStates()
        {
            int[] eqClass1 = new int[this.stateCount];
            int[] eqClass2 = new int[this.stateCount];
            SortedDictionary<GenericString<int>, int> eqMap =
                new SortedDictionary<GenericString<int>, int>();
            int classAcc = 0;
            int classRej = 0;
            if (this.acceptStates[0])
            {
                classAcc = 0;
                classRej = 1;
            }
            else
            {
                classAcc = 1;
                classRej = 0;
            }
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
            {
                if (this.acceptStates[stateIndex])
                    eqClass1[stateIndex] = classAcc;
                else
                    eqClass1[stateIndex] = classRej;
                eqClass2[stateIndex] = 0;
            }
            int classCount = 2;
            int oldCount = 0;
            while (classCount > oldCount)
            {
                oldCount = classCount;
                classCount = 0;
                eqMap.Clear();
                for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                {
                    int[] eqVector = new int[this.alphabetCount + 1];
                    eqVector[0] = eqClass1[stateIndex];
                    for (int tokenIndex = 0; tokenIndex < this.alphabetCount; ++tokenIndex)
                        eqVector[tokenIndex + 1] = eqClass1[this.transitionTable[stateIndex, tokenIndex]];
                    GenericString<int> eqVectorWrapper = GenericString<int>.Wrap(eqVector);
                    if (!eqMap.ContainsKey(eqVectorWrapper))
                        eqMap.Add(eqVectorWrapper, classCount++);
                    eqClass2[stateIndex] = eqMap[eqVectorWrapper];
                }
                int[] eqClassSwap = eqClass1;
                eqClass1 = eqClass2;
                eqClass2 = eqClassSwap;
            }
            DFA<T> destDFA = new DFA<T>(this.alphabet, classCount);
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                if (acceptStates[stateIndex])
                    destDFA.acceptStates[eqClass1[stateIndex]] = true;
            foreach (GenericString<int> eqRow in eqMap.Keys)
                for (int tokenIndex = 0; tokenIndex < this.alphabetCount; ++tokenIndex)
                    destDFA.transitionTable[eqRow[0], tokenIndex] = eqRow[tokenIndex + 1];
            return destDFA;
        }

        /// <summary>
        /// Returns the smallest generic word (with respect to the standard enumeration of strings)
        /// that is accepted in one automaton and rejected in the other automaton or
        /// null if these automata are equivalent.
        /// </summary>
        public GenericString<T> FindConflictingWord(DFA<T> dfa)
        {
            return DFA<T>.FindConflictingWord(this, dfa);
        }

        /// <summary>
        /// Returns a String that represents the current Object. (Inherited from Object.)
        /// </summary>
        public override string ToString()
        {
            const int minimalCapacity = 512;
            StringBuilder stringBuilder = new StringBuilder(minimalCapacity);
            stringBuilder.Append("   \t     ");
            foreach (T token in this.alphabet)
                stringBuilder.Append(String.Format("{0,5}", token.ToString()));
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
            {
                string prefix = null;
                if (this.acceptStates[stateIndex])
                {
                    if (stateIndex == 0) prefix = "\r\n<->\t";
                    else prefix = "\r\n<- \t";
                }
                else
                {
                    if (stateIndex == 0) prefix = "\r\n ->\t";
                    else prefix = "\r\n   \t";
                }
                stringBuilder.Append(prefix);
                stringBuilder.Append(String.Format("{0,5}", stateIndex.ToString()));
                for (int tokenIndex = 0; tokenIndex < this.alphabetCount; ++tokenIndex)
                    stringBuilder.Append(String.Format("{0,5}", this.transitionTable[stateIndex, tokenIndex]));
            }
            return stringBuilder.ToString();
        }
        #endregion

        #region Private Methods
        /// <summary>
        /// Checks the consistency of the private data members.
        /// </summary>
        private void CheckConsistency()
        {
            if (stateCount <= 0)
                throw new InvalidProgramException("The DFA must contain at least one state.");
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                for (int tokenIndex = 0; tokenIndex < this.alphabetCount; ++tokenIndex)
                    if (this.transitionTable[stateIndex, tokenIndex] < 0 ||
                        this.transitionTable[stateIndex, tokenIndex] >= this.stateCount)
                        throw new InvalidProgramException("Invalid transition table.");
        }
        #endregion

        #region Static Methods
        private struct StatePair : IComparable<StatePair>
        {
            public int state1;
            public int state2;
            public StatePair(int state1, int state2)
            {
                this.state1 = state1;
                this.state2 = state2;
            }
            public int CompareTo(StatePair other)
            {
                if (this.state1 == other.state1)
                    return this.state2 - other.state2;
                else return this.state1 - other.state1;
            }
        }

        /// <summary>
        /// Returns the smallest generic word (with respect to the standard enumeration of strings)
        /// that is accepted in one automaton and rejected in the other automaton.
        /// </summary>
        /// <param name="dfa1">First automaton.</param>
        /// <param name="dfa2">Second automaton.</param>
        /// <returns>
        /// GenericString&lt;T&gt; that is accepted in one automaton and rejected in the other automaton.
        /// Null if these automata are equivalent.
        /// </returns>
        public static GenericString<T> FindConflictingWord(DFA<T> dfa1, DFA<T> dfa2)
        {
            if (!dfa1.alphabet.Equals(dfa2.alphabet))
                return null;
            Alphabet<T> alphabet = dfa1.alphabet;
            int alphabetCount = dfa1.alphabetCount;
            SortedDictionary<StatePair, T[]> statePairMap = new SortedDictionary<StatePair, T[]>();
            Queue<StatePair> statePairQueue = new Queue<StatePair>(dfa1.stateCount * dfa2.stateCount);
            StatePair startStatePair = new StatePair(0, 0);
            statePairMap.Add(startStatePair, new T[0]);
            statePairQueue.Enqueue(startStatePair);
            while (statePairQueue.Count > 0)
            {
                StatePair statePair = statePairQueue.Dequeue();
                T[] statePairTokens = statePairMap[statePair];
                bool acceptState1 = dfa1.acceptStates[statePair.state1];
                bool acceptState2 = dfa2.acceptStates[statePair.state2];
                if (acceptState1 != acceptState2)
                    return GenericString<T>.Wrap(statePairTokens);
                for (int tokenIndex = 0; tokenIndex < alphabetCount; ++tokenIndex)
                {
                    StatePair newStatePair = new StatePair(
                        dfa1.transitionTable[statePair.state1, tokenIndex],
                        dfa2.transitionTable[statePair.state2, tokenIndex]);
                    if (!statePairMap.ContainsKey(newStatePair))
                    {
                        T[] newStatePairTokens = new T[statePairTokens.Length + 1];
                        statePairTokens.CopyTo(newStatePairTokens, 0);
                        newStatePairTokens[statePairTokens.Length] = alphabet[tokenIndex];
                        statePairMap.Add(newStatePair, newStatePairTokens);
                        statePairQueue.Enqueue(newStatePair);
                    }
                }
            }
            return null;
        }
        #endregion

        #region Private Data Members
        private Alphabet<T> alphabet;
        private int alphabetCount;
        private int stateCount;
        private bool[] acceptStates;
        private int[,] transitionTable;
        private string comment;
        #endregion
    }

    /// <summary>
    /// Struct that encapsulates a basic piece of information in the knowledge base.
    /// (A word together with the information whether it is accepted or rejected.)
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public struct WordAccept<T> :
        IComparable<WordAccept<T>>,
        IEquatable<WordAccept<T>>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of the WordAccept&lt;T&gt; struct.
        /// </summary>
        public WordAccept(GenericString<T> word, bool accept)
        {
            this.word = word;
            this.accept = accept;
        }
        #endregion

        /// Word in the knowledge base.
        public GenericString<T> word;
        /// Acceptance of the word.
        public bool accept;

        #region IComparable Interface
        /// <summary>
        /// Standard enumeration based only on the generic word.
        /// </summary>
        public int CompareTo(WordAccept<T> wordAccept)
        {
            return this.word.CompareTo(wordAccept.word);
        }
        #endregion

        #region IEquatable Interface
        /// <summary>
        /// Overloaded. Determines whether two Object instances are equal. (Inherited from Object.)
        /// </summary>
        public override bool Equals(object obj)
        {
            if (obj is WordAccept<T>)
                return this.Equals((WordAccept<T>)obj);
            else
                return false;
        }

        /// <summary>
        /// Determines whether two WordAccept&lt;T&gt; instances are equal (based only on the generic word).
        /// </summary>
        public bool Equals(WordAccept<T> wordAccept)
        {
            return this.word.Equals(wordAccept.word);
        }

        /// <summary>
        /// Serves as a hash function for a particular type. GetHashCode is suitable for use in hashing algorithms
        /// and data structures like a hash table. (Inherited from Object.)
        /// </summary>
        public override int GetHashCode()
        {
            return this.word.GetHashCode();
        }
        #endregion

        #region Methods
        /// <summary>
        /// Returns a String that represents the current Object. (Inherited from Object.)
        /// </summary>
        public override string ToString()
        {
            return this.ToString(null);
        }

        /// <summary>
        /// ToString method with given separator of tokens.
        /// </summary>
        public string ToString(string separator)
        {
            if (this.accept)
                return "accept " + this.word.ToString(separator);
            else return "reject " + this.word.ToString(separator);
        }
        #endregion
    }

    /// <summary>
    /// KnowledgeBaseChangeType enum represents a type of the KnowledgeBaseChanged event that
    /// can occur in the KnowledgeBase&lt;T&gt; class.
    /// </summary>
    public enum KnowledgeBaseChangeType
    {
        /// Knowledge (word + accept) was added to the knowledge base.
        KnowledgeAdded,
        /// Acceptance of a word was changed in the knowledge base.
        KnowledgeModified,
        /// Knowledge base was cleared.
        KnowledgeBaseCleared,
        /// Knowledge base was sorted.
        KnowledgeBaseSorted
    }

    /// <summary>
    /// KnowledgeBaseChangedEventHandler is a delegate for the KnowledgeBaseChanged event
    /// in the KnowledgeBase&lt;T&gt; class.
    /// </summary>
    /// <param name="knowledgeBaseChangeType">Type of the change that occured in the knowledge base.</param>
    public delegate void KnowledgeBaseChangedEventHandler(
        KnowledgeBaseChangeType knowledgeBaseChangeType);

    /// <summary>
    /// Class that encapsulates a knowledge base (a list of words together with 
    /// the information whether they are accepted or rejected).
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class KnowledgeBase<T> :
        IKnowledgeBaseData<T>,
        ICloneable,
        IEnumerable<WordAccept<T>>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of the KnowledgeBase&lt;T&gt; class with given alphabet.
        /// </summary>
        /// <param name="alphabet">Alphabet of the knowledge base.</param>
        public KnowledgeBase(Alphabet<T> alphabet)
        {
            const int minimalCapacity = 32;
            this.alphabet = alphabet;
            this.data = new List<WordAccept<T>>(minimalCapacity);
            this.cache = new SortedDictionary<GenericString<T>, bool>();
            this.comment = "";
            this.CheckConsistency();
        }

        /// <summary>
        /// Initializes a new instance of the KnowledgeBase&lt;T&gt; class
        /// based on the KnowledgeBase data provider which implements
        /// IKnowledgeBaseData&lt;T&gt; interface.
        /// </summary>
        /// <param name="knowledgeBaseData">KnowledgeBase data provider.</param>
        public KnowledgeBase(IKnowledgeBaseData<T> knowledgeBaseData)
        {
            this.alphabet = new Alphabet<T>(knowledgeBaseData);
            WordAcceptData<T>[] wordAcceptDataList = knowledgeBaseData.GetKnowledgeBaseData();
            this.data = new List<WordAccept<T>>(wordAcceptDataList.Length);
            this.cache = new SortedDictionary<GenericString<T>, bool>();
            foreach (WordAcceptData<T> wordAcceptData in wordAcceptDataList)
            {
                WordAccept<T> wordAccept = new WordAccept<T>();
                wordAccept.word = GenericString<T>.Wrap(wordAcceptData.word);
                wordAccept.accept = wordAcceptData.accept;
                this.data.Add(wordAccept);
                this.cache.Add(wordAccept.word, wordAccept.accept);
            }
            this.comment = knowledgeBaseData.GetKnowledgeBaseCommentData();
            this.CheckConsistency();
        }
        #endregion

        #region IKnowledgeBaseData Interface
        /// <summary>
        /// Returns data describing the alphabet.
        /// </summary>
        public T[] GetAlphabetData()
        {
            return this.alphabet.GetAlphabetData();
        }

        /// <summary>
        /// Returns an array of WordAcceptData&lt;T&gt; structures describing the whole knowledge base.
        /// </summary>
        public WordAcceptData<T>[] GetKnowledgeBaseData()
        {
            List<WordAcceptData<T>> wordAcceptDataList =
                new List<WordAcceptData<T>>(this.data.Count);
            foreach (WordAccept<T> wordAccept in this.data)
            {
                WordAcceptData<T> wordAcceptData = new WordAcceptData<T>();
                wordAcceptData.word = wordAccept.word.ToArray();
                wordAcceptData.accept = wordAccept.accept;
                wordAcceptDataList.Add(wordAcceptData);
            }
            return wordAcceptDataList.ToArray();
        }

        /// <summary>
        /// Returns a comment (a short description) of the knowledge base.
        /// </summary>
        public string GetKnowledgeBaseCommentData()
        {
            return this.comment;
        }
        #endregion

        #region ICloneable Interface
        /// <summary>
        /// Returns a new instance of the KnowledgeBase&lt;T&gt; class that is a deep copy of this instance.
        /// </summary>
        public object Clone()
        {
            KnowledgeBase<T> knowledgeBase = new KnowledgeBase<T>(this.alphabet);
            foreach (WordAccept<T> wordAccept in this.data)
            {
                knowledgeBase.data.Add(wordAccept);
                knowledgeBase.cache.Add(wordAccept.word, wordAccept.accept);
            }
            knowledgeBase.comment = this.comment;
            return knowledgeBase;
        }
        #endregion

        #region IEnumerable Interface
        /// <summary>
        /// Returns an enumerator that iterates through the KnowledgeBase&lt;T&gt;. 
        /// </summary>
        IEnumerator<WordAccept<T>> IEnumerable<WordAccept<T>>.GetEnumerator()
        {
            foreach (WordAccept<T> wordAccept in this.data)
                yield return wordAccept;
        }

        System.Collections.IEnumerator
            System.Collections.IEnumerable.GetEnumerator()
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Class that filters the words in the knowledge base by their acceptance.
        /// </summary>
        private class WordAcceptFilter :
            IEnumerable<GenericString<T>>
        {
            public WordAcceptFilter(IEnumerable<WordAccept<T>> wordAcceptCollection, bool accept)
            {
                this.wordAcceptCollection = wordAcceptCollection;
                this.accept = accept;
            }

            IEnumerator<GenericString<T>> IEnumerable<GenericString<T>>.GetEnumerator()
            {
                foreach (WordAccept<T> wordAccept in this.wordAcceptCollection)
                    if (wordAccept.accept == this.accept)
                        yield return wordAccept.word;
            }

            System.Collections.IEnumerator
                System.Collections.IEnumerable.GetEnumerator()
            {
                throw new NotImplementedException();
            }

            private IEnumerable<WordAccept<T>> wordAcceptCollection;
            private bool accept;
        }
        #endregion

        #region Indexer
        /// <summary>
        /// Gets an element at the specified index.
        /// </summary>
        public WordAccept<T> this[int index]
        {
            get
            {
                return this.data[index];
            }
        }

        /// <summary>
        /// Gets the acceptance of the specified generic word.
        /// </summary>
        public bool this[GenericString<T> word]
        {
            get
            {
                return this.cache[word];
            }
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets the alphabet of the KnowledgeBase&lt;T&gt; class.
        /// </summary>
        public Alphabet<T> Alphabet
        {
            get
            {
                return this.alphabet;
            }
        }

        /// <summary>
        /// Gets the number of elements actually contained in the KnowledgeBase&lt;T&gt;.
        /// </summary>
        public int Count
        {
            get
            {
                return this.data.Count;
            }
        }

        /// <summary>
        /// Gets the read-only collection of known words.
        /// </summary>
        public ReadOnlyCollection<WordAccept<T>> Knowledge
        {
            get
            {
                return new ReadOnlyCollection<WordAccept<T>>(this.data);
            }
        }

        /// <summary>
        /// Gets the collection of accepted words.
        /// </summary>
        public IEnumerable<GenericString<T>> AcceptedWords
        {
            get
            {
                return new WordAcceptFilter(this.data, true);
            }
        }

        /// <summary>
        /// Gets the collection of rejected words.
        /// </summary>
        public IEnumerable<GenericString<T>> RejectedWords
        {
            get
            {
                return new WordAcceptFilter(this.data, false);
            }
        }

        /// <summary>
        /// Gets or sets the comment (a short description) of the knowledge base.
        /// </summary>
        public string Comment
        {
            get
            {
                return this.comment;
            }
            set
            {
                this.comment = value;
            }
        }
        #endregion

        #region Events
        /// <summary>
        /// Event that occurs when the knowledge base is changed.
        /// </summary>
        public event KnowledgeBaseChangedEventHandler KnowledgeBaseChanged;

        /// <summary>
        /// Fires a KnowledgeBaseChanged event.
        /// </summary>
        /// <param name="knowledgeBaseChangeType">Type of the change that occured in the knowledge base.</param>
        public void OnKnowledgeBaseChanged(KnowledgeBaseChangeType knowledgeBaseChangeType)
        {
            if (this.KnowledgeBaseChanged != null)
                this.KnowledgeBaseChanged(knowledgeBaseChangeType);
        }
        #endregion

        #region Methods
        /// <summary>
        /// Represents the result of adding information to the knowledge base.
        /// </summary>
        public enum AddResult
        {
            /// Indicates that the information was already present in the knowledge base and nothing was added.
            NothingChanged,
            /// Indicates that the information was added to the knowledge base.
            ItemAdded,
            /// Indicates that the generic word was in the knowledge base and the acceptance of the word has been changed.
            ItemChanged
        }

        /// <summary>
        /// Adds a basic piece of information (a generic word and whether it is accepted or rejected)
        /// to the knowledge base.
        /// This method modifies the content of the knowledge base,
        /// so it can fire a KnowledgeBaseChanged event.
        /// </summary>
        /// <param name="word">Generic word.</param>
        /// <param name="accept">Whether the generic word is accepted or rejected.</param>
        /// <returns>
        /// Returns NothingChanged if and only if the information was already present in the knowledge base,
        /// ItemAdded if and only if the information has been added to the knowledge base,
        /// ItemChanged if and only if the generic word was in the knowledge base, but
        /// the acceptance of the word has been changed.
        /// </returns>
        public AddResult Add(GenericString<T> word, bool accept)
        {
            this.alphabet.CheckWord(word);
            WordAccept<T> wordAccept = new WordAccept<T>();
            wordAccept.word = word;
            wordAccept.accept = accept;
            if (this.cache.ContainsKey(word))
            {
                if (this.cache[word] != accept)
                {
                    int index = this.data.IndexOf(wordAccept);
                    this.data[index] = wordAccept;
                    this.cache[word] = accept;
                    this.OnKnowledgeBaseChanged(KnowledgeBaseChangeType.KnowledgeModified);
                    return AddResult.ItemChanged;
                }
                else return AddResult.NothingChanged;
            }
            else
            {
                this.data.Add(wordAccept);
                this.cache.Add(word, accept);
                this.OnKnowledgeBaseChanged(KnowledgeBaseChangeType.KnowledgeAdded);
                return AddResult.ItemAdded;
            }
        }

        /// <summary>
        /// Returns true if and only if the word is in the knowledge base.
        /// </summary>
        /// <param name="word">Generic word.</param>
        public bool Contains(GenericString<T> word)
        {
            return this.cache.ContainsKey(word);
        }

        /// <summary>
        /// Clears the entire knowledge base.
        /// This method modifies the content of the knowledge base,
        /// so it can fire a KnowledgeBaseChanged event.
        /// </summary>
        public void Clear()
        {
            this.data.Clear();
            this.cache.Clear();
            this.OnKnowledgeBaseChanged(KnowledgeBaseChangeType.KnowledgeBaseCleared);
        }

        /// <summary>
        /// Sorts the knowledge base. (Based on the standard enumeration.)
        /// This method modifies the content of the knowledge base,
        /// so it can fire a KnowledgeBaseChanged event.
        /// </summary>
        public void Sort()
        {
            this.data.Sort();
            this.OnKnowledgeBaseChanged(KnowledgeBaseChangeType.KnowledgeBaseSorted);
        }

        /// <summary>
        /// Returns a String that represents the current Object. (Inherited from Object.)
        /// </summary>
        public override string ToString()
        {
            return this.ToString(null);
        }

        /// <summary>
        /// ToString method with given separator of tokens.
        /// </summary>
        public string ToString(string separator)
        {
            StringBuilder stringBuilder = new StringBuilder();
            foreach (WordAccept<T> wordAccept in this.data)
            {
                stringBuilder.Append(wordAccept.ToString(separator));
                stringBuilder.Append("\r\n");
            }
            return stringBuilder.ToString().TrimEnd();
        }
        #endregion

        #region Private Methods
        /// <summary>
        /// Checks the consistency of the private data members.
        /// </summary>
        private void CheckConsistency()
        {
            foreach (WordAccept<T> wordAccept in this.data)
            {
                this.alphabet.CheckWord(wordAccept.word);
                if (!this.cache.ContainsKey(wordAccept.word) ||
                     this.cache[wordAccept.word] != wordAccept.accept)
                    throw new InvalidProgramException("Inconsistent knowledge base. " +
                        "Error is connected with the word: " + wordAccept.word.ToString());
            }
        }
        #endregion

        #region Private Data Members
        private Alphabet<T> alphabet;
        private List<WordAccept<T>> data;
        private SortedDictionary<GenericString<T>, bool> cache;
        private string comment;
        #endregion
    }

    /// <summary>
    /// PositiveNegativeSamplesOptions enum represents
    /// special options for PositiveNegativeSamples&lt;T&gt; class.
    /// </summary>
    public enum PositiveNegativeSamplesOptions
    {
        /// All samples (positive and negative) are allowed.
        AllSamplesEnabled,
        /// Only positive samples are allowed.
        PositiveSamplesEnabled,
        /// Only negative samples are allowed.
        NegativeSamplesEnabled
    }

    /// <summary>
    /// PositiveNegativeSamplesChangeType enum represents
    /// a type of the PositiveNegativeSamplesChanged event
    /// that can occur in the PositiveNegativeSamples&lt;T&gt; class.
    /// </summary>
    public enum PositiveNegativeSamplesChangeType
    {
        /// Positive sample was added to the PositiveNegativeSamples object.
        PositiveSampleAdded,
        /// Negative sample was added to the PositiveNegativeSamples object.
        NegativeSampleAdded,
        /// All positive and negative samples were cleared.
        PositiveNegativeSamplesCleared
    }

    /// <summary>
    /// PositiveNegativeSamplesChangedEventHandler is a delegate
    /// for the PositiveNegativeSamplesChanged event
    /// in the PositiveNegativeSamples&lt;T&gt; class.
    /// </summary>
    /// <param name="positiveNegativeSamplesChangeType">
    /// Type of the change that occured in the PositiveNegativeSamples object.
    /// </param>
    public delegate void PositiveNegativeSamplesChangedEventHandler(
        PositiveNegativeSamplesChangeType positiveNegativeSamplesChangeType);

    /// <summary>
    /// PositiveSampleAddedEventHandler is a delegate for PositiveSampleAdded 
    /// event in the PositiveNegativeSamples&lt;T&gt; class.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    /// <param name="positiveSample">Positive sample.</param>
    public delegate void PositiveSampleAddedEventHandler<T>(GenericString<T> positiveSample)
        where T : struct, IComparable<T>, IEquatable<T>;

    /// <summary>
    /// NegativeSampleAddedEventHandler is a delegate for NegativeSampleAdded 
    /// event in the PositiveNegativeSamples&lt;T&gt; class.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    /// <param name="negativeSample">Negative sample.</param>
    public delegate void NegativeSampleAddedEventHandler<T>(GenericString<T> negativeSample)
        where T : struct, IComparable<T>, IEquatable<T>;

    /// <summary>
    /// Class that encapsulates two lists: one for the positive samples 
    /// and the other for the negative samples.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class PositiveNegativeSamples<T> :
        IPositiveNegativeSamplesData<T>,
        ICloneable
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of the PositiveNegativeSamples&lt;T&gt; class with given alphabet.
        /// </summary>
        /// <param name="alphabet">Alphabet of the PositiveNegativeSamples&lt;T&gt; class.</param>
        /// <param name="options">Special options of PositiveNegativeSamples&lt;T&gt; class.</param>
        public PositiveNegativeSamples(Alphabet<T> alphabet,
            PositiveNegativeSamplesOptions options)
        {
            const int minimalCapacity = 32;
            this.alphabet = alphabet;
            this.options = options;
            this.positiveSamples = new List<GenericString<T>>(minimalCapacity);
            this.negativeSamples = new List<GenericString<T>>(minimalCapacity);
            this.cache = new SortedDictionary<GenericString<T>, bool>();
            this.comment = "";
            this.CheckConsistency();
        }

        /// <summary>
        /// Initializes a new instance of the PositiveNegativeSamples&lt;T&gt; class
        /// based on the PositiveNegativeSamples data provider which implements
        /// IPositiveNegativeSamplesData&lt;T&gt; interface.
        /// </summary>
        /// <param name="positiveNegativeSamplesData">PositiveNegativeSamples data provider.</param>
        public PositiveNegativeSamples(IPositiveNegativeSamplesData<T> positiveNegativeSamplesData)
        {
            this.alphabet = new Alphabet<T>(positiveNegativeSamplesData);
            this.options = positiveNegativeSamplesData.GetPNSamplesOptions();
            this.cache = new SortedDictionary<GenericString<T>, bool>();
            SampleData<T>[] positiveSamplesData = positiveNegativeSamplesData.GetPositiveSamplesData();
            SampleData<T>[] negativeSamplesData = positiveNegativeSamplesData.GetNegativeSamplesData();
            this.positiveSamples = new List<GenericString<T>>(positiveSamplesData.Length);
            this.negativeSamples = new List<GenericString<T>>(negativeSamplesData.Length);
            foreach (SampleData<T> sampleData in positiveSamplesData)
                this.AddPositiveSample(GenericString<T>.Wrap(sampleData.word));
            foreach (SampleData<T> sampleData in negativeSamplesData)
                this.AddNegativeSample(GenericString<T>.Wrap(sampleData.word));
            this.comment = positiveNegativeSamplesData.GetPNSamplesCommentData();
            this.CheckConsistency();
        }

        /// <summary>
        /// Initializes a new instance of the PositiveNegativeSamples&lt;T&gt; class
        /// based on the PositiveNegativeSamples data provider which implements
        /// IPositiveNegativeSamplesData&lt;T&gt; interface and additional special options.
        /// </summary>
        /// <param name="positiveNegativeSamplesData">PositiveNegativeSamples data provider.</param>
        /// <param name="options">Special options of PositiveNegativeSamples&lt;T&gt; class.</param>
        public PositiveNegativeSamples(IPositiveNegativeSamplesData<T> positiveNegativeSamplesData,
            PositiveNegativeSamplesOptions options)
        {
            this.alphabet = new Alphabet<T>(positiveNegativeSamplesData);
            this.options = options;
            this.cache = new SortedDictionary<GenericString<T>, bool>();
            SampleData<T>[] positiveSamplesData = positiveNegativeSamplesData.GetPositiveSamplesData();
            SampleData<T>[] negativeSamplesData = positiveNegativeSamplesData.GetNegativeSamplesData();
            this.positiveSamples = new List<GenericString<T>>(positiveSamplesData.Length);
            this.negativeSamples = new List<GenericString<T>>(negativeSamplesData.Length);
            foreach (SampleData<T> sampleData in positiveSamplesData)
                if (options != PositiveNegativeSamplesOptions.NegativeSamplesEnabled)
                    this.AddPositiveSample(GenericString<T>.Wrap(sampleData.word));
            foreach (SampleData<T> sampleData in negativeSamplesData)
                if (options != PositiveNegativeSamplesOptions.PositiveSamplesEnabled)
                    this.AddNegativeSample(GenericString<T>.Wrap(sampleData.word));
            this.comment = positiveNegativeSamplesData.GetPNSamplesCommentData();
            this.CheckConsistency();
        }

        /// <summary>
        /// Initializes a new instance of the PositiveNegativeSamples&lt;T&gt; class based 
        /// on the KnowledgeBase data provider which implements IKnowledgeBaseData&lt;T&gt; interface.
        /// </summary>
        /// <param name="knowledgeBaseData">KnowledgeBase data provider.</param>
        public PositiveNegativeSamples(IKnowledgeBaseData<T> knowledgeBaseData)
        {
            this.alphabet = new Alphabet<T>(knowledgeBaseData);
            this.options = PositiveNegativeSamplesOptions.AllSamplesEnabled;
            WordAcceptData<T>[] wordAcceptDataList = knowledgeBaseData.GetKnowledgeBaseData();
            this.positiveSamples = new List<GenericString<T>>(wordAcceptDataList.Length);
            this.negativeSamples = new List<GenericString<T>>(wordAcceptDataList.Length);
            this.cache = new SortedDictionary<GenericString<T>, bool>();
            foreach (WordAcceptData<T> wordAcceptData in wordAcceptDataList)
            {
                if (wordAcceptData.accept) this.AddPositiveSample(GenericString<T>.Wrap(wordAcceptData.word));
                else this.AddNegativeSample(GenericString<T>.Wrap(wordAcceptData.word));
            }
            this.comment = knowledgeBaseData.GetKnowledgeBaseCommentData();
            this.CheckConsistency();
        }

        /// <summary>
        /// Initializes a new instance of the PositiveNegativeSamples&lt;T&gt; class based 
        /// on the KnowledgeBase data provider which implements IKnowledgeBaseData&lt;T&gt; interface
        /// and additional special options.
        /// </summary>
        /// <param name="knowledgeBaseData">KnowledgeBase data provider.</param>
        /// <param name="options">Special options of PositiveNegativeSamples&lt;T&gt; class.</param>
        public PositiveNegativeSamples(IKnowledgeBaseData<T> knowledgeBaseData,
            PositiveNegativeSamplesOptions options)
        {
            this.alphabet = new Alphabet<T>(knowledgeBaseData);
            this.options = options;
            WordAcceptData<T>[] wordAcceptDataList = knowledgeBaseData.GetKnowledgeBaseData();
            this.positiveSamples = new List<GenericString<T>>(wordAcceptDataList.Length);
            this.negativeSamples = new List<GenericString<T>>(wordAcceptDataList.Length);
            this.cache = new SortedDictionary<GenericString<T>, bool>();
            foreach (WordAcceptData<T> wordAcceptData in wordAcceptDataList)
            {
                if (wordAcceptData.accept && options != PositiveNegativeSamplesOptions.NegativeSamplesEnabled)
                    this.AddPositiveSample(GenericString<T>.Wrap(wordAcceptData.word));
                else if (options != PositiveNegativeSamplesOptions.PositiveSamplesEnabled)
                    this.AddNegativeSample(GenericString<T>.Wrap(wordAcceptData.word));
            }
            this.comment = knowledgeBaseData.GetKnowledgeBaseCommentData();
            this.CheckConsistency();
        }
        #endregion

        #region IPositiveNegativeSamplesData Interface
        /// <summary>
        /// Returns data describing the alphabet.
        /// </summary>
        public T[] GetAlphabetData()
        {
            return this.alphabet.GetAlphabetData();
        }

        /// <summary>
        /// Returns special options for positive and negative samples.
        /// </summary>
        public PositiveNegativeSamplesOptions GetPNSamplesOptions()
        {
            return this.options;
        }

        /// <summary>
        /// Returns an array of SampleData&lt;T&gt; structures describing the set of positive samples.
        /// </summary>
        public SampleData<T>[] GetPositiveSamplesData()
        {
            List<SampleData<T>> positiveSamplesDataList =
                new List<SampleData<T>>(this.positiveSamples.Count);
            foreach (GenericString<T> positiveSample in this.positiveSamples)
            {
                SampleData<T> sampleData = new SampleData<T>();
                sampleData.word = positiveSample.ToArray();
                positiveSamplesDataList.Add(sampleData);
            }
            return positiveSamplesDataList.ToArray();
        }

        /// <summary>
        /// Returns an array of SampleData&lt;T&gt; structures describing the set of negative samples.
        /// </summary>
        public SampleData<T>[] GetNegativeSamplesData()
        {
            List<SampleData<T>> negativeSamplesDataList =
                new List<SampleData<T>>(this.negativeSamples.Count);
            foreach (GenericString<T> negativeSample in this.negativeSamples)
            {
                SampleData<T> sampleData = new SampleData<T>();
                sampleData.word = negativeSample.ToArray();
                negativeSamplesDataList.Add(sampleData);
            }
            return negativeSamplesDataList.ToArray();
        }

        /// <summary>
        /// Returns a comment (a short description) of the positive and negative samples.
        /// </summary>
        public string GetPNSamplesCommentData()
        {
            return this.comment;
        }
        #endregion

        #region ICloneable Interface
        /// <summary>
        /// Returns a new instance of the PositiveNegativeSamples&lt;T&gt; class that is a deep copy of this instance.
        /// </summary>
        public object Clone()
        {
            PositiveNegativeSamples<T> positiveNegativeSamples
                = new PositiveNegativeSamples<T>(this.alphabet, this.options);
            foreach (GenericString<T> positiveSample in this.positiveSamples)
                positiveNegativeSamples.AddPositiveSample(positiveSample);
            foreach (GenericString<T> negativeSample in this.negativeSamples)
                positiveNegativeSamples.AddNegativeSample(negativeSample);
            positiveNegativeSamples.comment = this.comment;
            return positiveNegativeSamples;
        }
        #endregion

        #region Indexer
        /// <summary>
        /// Gets the acceptance of the specified generic word.
        /// </summary>
        public bool this[GenericString<T> word]
        {
            get
            {
                return this.cache[word];
            }
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets the alphabet of the PositiveNegativeSamples&lt;T&gt; class.
        /// </summary>
        public Alphabet<T> Alphabet
        {
            get
            {
                return this.alphabet;
            }
        }

        /// <summary>
        /// Gets special options for positive and negative samples.
        /// </summary>
        public PositiveNegativeSamplesOptions Options
        {
            get
            {
                return this.options;
            }
        }

        /// <summary>
        /// Gets the collection of positive samples.
        /// </summary>
        public IEnumerable<GenericString<T>> PositiveSamples
        {
            get
            {
                return this.positiveSamples;
            }
        }

        /// <summary>
        /// Gets the collection of negative samples.
        /// </summary>
        public IEnumerable<GenericString<T>> NegativeSamples
        {
            get
            {
                return this.negativeSamples;
            }
        }

        /// <summary>
        /// Gets or sets the comment (a short description) of the positive and negative samples.
        /// </summary>
        public string Comment
        {
            get
            {
                return this.comment;
            }
            set
            {
                this.comment = value;
            }
        }
        #endregion

        #region Events
        /// <summary>
        /// Event that occurs when the positive or negative samples are changed.
        /// </summary>
        public event PositiveNegativeSamplesChangedEventHandler
            PositiveNegativeSamplesChanged;

        /// <summary>
        /// Event that occurs when the positive sample is added.
        /// </summary>
        public event PositiveSampleAddedEventHandler<T> PositiveSampleAdded;

        /// <summary>
        /// Event that occurs when the negative sample is added.
        /// </summary>
        public event NegativeSampleAddedEventHandler<T> NegativeSampleAdded;

        /// <summary>
        /// Fires a PositiveNegativeSamplesChanged event.
        /// </summary>
        /// <param name="positiveNegativeSamplesChangeType">
        /// Type of the change that occured in the PositiveNegativeSamples object.
        /// </param>
        public void OnPositiveNegativeSamplesChanged(
            PositiveNegativeSamplesChangeType positiveNegativeSamplesChangeType)
        {
            if (this.PositiveNegativeSamplesChanged != null)
                this.PositiveNegativeSamplesChanged(
                    positiveNegativeSamplesChangeType);
        }

        /// <summary>
        /// Fires a PositiveSampleAdded event.
        /// </summary>
        /// <param name="positiveSample">Positive sample.</param>
        public void OnPositiveSampleAdded(GenericString<T> positiveSample)
        {
            if (this.PositiveSampleAdded != null)
                this.PositiveSampleAdded(positiveSample);
        }

        /// <summary>
        /// Fires a NegativeSampleAdded event.
        /// </summary>
        /// <param name="negativeSample">Negative sample.</param>
        public void OnNegativeSampleAdded(GenericString<T> negativeSample)
        {
            if (this.NegativeSampleAdded != null)
                this.NegativeSampleAdded(negativeSample);
        }
        #endregion

        #region Methods
        /// <summary>
        /// Adds a positive sample to the PositiveNegativeSamples&lt;T&gt; class.
        /// </summary>
        public void AddPositiveSample(GenericString<T> positiveSample)
        {
            this.alphabet.CheckWord(positiveSample);
            if (this.options == PositiveNegativeSamplesOptions.NegativeSamplesEnabled)
                throw new InvalidProgramException("Only negative samples are allowed.");
            if (this.cache.ContainsKey(positiveSample))
            {
                if (this.cache[positiveSample] == false)
                    throw new InvalidProgramException(
                        "Error: Sample " + positiveSample.ToString() +
                        " is already set to be negative!");
            }
            else
            {
                this.positiveSamples.Add(positiveSample);
                this.cache[positiveSample] = true;
                this.OnPositiveNegativeSamplesChanged(
                    PositiveNegativeSamplesChangeType.PositiveSampleAdded);
                this.OnPositiveSampleAdded(positiveSample);
            }
        }

        /// <summary>
        /// Adds a negative sample to the PositiveNegativeSamples&lt;T&gt; class.
        /// </summary>
        public void AddNegativeSample(GenericString<T> negativeSample)
        {
            this.alphabet.CheckWord(negativeSample);
            if (this.options == PositiveNegativeSamplesOptions.PositiveSamplesEnabled)
                throw new InvalidProgramException("Only positive samples are allowed.");
            if (this.cache.ContainsKey(negativeSample))
            {
                if (this.cache[negativeSample] == true)
                    throw new InvalidProgramException(
                        "Error: Sample " + negativeSample.ToString() +
                        " is already set to be positive!");
            }
            else
            {
                this.negativeSamples.Add(negativeSample);
                this.cache[negativeSample] = false;
                this.OnPositiveNegativeSamplesChanged(
                    PositiveNegativeSamplesChangeType.NegativeSampleAdded);
                this.OnNegativeSampleAdded(negativeSample);
            }
        }

        /// <summary>
        /// Returns true if and only if the word is found inside the positive or the negative samples.
        /// </summary>
        /// <param name="word">Generic word.</param>
        public bool Contains(GenericString<T> word)
        {
            return this.cache.ContainsKey(word);
        }

        /// <summary>
        /// Clears all positive and negative samples.
        /// </summary>
        public void Clear()
        {
            this.positiveSamples.Clear();
            this.negativeSamples.Clear();
            this.cache.Clear();
            this.OnPositiveNegativeSamplesChanged(
                PositiveNegativeSamplesChangeType.PositiveNegativeSamplesCleared);
        }

        /// <summary>
        /// Returns a String that represents the current Object. (Inherited from Object.)
        /// </summary>
        public override string ToString()
        {
            return this.ToString(null);
        }

        /// <summary>
        /// ToString method with given separator of tokens.
        /// </summary>
        public string ToString(string separator)
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.Append("Positive Samples:\r\n");
            foreach (GenericString<T> positiveSample in this.positiveSamples)
            {
                stringBuilder.Append("  ");
                stringBuilder.Append(positiveSample.ToString(separator));
                stringBuilder.Append("\r\n");
            }
            stringBuilder.Append("Negative Samples:\r\n");
            foreach (GenericString<T> negativeSample in this.negativeSamples)
            {
                stringBuilder.Append("  ");
                stringBuilder.Append(negativeSample.ToString(separator));
                stringBuilder.Append("\r\n");
            }
            return stringBuilder.ToString().TrimEnd();
        }
        #endregion

        #region Private Methods
        /// <summary>
        /// Checks the consistency of the private data members.
        /// </summary>
        private void CheckConsistency()
        {
            foreach (GenericString<T> positiveSample in this.positiveSamples)
            {
                this.alphabet.CheckWord(positiveSample);
                if (this.options == PositiveNegativeSamplesOptions.NegativeSamplesEnabled)
                    throw new InvalidProgramException("Only negative samples are allowed.");
                if (!this.cache.ContainsKey(positiveSample) ||
                     this.cache[positiveSample] == false)
                    throw new InvalidProgramException("Inconsistent positive negatives samples. " +
                        "Error is connected with the positive sample: " + positiveSample.ToString());
            }
            foreach (GenericString<T> negativeSample in this.negativeSamples)
            {
                this.alphabet.CheckWord(negativeSample);
                if (this.options == PositiveNegativeSamplesOptions.PositiveSamplesEnabled)
                    throw new InvalidProgramException("Only positive samples are allowed.");
                if (!this.cache.ContainsKey(negativeSample) ||
                     this.cache[negativeSample] == true)
                    throw new InvalidProgramException("Inconsistent positive negatives samples. " +
                        "Error is connected with the negative sample: " + negativeSample.ToString());
            }
        }
        #endregion

        #region Private Data Members
        private Alphabet<T> alphabet;
        private PositiveNegativeSamplesOptions options;
        private List<GenericString<T>> positiveSamples;
        private List<GenericString<T>> negativeSamples;
        private SortedDictionary<GenericString<T>, bool> cache;
        private string comment;
        #endregion
    }

    /// <summary>
    /// Class that encapsulates Dana Angluin's L* algorithm.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class LStar<T> :
        ILStarData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of the LStar&lt;T&gt; class with given alphabet.
        /// </summary>
        /// <param name="alphabet">Alphabet of the LStar object.</param>
        public LStar(Alphabet<T> alphabet)
        {
            this.alphabet = alphabet;
            this.prefixes = new List<GenericString<T>>(this.alphabet.Count + 1);
            this.postfixes = new List<GenericString<T>>();
            this.knowledgeBase = new KnowledgeBase<T>(this.alphabet);
            this.knowledgeBase.KnowledgeBaseChanged += this.LStar_OnKnowledgeBaseChanged;
            this.unknownWords = new List<GenericString<T>>(this.alphabet.Count + 1);
            this.rowCache = new SortedDictionary<GenericString<T>, GenericString<bool>>();
            this.Initialize();
            this.CheckConsistency();
        }

        /// <summary>
        /// Initializes a new instance of the LStar&lt;T&gt; class
        /// based on the LStar data provider which implements
        /// ILStarData&lt;T&gt; interface.
        /// </summary>
        /// <param name="lstarData">LStar data provider.</param>
        public LStar(ILStarData<T> lstarData)
        {
            this.alphabet = new Alphabet<T>(lstarData);
            this.prefixes = new List<GenericString<T>>(this.alphabet.Count + 1);
            this.postfixes = new List<GenericString<T>>();
            this.unknownWords = new List<GenericString<T>>(this.alphabet.Count + 1);
            this.rowCache = new SortedDictionary<GenericString<T>, GenericString<bool>>();
            this.Initialize();
            this.KnowledgeBase = new KnowledgeBase<T>(lstarData);
            if (!this.CanConjecture)
                throw new InvalidProgramException("Invalid LStar data provider: Conjecture cannot be constructed.");
            this.CheckConsistency();
        }
        #endregion

        #region ILStarData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.alphabet.GetAlphabetData(); }
        /// Returns data describing the knowledge base.
        public WordAcceptData<T>[] GetKnowledgeBaseData() { return this.knowledgeBase.GetKnowledgeBaseData(); }
        /// Returns a comment (a short description) of the knowledge base.
        public string GetKnowledgeBaseCommentData() { return this.knowledgeBase.GetKnowledgeBaseCommentData(); }
        #endregion

        #region Properties
        /// <summary>
        /// Gets the alphabet of the LStar class.
        /// </summary>
        public Alphabet<T> Alphabet
        {
            get
            {
                return this.alphabet;
            }
        }

        /// <summary>
        /// Gets the read-only collection of prefixes.
        /// </summary>
        public ReadOnlyCollection<GenericString<T>> Prefixes
        {
            get
            {
                return new ReadOnlyCollection<GenericString<T>>(this.prefixes);
            }
        }

        /// <summary>
        /// Gets the read-only collection of postfixes.
        /// </summary>
        public ReadOnlyCollection<GenericString<T>> Postfixes
        {
            get
            {
                return new ReadOnlyCollection<GenericString<T>>(this.postfixes);
            }
        }

        /// <summary>
        /// Gets or sets the knowledge base.
        /// </summary>
        public KnowledgeBase<T> KnowledgeBase
        {
            get
            {
                return this.knowledgeBase;
            }
            set
            {
                this.alphabet.CheckAlphabet(value.Alphabet);
                this.knowledgeBase = value;
                this.knowledgeBase.KnowledgeBaseChanged += this.LStar_OnKnowledgeBaseChanged;
                this.Rebuild();
            }
        }

        /// <summary>
        /// Gets the read-only collection of unknown words that have to be set.
        /// </summary>
        public ReadOnlyCollection<GenericString<T>> UnknownWords
        {
            get
            {
                return new ReadOnlyCollection<GenericString<T>>(this.unknownWords);
            }
        }

        /// <summary>
        /// Returns true if and only if you can request a conjecture (dfa).
        /// </summary>
        public bool CanConjecture
        {
            get
            {
                return this.canConjecture;
            }
        }

        /// <summary>
        /// Gets an estimated dfa (only if you can request a conjecture).
        /// </summary>
        public DFA<T> Conjecture
        {
            get
            {
                this.dfaCache.Comment = this.knowledgeBase.Comment;
                return this.dfaCache;
            }
        }
        #endregion

        #region Event Handlers
        /// <summary>
        /// Event handler for KnowledgeBaseChanged event.
        /// </summary>
        private void LStar_OnKnowledgeBaseChanged(KnowledgeBaseChangeType knowledgeBaseChangeType)
        {
            switch (knowledgeBaseChangeType)
            {
                case KnowledgeBaseChangeType.KnowledgeAdded:
                    this.Update();
                    break;
                case KnowledgeBaseChangeType.KnowledgeModified:
                case KnowledgeBaseChangeType.KnowledgeBaseCleared:
                case KnowledgeBaseChangeType.KnowledgeBaseSorted:
                default:
                    this.Rebuild();
                    break;
            }
        }
        #endregion

        #region Methods
        /// <summary>
        /// Resets the LStar&lt;T&gt; and rebuilds the conjecture.
        /// </summary>
        /// <returns>True if you can request a conjecture.</returns>
        public bool Rebuild()
        {
            this.Initialize();
            return this.Update();
        }
        #endregion

        #region Private Methods
        /// <summary>
        /// Resets the LStar&lt;T&gt; except the knowledge base.
        /// </summary>
        private void Initialize()
        {
            this.prefixes.Clear();
            this.postfixes.Clear();
            this.prefixes.Add(GenericString<T>.EmptyWord);
            this.postfixes.Add(GenericString<T>.EmptyWord);
            this.rowCache.Clear();
            this.unknownWords.Clear();
            this.unknownWords.Add(GenericString<T>.EmptyWord);
            foreach (T token in this.alphabet)
                this.unknownWords.Add(GenericString<T>.Wrap(token));
            this.dfaCache = null;
            this.canConjecture = false;
        }

        /// <summary>
        /// Returns a boolean row corresponding to the input prefix.
        /// Only if there are no unknown words.
        /// </summary>
        private GenericString<bool> GetRowVector(GenericString<T> prefix)
        {
            if (this.rowCache.ContainsKey(prefix))
                return this.rowCache[prefix];
            int index = 0;
            bool[] buffer = new bool[this.postfixes.Count];
            foreach (GenericString<T> postfix in this.postfixes)
                buffer[index++] = this.knowledgeBase[prefix + postfix];
            GenericString<bool> row = GenericString<bool>.Wrap(buffer);
            this.rowCache.Add(prefix, row);
            return row;
        }

        /// <summary>
        /// Adds a new prefix.
        /// </summary>
        private void NewPrefix(GenericString<T> prefix)
        {
            this.prefixes.Add(prefix);
            foreach (GenericString<T> postfix in this.postfixes)
            {
                GenericString<T> compound = prefix + postfix;
                if (!this.knowledgeBase.Contains(compound))
                    this.unknownWords.Add(compound);
                foreach (T token in this.alphabet)
                {
                    compound = prefix + GenericString<T>.Wrap(token) + postfix;
                    if (!this.knowledgeBase.Contains(compound))
                        this.unknownWords.Add(compound);
                }
            }
            this.unknownWords.Sort();
            this.unknownWords = Unique<GenericString<T>>.UniqueList(this.unknownWords);
        }

        /// <summary>
        /// Adds a new postfix.
        /// </summary>
        private void NewPostfix(GenericString<T> postfix)
        {
            this.rowCache.Clear();
            this.postfixes.Add(postfix);
            foreach (GenericString<T> prefix in this.prefixes)
            {
                GenericString<T> compound = prefix + postfix;
                if (!this.knowledgeBase.Contains(compound))
                    this.unknownWords.Add(compound);
                foreach (T token in this.alphabet)
                {
                    compound = prefix + GenericString<T>.Wrap(token) + postfix;
                    if (!this.knowledgeBase.Contains(compound))
                        this.unknownWords.Add(compound);
                }
            }
            this.unknownWords.Sort();
            this.unknownWords = Unique<GenericString<T>>.UniqueList(this.unknownWords);
        }

        /// <summary>
        /// Adds a counter example and all its substrings to the prefixes.
        /// </summary>
        private void CounterExample(GenericString<T> counterExample)
        {
            List<GenericString<T>> sortedPrefixes =
                new List<GenericString<T>>(this.prefixes.Count);
            sortedPrefixes.AddRange(this.prefixes);
            sortedPrefixes.Sort();
            for (int count = 1; count <= counterExample.Length; ++count)
            {
                GenericString<T> subString = counterExample.SubString(count);
                if (sortedPrefixes.BinarySearch(subString) < 0)
                    this.NewPrefix(subString);
            }
        }

        /// <summary>
        /// Removes all known words from the set of unknown words.
        /// </summary>
        /// <returns>True if and only if the set of unknown words is empty.</returns>
        private bool RefreshUnknownWords()
        {
            if (this.unknownWords.Count == 0)
                return true;
            List<GenericString<T>> unknownRefresh =
                new List<GenericString<T>>(this.unknownWords.Count);
            foreach (GenericString<T> unknownWord in this.unknownWords)
                if (!this.knowledgeBase.Contains(unknownWord))
                    unknownRefresh.Add(unknownWord);
            this.unknownWords = unknownRefresh;
            return this.unknownWords.Count == 0;
        }

        /// <summary>
        /// Generates a conjecture.
        /// Only if there are no unknown words and the table is consistent and closed.
        /// </summary>
        private void UpdateConjecture()
        {
            SortedDictionary<GenericString<bool>, int> stateMap =
                new SortedDictionary<GenericString<bool>, int>();
            foreach (GenericString<T> prefix in this.prefixes)
                stateMap[this.GetRowVector(prefix)] = -1;
            int stateCount = 0;
            DFA<T> dfa = new DFA<T>(this.alphabet, stateMap.Count);
            List<int> acceptStates = new List<int>(stateMap.Count);
            foreach (GenericString<T> prefix in this.prefixes)
            {
                int fromState = 0;
                int toState = 0;
                GenericString<bool> fromRow =
                    this.GetRowVector(prefix);
                fromState = stateMap[fromRow];
                if (fromState == -1)
                    fromState = stateMap[fromRow] = stateCount++;
                if (fromRow[0])
                    acceptStates.Add(fromState);
                foreach (T token in this.alphabet)
                {
                    GenericString<bool> toRow =
                        this.GetRowVector(prefix + GenericString<T>.Wrap(token));
                    toState = stateMap[toRow];
                    if (toState == -1)
                        toState = stateMap[toRow] = stateCount++;
                    dfa.SetArc(fromState, toState, token);
                }
            }
            dfa.SetAcceptStates(acceptStates);
            this.dfaCache = dfa.Normalize();
        }

        /// <summary>
        /// Makes one logical step in Dana Angluin's L* algorithm.
        /// Only if there are no unknown words.
        /// </summary>
        /// <returns>True if you can request a conjecture.</returns>
        private bool UpdateStep()
        {
            this.dfaCache = null;
            this.canConjecture = false;
            int prefixesCount = this.prefixes.Count;
            int postfixesCount = this.postfixes.Count;
            int prefixIndex = 0;
            int prefixIndex1 = 0;
            int prefixIndex2 = 0;
            int postfixIndex = 0;
            List<GenericString<bool>> prefixRows =
                new List<GenericString<bool>>(prefixesCount);
            foreach (GenericString<T> prefix in this.prefixes)
                prefixRows.Add(this.GetRowVector(prefix));
            // Check consistency
            for (prefixIndex1 = 0; prefixIndex1 < prefixesCount; ++prefixIndex1)
                for (prefixIndex2 = prefixIndex1 + 1; prefixIndex2 < prefixesCount; ++prefixIndex2)
                    if (prefixRows[prefixIndex1].Equals(prefixRows[prefixIndex2]))
                    {
                        GenericString<T> prefix1 = this.prefixes[prefixIndex1];
                        GenericString<T> prefix2 = this.prefixes[prefixIndex2];
                        foreach (T token in this.alphabet)
                            for (postfixIndex = 0; postfixIndex < postfixesCount; ++postfixIndex)
                            {
                                GenericString<T> postfix =
                                    GenericString<T>.Wrap(token) + this.postfixes[postfixIndex];
                                if (this.knowledgeBase[prefix1 + postfix] != this.knowledgeBase[prefix2 + postfix])
                                {
                                    this.NewPostfix(postfix);
                                    return false;
                                }
                            }
                    }
            // Check closure
            prefixRows.Sort();
            for (prefixIndex = 0; prefixIndex < prefixesCount; ++prefixIndex)
                foreach (T token in this.alphabet)
                {
                    GenericString<T> prefix =
                        this.prefixes[prefixIndex] + GenericString<T>.Wrap(token);
                    if (prefixRows.BinarySearch(this.GetRowVector(prefix)) < 0)
                    {
                        this.NewPrefix(prefix);
                        return false;
                    }
                }
            this.UpdateConjecture();
            foreach (WordAccept<T> wordAccept in this.knowledgeBase.Knowledge)
                if (dfaCache.Decide(wordAccept.word) != wordAccept.accept)
                {
                    this.CounterExample(wordAccept.word);
                    return false;
                }
            this.dfaCache.Comment = this.knowledgeBase.Comment;
            this.canConjecture = true;
            return true;
        }

        /// <summary>
        /// Makes as many logical steps in Dana Angluin's L* algorithm as possible.
        /// </summary>
        /// <returns>True if you can request a conjecture.</returns>
        private bool Update()
        {
            if (!this.RefreshUnknownWords())
                return false;
            while (this.unknownWords.Count == 0)
            {
                if (this.UpdateStep())
                    return true;
            }
            return false;
        }

        /// <summary>
        /// Checks the consistency of the private data members.
        /// </summary>
        private void CheckConsistency()
        {
            this.alphabet.CheckAlphabet(this.knowledgeBase.Alphabet);
        }
        #endregion

        #region Private Data Members
        private Alphabet<T> alphabet;
        private List<GenericString<T>> prefixes;
        private List<GenericString<T>> postfixes;
        private KnowledgeBase<T> knowledgeBase;
        private List<GenericString<T>> unknownWords;
        private SortedDictionary<GenericString<T>, GenericString<bool>> rowCache;
        private DFA<T> dfaCache;
        private bool canConjecture;
        #endregion
    }

    /// <summary>
    /// Class that encapsulates Union By Rank algorithm.
    /// </summary>
    public sealed class DisjointSet :
        ICloneable
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of the DisjointSet class with given count of vertices.
        /// </summary>
        /// <param name="count">The number of vertices.</param>
        public DisjointSet(int count)
        {
            if (count <= 0)
                throw new ArgumentException("The DisjointSet must contain at least one vertex.");
            this.count = count;
            this.tree = new int[this.count];
            this.rank = new int[this.count];
            this.Clear();
        }
        #endregion

        #region ICloneable interface
        /// <summary>
        /// Returns a new instance of the DisjointSet class that is a deep copy of this instance.
        /// </summary>
        public object Clone()
        {
            DisjointSet disjointSet = new DisjointSet(this.count);
            for (int index = 0; index < this.count; ++index)
            {
                disjointSet.tree[index] = this.tree[index];
                disjointSet.rank[index] = this.rank[index];
            }
            return disjointSet;
        }
        #endregion

        #region Indexer
        /// <summary>
        /// Gets an element at the specified index.
        /// </summary>
        public int this[int index]
        {
            get
            {
                return this.Find(index);
            }
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets the number of vertices.
        /// </summary>
        public int Count
        {
            get
            {
                return this.count;
            }
        }
        #endregion

        #region Methods
        /// <summary>
        /// Initializes the Union By Rank algorithm.
        /// </summary>
        public void Clear()
        {
            for (int index = 0; index < this.count; ++index)
            {
                this.tree[index] = index;
                this.rank[index] = 0;
            }
        }

        /// <summary>
        /// Returns the representant of x.
        /// </summary>
        public int Find(int x)
        {
            int y = x;
            int z = 0;
            while (y != this.tree[y])
                y = this.tree[y];
            while (x != this.tree[x])
            {
                z = this.tree[x];
                this.tree[x] = y;
                x = z;
            }
            return x;
        }

        /// <summary>
        /// Unions one set that contains x with the other set that contains y.
        /// </summary>
        public void Union(int x, int y)
        {
            this.Link(this.Find(x), this.Find(y));
        }

        /// <summary>
        /// Copies the elements of the internal representation of the DisjointSet to a new array. 
        /// </summary>
        public int[] ToArray()
        {
            return this.tree.Clone() as int[];
        }

        /// <summary>
        /// Returns a String that represents the current Object. (Inherited from Object.)
        /// </summary>
        public override string ToString()
        {
            List<string> stringList = new List<string>(this.count);
            foreach (int x in this.tree)
                stringList.Add(x.ToString());
            return String.Join(" ", stringList.ToArray());
        }
        #endregion

        #region Private Methods
        /// <summary>
        /// Unions two sets with representants x and y.
        /// </summary>
        private void Link(int x, int y)
        {
            if (x == y) return;
            if (this.rank[x] > this.rank[y])
                this.tree[y] = x;
            else
            {
                this.tree[x] = y;
                if (this.rank[x] == this.rank[y])
                    ++this.rank[y];
            }
        }
        #endregion

        #region Private Data Members
        private int count;
        private int[] tree;
        private int[] rank;
        #endregion
    }

    /// <summary>
    /// Class that encapsulates the RPNI algorithm.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class RPNI<T> :
        IRPNIData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of the RPNI&lt;T&gt; class with given alphabet.
        /// </summary>
        /// <param name="alphabet">Alphabet of the RPNI&lt;T&gt;.</param>
        public RPNI(Alphabet<T> alphabet)
        {
            this.alphabet = alphabet;
            this.positiveNegativeSamples = 
                new PositiveNegativeSamples<T>(this.alphabet, 
                        PositiveNegativeSamplesOptions.AllSamplesEnabled);
            this.positivePrefixes = null;
            this.alphabetCount = alphabet.Count;
            this.stateCount = 0;
            this.prefixTree = null;
            this.CheckConsistency();
        }

        /// <summary>
        /// Initializes a new instance of the RPNI&lt;T&gt; class
        /// based on the RPNI data provider which implements
        /// IRPNIData&lt;T&gt; interface.
        /// </summary>
        /// <param name="rpniData">RPNI data provider.</param>
        public RPNI(IRPNIData<T> rpniData)
        {
            this.alphabet = new Alphabet<T>(rpniData);
            this.positiveNegativeSamples = new PositiveNegativeSamples<T>(rpniData);
            this.positivePrefixes = null;
            this.alphabetCount = alphabet.Count;
            this.stateCount = 0;
            this.prefixTree = null;
            this.CheckConsistency();
        }
        #endregion

        #region IRPNIData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.alphabet.GetAlphabetData(); }
        /// Returns special options for positive and negative samples.
        public PositiveNegativeSamplesOptions GetPNSamplesOptions() { return this.positiveNegativeSamples.GetPNSamplesOptions(); }
        /// Returns data describing the positive samples.
        public SampleData<T>[] GetPositiveSamplesData() { return this.positiveNegativeSamples.GetPositiveSamplesData(); }
        /// Returns data describing the negative samples.
        public SampleData<T>[] GetNegativeSamplesData() { return this.positiveNegativeSamples.GetNegativeSamplesData(); }
        /// Returns a comment (a short description) of the positive and negative samples.
        public string GetPNSamplesCommentData() { return this.positiveNegativeSamples.GetPNSamplesCommentData(); }
        #endregion

        #region Properties
        /// <summary>
        /// Gets the alphabet of the RPNI&lt;T&gt; class.
        /// </summary>
        public Alphabet<T> Alphabet
        {
            get
            {
                return this.alphabet;
            }
        }

        /// <summary>
        /// Gets or sets the positive and negative samples.
        /// </summary>
        public PositiveNegativeSamples<T> Samples
        {
            get
            {
                return this.positiveNegativeSamples;
            }
            set
            {
                this.alphabet.CheckAlphabet(value.Alphabet);
                this.positiveNegativeSamples = value;
            }
        }

        /// <summary>
        /// Gets a new DFA&lt;T&gt; that is consistent with given
        /// positive and negative samples in the knowledge base.
        /// </summary>
        public DFA<T> Conjecture
        {
            get
            {
                const int minimalCapacity = 32;
                List<GenericString<T>> positiveSamples = new List<GenericString<T>>(minimalCapacity);
                List<GenericString<T>> negativeSamples = new List<GenericString<T>>(minimalCapacity);
                positiveSamples.AddRange(this.positiveNegativeSamples.PositiveSamples);
                negativeSamples.AddRange(this.positiveNegativeSamples.NegativeSamples);
                if (negativeSamples.Count == 0)
                    return DFA<T>.CreateAcceptingDFA(this.alphabet);
                if (positiveSamples.Count == 0)
                    return DFA<T>.CreateRejectingDFA(this.alphabet);
                this.SetPositivePrefixes(positiveSamples);
                this.SetPrefixTree();
                DisjointSet disjointSet = new DisjointSet(this.stateCount);
                DFA<T> dfa = this.GetDFA(disjointSet);
                for (int stateIndex1 = 1; stateIndex1 < this.stateCount; ++stateIndex1)
                    for (int stateIndex2 = 0; stateIndex2 < stateIndex1; ++stateIndex2)
                    {
                        DisjointSet disjointSetNew =
                            disjointSet.Clone() as DisjointSet;
                        disjointSetNew.Union(stateIndex1, stateIndex2);
                        this.DeterministicMerge(disjointSetNew);
                        DFA<T> dfaNew = this.GetDFA(disjointSetNew);
                        bool consistent = true;
                        foreach (GenericString<T> negativeSample in negativeSamples)
                            if (dfaNew.Decide(negativeSample))
                            {
                                consistent = false;
                                break;
                            }
                        if (consistent)
                        {
                            disjointSet = disjointSetNew;
                            dfa = dfaNew;
                        }
                    }
                dfa.Comment = this.positiveNegativeSamples.Comment;
                return dfa;
            }
        }

        /// <summary>
        /// Gets a new DFA&lt;T&gt; that represents a prefix tree 
        /// consistent with given positive (and negative) samples 
        /// in the knowledge base.
        /// </summary>
        public DFA<T> PrefixTree
        {
            get
            {
                const int minimalCapacity = 32;
                List<GenericString<T>> positiveSamples = new List<GenericString<T>>(minimalCapacity);
                List<GenericString<T>> negativeSamples = new List<GenericString<T>>(minimalCapacity);
                positiveSamples.AddRange(this.positiveNegativeSamples.PositiveSamples);
                negativeSamples.AddRange(this.positiveNegativeSamples.NegativeSamples);
                if (negativeSamples.Count == 0)
                    return DFA<T>.CreateAcceptingDFA(this.alphabet);
                if (positiveSamples.Count == 0)
                    return DFA<T>.CreateRejectingDFA(this.alphabet);
                this.SetPositivePrefixes(positiveSamples);
                this.SetPrefixTree();
                DisjointSet disjointSet = new DisjointSet(this.stateCount);
                return this.GetDFA(disjointSet);
            }
        }
        #endregion

        #region Private Methods
        /// <summary>
        /// Finds all prefixes of positive samples.
        /// </summary>
        private void SetPositivePrefixes(IEnumerable<GenericString<T>> positiveSamples)
        {
            this.positivePrefixes = new SortedDictionary<GenericString<T>, bool>();
            foreach (GenericString<T> positiveSample in positiveSamples)
            {
                for (int substrLength = 0; substrLength < positiveSample.Length; ++substrLength)
                {
                    GenericString<T> postiveSampleSubstring = positiveSample.SubString(substrLength);
                    if (!positivePrefixes.ContainsKey(postiveSampleSubstring))
                        this.positivePrefixes[postiveSampleSubstring] = false;
                }
                this.positivePrefixes[positiveSample] = true;
            }
        }

        /// <summary>
        /// Builds a prefix tree.
        /// </summary>
        private void SetPrefixTree()
        {
            this.stateCount = this.positivePrefixes.Count;
            this.prefixTree = new int[this.stateCount, this.alphabetCount];
            this.acceptState = new bool[this.stateCount];
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                for (int tokenIndex = 0; tokenIndex < this.alphabetCount; ++tokenIndex)
                    this.prefixTree[stateIndex, tokenIndex] = this.stateCount;
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                this.acceptState[stateIndex] = false;
            int topState = 0;
            foreach (KeyValuePair<GenericString<T>, bool> positivePrefixPair in this.positivePrefixes)
            {
                int actualState = 0;
                foreach (T token in positivePrefixPair.Key)
                {
                    if (this.prefixTree[actualState, this.alphabet.IndexOf(token)] == this.stateCount)
                        this.prefixTree[actualState, this.alphabet.IndexOf(token)] = ++topState;
                    actualState = this.prefixTree[actualState, this.alphabet.IndexOf(token)];
                }
                this.acceptState[actualState] = positivePrefixPair.Value;
            }
        }

        /// <summary>
        /// ConflictingPair is a pair of states that should be merged.
        /// </summary>
        private class ConflictPair
        {
            /// <summary>
            /// Initializes a new instance of ConflictPair class.
            /// </summary>
            public ConflictPair(int x, int y)
            {
                this.x = x;
                this.y = y;
            }
            /// First state.
            public int x;
            /// Second state.
            public int y;
        }

        /// <summary>
        /// Returns a conflicting pair for given disjoint set or
        /// null if there are no conflicts.
        /// </summary>
        private ConflictPair GetConflictPair(DisjointSet disjointSet)
        {
            int[,] fsaTable = new int[this.stateCount, this.alphabetCount];
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
                for (int tokenIndex = 0; tokenIndex < this.alphabetCount; ++tokenIndex)
                    fsaTable[stateIndex, tokenIndex] = this.stateCount;
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
            {
                int stateIndexRoot = disjointSet.Find(stateIndex);
                for (int tokenIndex = 0; tokenIndex < this.alphabetCount; ++tokenIndex)
                {
                    int toState = this.prefixTree[stateIndex, tokenIndex];
                    if (toState != this.stateCount)
                    {
                        int toStateRoot = disjointSet.Find(toState);
                        int exStateRoot = fsaTable[stateIndexRoot, tokenIndex];
                        if (exStateRoot == stateCount)
                            fsaTable[stateIndexRoot, tokenIndex] = toStateRoot;
                        else if (toStateRoot != exStateRoot)
                            return new ConflictPair(exStateRoot, toStateRoot);
                    }
                }
            }
            return null;
        }

        /// <summary>
        /// Merges the states until the resulting automaton is deterministic.
        /// </summary>
        private void DeterministicMerge(DisjointSet disjointSet)
        {
            bool deterministic = false;
            while (!deterministic)
            {
                ConflictPair conflictPair = this.GetConflictPair(disjointSet);
                if (conflictPair == null) deterministic = true;
                else disjointSet.Union(conflictPair.x, conflictPair.y);
            }
        }

        /// <summary>
        /// Returns DFA based on a disjoint set and a prefix tree.
        /// </summary>
        private DFA<T> GetDFA(DisjointSet disjointSet)
        {
            DFA<T> dfa = new DFA<T>(this.alphabet, this.stateCount + 1);
            int[] swap = new int[this.stateCount + 1];
            for (int swapIndex = 0; swapIndex <= this.stateCount; ++swapIndex)
                swap[swapIndex] = swapIndex;
            int swapStart = disjointSet.Find(0);
            swap[swapStart] = 0;
            swap[0] = swapStart;
            for (int stateIndex = 0; stateIndex <= this.stateCount; ++stateIndex)
                for (int tokenIndex = 0; tokenIndex < this.alphabetCount; ++tokenIndex)
                    dfa.SetArc(stateIndex, stateCount, this.alphabet[tokenIndex]);
            for (int stateIndex = 0; stateIndex < this.stateCount; ++stateIndex)
            {
                int stateIndexRoot = swap[disjointSet.Find(stateIndex)];
                if (this.acceptState[stateIndex])
                    dfa.SetAcceptState(stateIndexRoot);
                for (int tokenIndex = 0; tokenIndex < this.alphabetCount; ++tokenIndex)
                {
                    int toState = this.prefixTree[stateIndex, tokenIndex];
                    if (toState != stateCount)
                    {
                        int toStateRoot = swap[disjointSet.Find(toState)];
                        int exStateRoot = dfa.GetArc(stateIndexRoot, this.alphabet[tokenIndex]);
                        if (exStateRoot == this.stateCount)
                            dfa.SetArc(stateIndexRoot, toStateRoot, this.alphabet[tokenIndex]);
                        else if (toStateRoot != exStateRoot)
                            throw new Exception("Cannot return nondeterministic finite-state automaton.");
                    }
                }
            }
            return dfa.Normalize();
        }

        /// <summary>
        /// Checks the consistency of the private data members.
        /// </summary>
        private void CheckConsistency()
        {
            this.alphabet.CheckAlphabet(this.positiveNegativeSamples.Alphabet);
        }
        #endregion

        #region Private Data Members
        private Alphabet<T> alphabet;
        private PositiveNegativeSamples<T> positiveNegativeSamples;
        private SortedDictionary<GenericString<T>, bool> positivePrefixes;
        private int alphabetCount;
        private int stateCount;
        private int[,] prefixTree;
        private bool[] acceptState;
        #endregion
    }

    /// <summary>
    /// Class designed for strictly locally testable languages.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class SLT<T> :
        ISLTData<T>,
        ICloneable
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of the SLT&lt;T&gt; class with given alphabet.
        /// </summary>
        /// <param name="alphabet">Alphabet of the SLT&lt;T&gt;.</param>
        /// <param name="k">k for k-SLT languages.</param>
        public SLT(Alphabet<T> alphabet, int k)
        {
            this.alphabet = alphabet;
            this.positiveNegativeSamples =
                new PositiveNegativeSamples<T>(alphabet, 
                        PositiveNegativeSamplesOptions.PositiveSamplesEnabled);
            this.k = k;
            this.InitializeEventHandlers();
            this.CreateInternalRepresentation();
            this.CheckConsistency();
        }

        /// <summary>
        /// Initializes a new instance of the SLT&lt;T&gt; class
        /// based on the SLT data provider which implements
        /// ISLTData&lt;T&gt; interface.
        /// </summary>
        /// <param name="sltData">SLT data provider.</param>
        public SLT(ISLTData<T> sltData)
        {
            this.alphabet = new Alphabet<T>(sltData);
            this.positiveNegativeSamples = new PositiveNegativeSamples<T>(sltData);
            this.k = sltData.GetSLTK();
            this.InitializeEventHandlers();
            this.CreateInternalRepresentation();
            this.CheckConsistency();
        }
        #endregion

        #region ISLTData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.alphabet.GetAlphabetData(); }
        /// Returns special options for positive and negative samples.
        public PositiveNegativeSamplesOptions GetPNSamplesOptions() { return this.positiveNegativeSamples.GetPNSamplesOptions(); }
        /// Returns data describing the positive samples.
        public SampleData<T>[] GetPositiveSamplesData() { return this.positiveNegativeSamples.GetPositiveSamplesData(); }
        /// Returns data describing the negative samples.
        public SampleData<T>[] GetNegativeSamplesData() { return this.positiveNegativeSamples.GetNegativeSamplesData(); }
        /// Returns a comment (a short description) of the positive and negative samples.
        public string GetPNSamplesCommentData() { return this.positiveNegativeSamples.GetPNSamplesCommentData(); }
        /// Returns k for k-SLT languages.
        public int GetSLTK() { return this.k; }
        #endregion

        #region ICloneable Interface
        /// <summary>
        /// Returns a new instance of the SLT&lt;T&gt; class that is a deep copy of this instance.
        /// </summary>
        public object Clone()
        {
            SLT<T> slt = new SLT<T>(this.alphabet, this.k);
            slt.Samples = this.positiveNegativeSamples.Clone() 
                as PositiveNegativeSamples<T>;
            return slt;
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets the alphabet of the SLT&lt;T&gt; class.
        /// </summary>
        public Alphabet<T> Alphabet
        {
            get
            {
                return this.alphabet;
            }
        }
        
        /// <summary>
        /// Gets or sets the positive and negative samples.
        /// </summary>
        public PositiveNegativeSamples<T> Samples
        {
            get
            {
                return this.positiveNegativeSamples;
            }
            set
            {
                this.alphabet.CheckAlphabet(value.Alphabet);
                if (this.positiveNegativeSamples.Options !=
                        PositiveNegativeSamplesOptions.PositiveSamplesEnabled)
                    throw new InvalidProgramException("Only positive samples should be allowed.");
                this.positiveNegativeSamples = value;
                this.InitializeEventHandlers();
                this.Rebuild();
            }
        }

        /// <summary>
        /// Gets k for k-SLT languages.
        /// </summary>
        public int K
        {
            get
            {
                return this.k;
            }
            set
            {
                this.k = value;
                if (this.k <= 0)
                    throw new InvalidProgramException("k shold be positive.");
                this.Rebuild();
            }
        }
        #endregion

        #region Event Handlers
        /// <summary>
        /// Event handler for PositiveNegativeSamplesChanged event.
        /// </summary>
        private void SLT_OnPositiveNegativeSamplesChanged(
            PositiveNegativeSamplesChangeType positiveNegativeSamplesChangeType)
        {
            switch (positiveNegativeSamplesChangeType)
            {
                case PositiveNegativeSamplesChangeType.PositiveSampleAdded: break;
                case PositiveNegativeSamplesChangeType.NegativeSampleAdded: break;
                case PositiveNegativeSamplesChangeType.PositiveNegativeSamplesCleared:
                    this.Clear();
                    break;
                default:
                    break;
            }
        }

        /// <summary>
        /// Event handler for PositiveSampleAdded event.
        /// </summary>
        private void SLT_OnPositiveSampleAdded(GenericString<T> positiveSample)
        {
            this.AddPositiveSample(positiveSample);
        }

        /// <summary>
        /// Event handler for NegativeSampleAdded event.
        /// </summary>
        private void SLT_OnNegativeSampleAdded(GenericString<T> negativeSample)
        {
            throw new InvalidProgramException("Only positive samples are allowed.");
        }
        #endregion

        #region Methods
        /// <summary>
        /// Returns true if and only if the input generic string is accepted by the k-SLT language.
        /// </summary>
        public bool Decide(GenericString<T> word)
        {
            int length = word.Length;
            if (length < this.k)
                return this.smallWords.ContainsKey(word);
            else if (length == this.k)
                return this.prefixes.ContainsKey(word) &&
                    this.postfixes.ContainsKey(word);
            else
            {
                GenericString<T> prefix = word.Prefix(this.k);
                GenericString<T> postfix = word.Postfix(this.k);
                if (!this.prefixes.ContainsKey(prefix)) return false;
                if (!this.postfixes.ContainsKey(postfix)) return false;
                if (length > this.k + 1)
                    for (int index = 0; index < length - this.k - 1; ++index)
                    {
                        GenericString<T> infix = word.SubString(index + 1, this.k);
                        if (!this.infixes.ContainsKey(infix)) return false;
                    }
                return true;
            }
        }

        /// <summary>
        /// Returns a String that represents the current Object. (Inherited from Object.)
        /// </summary>
        public override string ToString()
        {
            return this.ToString(null);
        }

        /// <summary>
        /// ToString method with given separator of tokens.
        /// </summary>
        public string ToString(string separator)
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.Append(this.k.ToString() + "-SLT\r\n");
            stringBuilder.Append("Small words:\r\n");
            foreach (GenericString<T> smallWord in this.smallWords.Keys)
            {
                stringBuilder.Append("  ");
                stringBuilder.Append(smallWord.ToString(separator));
                stringBuilder.Append("\r\n");
            }
            stringBuilder.Append("Prefixes:\r\n");
            foreach (GenericString<T> prefix in this.prefixes.Keys)
            {
                stringBuilder.Append("  ");
                stringBuilder.Append(prefix.ToString(separator));
                stringBuilder.Append("\r\n");
            }
            stringBuilder.Append("Postfixes:\r\n");
            foreach (GenericString<T> postfix in this.postfixes.Keys)
            {
                stringBuilder.Append("  ");
                stringBuilder.Append(postfix.ToString(separator));
                stringBuilder.Append("\r\n");
            }
            stringBuilder.Append("Infixes:\r\n");
            foreach (GenericString<T> infix in this.infixes.Keys)
            {
                stringBuilder.Append("  ");
                stringBuilder.Append(infix.ToString(separator));
                stringBuilder.Append("\r\n");
            }
            return stringBuilder.ToString().TrimEnd();
        }
        #endregion

        #region Private Methods
        /// <summary>
        /// Creates the internal representation of SLT.
        /// </summary>
        private void CreateInternalRepresentation()
        {
            this.smallWords = new SortedDictionary<GenericString<T>, bool>();
            this.prefixes = new SortedDictionary<GenericString<T>, bool>();
            this.postfixes = new SortedDictionary<GenericString<T>, bool>();
            this.infixes = new SortedDictionary<GenericString<T>, bool>();
        }

        /// <summary>
        /// Initializes the event handlers.
        /// </summary>
        private void InitializeEventHandlers()
        {
            this.positiveNegativeSamples.PositiveNegativeSamplesChanged +=
                this.SLT_OnPositiveNegativeSamplesChanged;
            this.positiveNegativeSamples.PositiveSampleAdded +=
                this.SLT_OnPositiveSampleAdded;
            this.positiveNegativeSamples.NegativeSampleAdded +=
                this.SLT_OnNegativeSampleAdded;
        }

        /// <summary>
        /// Clears the internal representation of the SLT.
        /// </summary>
        private void Clear()
        {
            this.smallWords.Clear();
            this.prefixes.Clear();
            this.postfixes.Clear();
            this.infixes.Clear();
        }

        /// <summary>
        /// Rebuilds the internal representation of the SLT.
        /// </summary>
        private void Rebuild()
        {
            this.Clear();
            foreach (GenericString<T> positiveSample in this.positiveNegativeSamples.PositiveSamples)
                this.AddPositiveSample(positiveSample);
        }

        /// <summary>
        /// Adds a positive sample to the internal representation of the SLT.
        /// </summary>
        private void AddPositiveSample(GenericString<T> positiveSample)
        {
            int length = positiveSample.Length;
            if (length < this.k)
                this.smallWords[positiveSample] = true;
            else if (length == this.k)
            {
                this.prefixes[positiveSample] = true;
                this.postfixes[positiveSample] = true;
            }
            else
            {
                GenericString<T> prefix = positiveSample.Prefix(this.k);
                GenericString<T> postfix = positiveSample.Postfix(this.k);
                this.prefixes[prefix] = true;
                this.postfixes[postfix] = true;
                if (length > this.k + 1)
                    for (int index = 0; index < length - this.k - 1; ++index)
                    {
                        GenericString<T> infix = positiveSample.SubString(index + 1, this.k);
                        this.infixes[infix] = true;
                    }
            }
        }

        /// <summary>
        /// Checks the consistency of the private data members.
        /// </summary>
        private void CheckConsistency()
        {
            this.alphabet.CheckAlphabet(this.positiveNegativeSamples.Alphabet);
            if (this.positiveNegativeSamples.Options !=
                    PositiveNegativeSamplesOptions.PositiveSamplesEnabled)
                throw new InvalidProgramException("Only positive samples should be allowed.");
            if (this.k <= 0)
                throw new InvalidProgramException("k shold be positive.");
        }
        #endregion

        #region Private Data Members
        private Alphabet<T> alphabet;
        private PositiveNegativeSamples<T> positiveNegativeSamples;
        private int k;
        private SortedDictionary<GenericString<T>, bool> smallWords;
        private SortedDictionary<GenericString<T>, bool> prefixes;
        private SortedDictionary<GenericString<T>, bool> postfixes;
        private SortedDictionary<GenericString<T>, bool> infixes;
        #endregion
    }

    /// <summary>
    /// Interface that is implemented by every language.
    /// All the classes that implement this interface should be immutable.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface ILanguage<T> :
        ILanguageData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// <summary>
        /// Gets the alphabet of the language.
        /// </summary>
        Alphabet<T> Alphabet { get; }

        /// <summary>
        /// Decides whether the given generic word belongs to the language.
        /// </summary>
        bool Decide(IEnumerable<T> word);
    }

    /// <summary>
    /// Immutable language class that encapsulates a DFA.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public class LanguageDFA<T> :
        ILanguage<T>,
        ILanguageDFAData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of LanguageDFA&lt;T&gt; class with given DFA.
        /// </summary>
        /// <param name="dfa">DFA of the LanguageDFA&lt;T&gt;.</param>
        public LanguageDFA(DFA<T> dfa)
        {
            this.dfa = dfa.Clone() as DFA<T>;
        }

        /// <summary>
        /// Initializes a new instance of the LanguageDFA&lt;T&gt; class
        /// based on the language data provider which implements
        /// ILanguageDFAData&lt;T&gt; interface.
        /// </summary>
        /// <param name="languageDFAData">Language data provider.</param>
        public LanguageDFA(ILanguageDFAData<T> languageDFAData)
        {
            this.dfa = new DFA<T>(languageDFAData);
        }
        #endregion

        #region ILanguage Interface
        /// <summary>
        /// Gets the alphabet of the LanguageDFA&lt;T&gt; class.
        /// </summary>
        public Alphabet<T> Alphabet
        {
            get
            {
                return this.dfa.Alphabet;
            }
        }

        /// <summary>
        /// Decides whether the given generic word belongs to the language.
        /// </summary>
        public bool Decide(IEnumerable<T> word)
        {
            this.dfa.Alphabet.CheckWord(word);
            return this.dfa.Decide(word);
        }
        #endregion

        #region ILanguageDFAData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.dfa.GetAlphabetData(); }
        /// Returns the number of states of the DFA.
        public int GetDFAStateCountData() { return this.dfa.GetDFAStateCountData(); }
        /// Returns the string that consist of the indices of the accepted states
        /// separated by a single space.
        public string GetDFAAcceptStatesData() { return this.dfa.GetDFAAcceptStatesData(); }
        /// Returns the content of the transition table as a string of numbers
        /// separated by a single space.
        public string GetDFATransitionTableData() { return this.dfa.GetDFATransitionTableData(); }
        /// Returns a comment (a short description) of the dfa.
        public string GetDFACommentData() { return this.dfa.GetDFACommentData(); }
        #endregion

        #region Methods
        /// <summary>
        /// Returns a String that represents the current Object. (Inherited from Object.)
        /// </summary>
        public override string ToString()
        {
            if (String.IsNullOrEmpty(this.dfa.Comment))
                return "DFA Language";
            else return this.dfa.Comment;
        }
        #endregion

        #region Private Data Members
        private DFA<T> dfa;
        #endregion
    }

    /// <summary>
    /// Immutable language class that encapsulates a DFA
    /// constructed by LStar (Dana Angluin's L*) algorithm.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class LanguageDFALStar<T> :
        LanguageDFA<T>,
        ILanguageDFALStarData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of LanguageDFALStar&lt;T&gt; class with given LStar object.
        /// </summary>
        /// <param name="lstar">LStar object.</param>
        public LanguageDFALStar(LStar<T> lstar) :
            base(lstar.Conjecture)
        {
            this.knowledgeBase = lstar.KnowledgeBase.Clone() as KnowledgeBase<T>;
        }

        /// <summary>
        /// Initializes a new instance of the LanguageDFALStar&lt;T&gt; class
        /// based on the language data provider which implements
        /// ILanguageDFALStarData&lt;T&gt; interface.
        /// </summary>
        /// <param name="languageDFALStarData">Language data provider.</param>
        public LanguageDFALStar(ILanguageDFALStarData<T> languageDFALStarData) :
            base(languageDFALStarData)
        {
            this.knowledgeBase = new KnowledgeBase<T>(languageDFALStarData);
        }
        #endregion

        #region ILanguageDFALStarData Interface
        /// Returns an array of WordAcceptData&lt;T&gt; structures describing the whole knowledge base.
        public WordAcceptData<T>[] GetKnowledgeBaseData() { return this.knowledgeBase.GetKnowledgeBaseData(); }
        /// Returns a comment (a short description) of the knowledge base.
        public string GetKnowledgeBaseCommentData() { return this.knowledgeBase.GetKnowledgeBaseCommentData(); }
        #endregion

        #region Methods
        /// <summary>
        /// Returns a String that represents the current Object. (Inherited from Object.)
        /// </summary>
        public override string ToString()
        {
            if (String.IsNullOrEmpty(this.knowledgeBase.Comment))
                return "LStar Language";
            else return this.knowledgeBase.Comment;
        }
        #endregion

        #region Private Data Members
        private KnowledgeBase<T> knowledgeBase;
        #endregion
    }

    /// <summary>
    /// Immutable language class that encapsulates a DFA
    /// constructed by RPNI algorithm.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class LanguageDFARPNI<T> :
        LanguageDFA<T>,
        ILanguageDFARPNIData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of LanguageDFARPNI&lt;T&gt; class with given RPNI object.
        /// </summary>
        /// <param name="rpni">RPNI object.</param>
        public LanguageDFARPNI(RPNI<T> rpni) :
            base(rpni.Conjecture)
        {
            this.positiveNegativeSamples = rpni.Samples.Clone() as PositiveNegativeSamples<T>;
        }

        /// <summary>
        /// Initializes a new instance of the LanguageDFARPNI&lt;T&gt; class
        /// based on the language data provider which implements
        /// ILanguageDFARPNIData&lt;T&gt; interface.
        /// </summary>
        /// <param name="languageDFARPNIData">Language data provider.</param>
        public LanguageDFARPNI(ILanguageDFARPNIData<T> languageDFARPNIData) :
            base(languageDFARPNIData)
        {
            this.positiveNegativeSamples = new PositiveNegativeSamples<T>(languageDFARPNIData);
        }
        #endregion

        #region ILanguageDFARPNIData Interface
        /// Returns special options for positive and negative samples.
        public PositiveNegativeSamplesOptions GetPNSamplesOptions() { return this.positiveNegativeSamples.GetPNSamplesOptions(); }
        /// Returns an array of SampleData&lt;T&gt; structures describing the set of positive samples.
        public SampleData<T>[] GetPositiveSamplesData() { return this.positiveNegativeSamples.GetPositiveSamplesData(); }
        /// Returns an array of SampleData&lt;T&gt; structures describing the set of negative samples.
        public SampleData<T>[] GetNegativeSamplesData() { return this.positiveNegativeSamples.GetNegativeSamplesData(); }
        /// Returns a comment (a short description) of the positive and negative samples.
        public string GetPNSamplesCommentData() { return this.positiveNegativeSamples.GetPNSamplesCommentData(); }
        #endregion

        #region Methods
        /// <summary>
        /// Returns a String that represents the current Object. (Inherited from Object.)
        /// </summary>
        public override string ToString()
        {
            if (String.IsNullOrEmpty(this.positiveNegativeSamples.Comment))
                return "RPNI Language";
            else return this.positiveNegativeSamples.Comment;
        }
        #endregion

        #region Private Data Members
        private PositiveNegativeSamples<T> positiveNegativeSamples;
        #endregion
    }

    /// <summary>
    /// Immutable language class that encapsulates a regular expression.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class LanguageRegex<T> :
        ILanguage<T>,
        ILanguageRegexData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of LanguageRegex&lt;T&gt; class with given alphabet
        /// and regex pattern.
        /// </summary>
        /// <param name="alphabet">Alphabet.</param>
        /// <param name="regexPattern">Regex pattern.</param>
        public LanguageRegex(Alphabet<T> alphabet, string regexPattern)
        {
            this.alphabet = alphabet;
            this.regexPattern = regexPattern;
            this.regex = new Regex(regexPattern);
            this.CheckConsistency();
        }

        /// <summary>
        /// Initializes a new instance of the LanguageRegex&lt;T&gt; class
        /// based on the language data provider which implements
        /// ILanguageRegexData&lt;T&gt; interface.
        /// </summary>
        /// <param name="languageRegexData">Language data provider.</param>
        public LanguageRegex(ILanguageRegexData<T> languageRegexData)
        {
            this.alphabet = new Alphabet<T>(languageRegexData);
            this.regexPattern = languageRegexData.GetRegexData();
            this.regex = new Regex(regexPattern);
            this.CheckConsistency();
        }
        #endregion

        #region ILanguage Interface
        /// <summary>
        /// Gets the alphabet of the LanguageRegex&lt;T&gt; class.
        /// </summary>
        public Alphabet<T> Alphabet
        {
            get
            {
                return this.alphabet;
            }
        }

        /// <summary>
        /// Decides whether the given generic word belongs to the language.
        /// </summary>
        public bool Decide(IEnumerable<T> word)
        {
            this.alphabet.CheckWord(word);
            GenericString<T> wordString = new GenericString<T>(word);
            return this.regex.IsMatch(wordString.ToString());
        }
        #endregion

        #region ILanguageRegexData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.alphabet.GetAlphabetData(); }
        /// Returns the string pattern of the regular expression.
        public string GetRegexData() { return this.regexPattern; }
        #endregion

        #region Methods
        /// <summary>
        /// Returns a String that represents the current Object. (Inherited from Object.)
        /// </summary>
        public override string ToString()
        {
            if (String.IsNullOrEmpty(this.regexPattern))
                return "Regex Language";
            else return this.regexPattern;
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets the pattern of the regular expression.
        /// </summary>
        public string RegexPattern
        {
            get
            {
                return this.regexPattern;
            }
        }
        #endregion

        #region Private Methods
        private void CheckConsistency()
        {
            if (typeof(T) != typeof(char))
                throw new InvalidProgramException("Only type char is supported.");
        }
        #endregion

        #region Private Data Members
        private Alphabet<T> alphabet;
        private string regexPattern;
        private Regex regex;
        #endregion
    }

    /// <summary>
    /// Immutable language class that encapsulates a k-SLT language.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class LanguageSLT<T> :
        ILanguage<T>,
        ILanguageSLTData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of LanguageSLT&lt;T&gt; class with given SLT.
        /// </summary>
        /// <param name="slt">SLT of the LanguageSLT&lt;T&gt;.</param>
        public LanguageSLT(SLT<T> slt)
        {
            this.slt = slt.Clone() as SLT<T>;
        }

        /// <summary>
        /// Initializes a new instance of the LanguageSLT&lt;T&gt; class
        /// based on the language data provider which implements
        /// ILanguageSLTData&lt;T&gt; interface.
        /// </summary>
        /// <param name="languageSLTData">Language data provider.</param>
        public LanguageSLT(ILanguageSLTData<T> languageSLTData)
        {
            this.slt = new SLT<T>(languageSLTData);
        }
        #endregion

        #region ILanguage Interface
        /// <summary>
        /// Gets the alphabet of the LanguageSLT&lt;T&gt; class.
        /// </summary>
        public Alphabet<T> Alphabet
        {
            get
            {
                return this.slt.Alphabet;
            }
        }

        /// <summary>
        /// Decides whether the given generic word belongs to the language.
        /// </summary>
        public bool Decide(IEnumerable<T> word)
        {
            this.slt.Alphabet.CheckWord(word);
            return this.slt.Decide(new GenericString<T>(word));
        }
        #endregion

        #region ILanguageSLTData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.slt.Alphabet.GetAlphabetData(); }
        /// Returns special options for positive and negative samples.
        public PositiveNegativeSamplesOptions GetPNSamplesOptions() { return this.slt.Samples.GetPNSamplesOptions(); }
        /// Returns data describing the positive samples.
        public SampleData<T>[] GetPositiveSamplesData() { return this.slt.Samples.GetPositiveSamplesData(); }
        /// Returns data describing the negative samples.
        public SampleData<T>[] GetNegativeSamplesData() { return this.slt.Samples.GetNegativeSamplesData(); }
        /// Returns a comment (a short description) of the positive and negative samples.
        public string GetPNSamplesCommentData() { return this.slt.Samples.GetPNSamplesCommentData(); }
        /// Returns k for k-SLT languages.
        public int GetSLTK() { return this.slt.GetSLTK(); }
        #endregion

        #region Methods
        /// <summary>
        /// Returns a String that represents the current Object. (Inherited from Object.)
        /// </summary>
        public override string ToString()
        {
            if (String.IsNullOrEmpty(this.slt.Samples.Comment))
                return "SLT Language";
            else return this.slt.Samples.Comment;
        }
        #endregion

        #region Private Data Members
        private SLT<T> slt;
        #endregion
    }

    /// <summary>
    /// Class that represents an accepting metainstruction of a restarting automaton.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class AcceptingMetainstruction<T> :
        IAcceptingMetainstructionData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of AcceptingMetainstruction&lt;T&gt; class with given alphabet.
        /// </summary>
        /// <param name="alphabet">Input alphabet.</param>
        public AcceptingMetainstruction(Alphabet<T> alphabet)
        {
            this.alphabet = alphabet;
            this.acceptingLanguage = null;
            this.name = "";
        }

        /// <summary>
        /// Initializes a new instance of the AcceptingMetainstruction&lt;T&gt; class
        /// based on the accepting metainstruction data provider which implements
        /// IAcceptingMetainstructionData&lt;T&gt; interface.
        /// </summary>
        /// <param name="acceptingMetainstructionData">Accepting metainstruction data provider.</param>
        public AcceptingMetainstruction(IAcceptingMetainstructionData<T> acceptingMetainstructionData)
        {
            this.alphabet = new Alphabet<T>(acceptingMetainstructionData);
            this.acceptingLanguage = LanguageFactory<T>.GetLanguageInstance(
                acceptingMetainstructionData.GetAcceptingLanguageData());
            this.name = acceptingMetainstructionData.GetAcceptingName();
            if (this.acceptingLanguage != null)
                this.alphabet.CheckAlphabet(this.acceptingLanguage.Alphabet);
        }
        #endregion

        #region IAcceptingMetainstructionData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.alphabet.GetAlphabetData(); }
        /// Returns an interface that provides all the neccessary data
        /// describing the accepting language of the accepting metainstruction.
        public ILanguageData<T> GetAcceptingLanguageData() { return this.acceptingLanguage; }
        /// Returns the name of the accepting metainstruction.
        public string GetAcceptingName() { return this.name; }
        #endregion

        #region Properties
        /// <summary>
        /// Gets the alphabet of the accepting metainstruction.
        /// </summary>
        public Alphabet<T> Alphabet
        {
            get
            {
                return this.alphabet;
            }
        }

        /// <summary>
        /// Gets or sets the accepting language.
        /// </summary>
        public ILanguage<T> AcceptingLanguage
        {
            get
            {
                return this.acceptingLanguage;
            }
            set
            {
                this.alphabet.CheckAlphabet(value.Alphabet);
                this.acceptingLanguage = value;
            }
        }

        /// <summary>
        /// Gets or sets the name of the accepting metainstruction.
        /// </summary>
        public string Name
        {
            get
            {
                return this.name;
            }
            set
            {
                this.name = value.Trim();
            }
        }
        #endregion

        #region Private Data Members
        private Alphabet<T> alphabet;
        private ILanguage<T> acceptingLanguage;
        private string name;
        #endregion
    }

    /// <summary>
    /// Class that represents a reducing metainstruction of a restarting automaton.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class ReducingMetainstruction<T> :
        IReducingMetainstructionData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of ReducingMetainstruction&lt;T&gt; class with given alphabet.
        /// </summary>
        /// <param name="alphabet">Input alphabet.</param>
        public ReducingMetainstruction(Alphabet<T> alphabet)
        {
            this.alphabet = alphabet;
            this.leftLanguage = null;
            this.rightLanguage = null;
            this.fromWord = null;
            this.toWord = null;
            this.name = "";
        }

        /// <summary>
        /// Initializes a new instance of the ReducingMetainstruction&lt;T&gt; class
        /// based on the reducing metainstruction data provider which implements
        /// IReducingMetainstructionData&lt;T&gt; interface.
        /// </summary>
        /// <param name="reducingMetainstructionData">Reducing metainstruction data provider.</param>
        public ReducingMetainstruction(IReducingMetainstructionData<T> reducingMetainstructionData)
        {
            this.alphabet = new Alphabet<T>(reducingMetainstructionData);
            this.leftLanguage = LanguageFactory<T>.GetLanguageInstance(
                reducingMetainstructionData.GetLeftLanguageData());
            this.rightLanguage = LanguageFactory<T>.GetLanguageInstance(
                reducingMetainstructionData.GetRightLanguageData());
            this.fromWord = GenericString<T>.Wrap(reducingMetainstructionData.GetFromWordData());
            this.toWord = GenericString<T>.Wrap(reducingMetainstructionData.GetToWordData());
            this.name = reducingMetainstructionData.GetReducingName();
            if (this.leftLanguage != null)
                this.alphabet.CheckAlphabet(this.leftLanguage.Alphabet);
            if (this.rightLanguage != null)
                this.alphabet.CheckAlphabet(this.rightLanguage.Alphabet);
            this.alphabet.CheckWord(this.fromWord);
            this.alphabet.CheckWord(this.toWord);
        }
        #endregion

        #region IReducingMetainstructionData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.alphabet.GetAlphabetData(); }
        /// Returns an interface that provides all the neccessary data
        /// describing the left language of the reducing metainstruction.
        public ILanguageData<T> GetLeftLanguageData() { return this.leftLanguage; }
        /// Returns an interface that provides all the neccessary data
        /// describing the right language of the reducing metainstruction.
        public ILanguageData<T> GetRightLanguageData() { return this.rightLanguage; }
        /// Returns data describing the from-word of the reducing metainstruction.
        public T[] GetFromWordData() { return this.fromWord.ToArray(); }
        /// Returns data describing the to-word of the reducing metainstruction.
        public T[] GetToWordData() { return this.toWord.ToArray(); }
        /// Returns the name of the reducing metainstruction.
        public string GetReducingName() { return this.name; }
        #endregion

        #region Properties
        /// <summary>
        /// Gets the alphabet of the reducing metainstruction.
        /// </summary>
        public Alphabet<T> Alphabet
        {
            get
            {
                return this.alphabet;
            }
        }

        /// <summary>
        /// Gets or sets the left language.
        /// </summary>
        public ILanguage<T> LeftLanguage
        {
            get
            {
                return this.leftLanguage;
            }
            set
            {
                this.alphabet.CheckAlphabet(value.Alphabet);
                this.leftLanguage = value;
            }
        }

        /// <summary>
        /// Gets or sets the right language.
        /// </summary>
        public ILanguage<T> RightLanguage
        {
            get
            {
                return this.rightLanguage;
            }
            set
            {
                this.alphabet.CheckAlphabet(value.Alphabet);
                this.rightLanguage = value;
            }
        }

        /// <summary>
        /// Gets or sets the from-word.
        /// </summary>
        public GenericString<T> FromWord
        {
            get { return this.fromWord; }
            set { this.alphabet.CheckWord(value);  this.fromWord = value; }
        }

        /// <summary>
        /// Gets or sets the to-word.
        /// </summary>
        public GenericString<T> ToWord
        {
            get { return this.toWord; }
            set { this.alphabet.CheckWord(value);  this.toWord = value; } 
        }

        /// <summary>
        /// Gets or sets the name of the accepting metainstruction.
        /// </summary>
        public string Name
        {
            get
            {
                return this.name;
            }
            set
            {
                this.name = value.Trim();
            }
        }
        #endregion

        #region Private Data Members
        private Alphabet<T> alphabet;
        private ILanguage<T> leftLanguage;
        private ILanguage<T> rightLanguage;
        private GenericString<T> fromWord;
        private GenericString<T> toWord;
        private string name;
        #endregion
    }

    /// <summary>
    /// Struct that holds all the necessary information
    /// about one reduction step in the reduction path.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public struct ReductionStep<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// The first (starting) word of the reduction step.
        public GenericString<T> word;
        /// The word that is reduced from the starting word.
        public GenericString<T> reducedWord;
        /// The first index of the substitution part of the starting word.
        public int subWordStart;
        /// The length of the substitution part of the starting word.
        public int subWordLength;
        /// The first index of the substituted part of the reduced word.
        public int reducedSubWordStart;
        /// The length of the substituted part of the reduced word.
        public int reducedSubWordLength;
        /// The index of used reducing metainstruction.
        public int reducingMetainstructionIndex;
        /// The name of used reducing metainstruction
        public string reducingMetainstructionName;
    }

    /// <summary>
    /// Struct that holds all the reduction paths from one word to another word.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public struct WordToWordPath<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// The starting word of the path.
        public GenericString<T> fromWord;
        /// The last word of the path.
        public GenericString<T> toWord;
        /// The array of reduction steps from the fromWord to the toWord.
        public ReductionStep<T>[] reductionPath;

        #region Methods
        /// <summary>
        /// Returns a String that represents the current Object. (Inherited from Object.)
        /// </summary>
        public override string ToString()
        {
            const int minimalCapacity = 256;
            StringBuilder stringBuilder = new StringBuilder(minimalCapacity);
            foreach (ReductionStep<T> reductionStep in reductionPath)
            {
                string word = reductionStep.word.ToString();
                stringBuilder.Append(word);
                stringBuilder.Append(" -");
                stringBuilder.Append(reductionStep.reducingMetainstructionName);
                stringBuilder.Append("-> ");
            }
            stringBuilder.Append(this.toWord.ToString());
            return stringBuilder.ToString();
        }

        /// <summary>
        /// Returns a String that represents the current Object.
        /// Adds special mark symbols [ and ].
        /// </summary>
        public string ToStringMarked()
        {
            const int minimalCapacity = 256;
            StringBuilder stringBuilder = new StringBuilder(minimalCapacity);
            foreach (ReductionStep<T> reductionStep in reductionPath)
            {
                string word = reductionStep.word.ToString();
                stringBuilder.Append(word.Substring(0, reductionStep.subWordStart));
                stringBuilder.Append("[");
                stringBuilder.Append(word.Substring(
                    reductionStep.subWordStart, reductionStep.subWordLength));
                stringBuilder.Append("]");
                stringBuilder.Append(word.Substring(
                    reductionStep.subWordStart + reductionStep.subWordLength));
                stringBuilder.Append(" -");
                stringBuilder.Append(reductionStep.reducingMetainstructionName);
                stringBuilder.Append("-> ");
            }
            stringBuilder.Append(this.toWord.ToString());
            return stringBuilder.ToString();
        }

        /// <summary>
        /// Appends the string representation of the current Object
        /// to the text of RichTextBox control.
        /// </summary>
        public void AppendToRichTextBox(System.Windows.Forms.RichTextBox richTextBox)
        {
            System.Drawing.Font normalFont = richTextBox.Font;
            System.Drawing.Font underlinedFont = new System.Drawing.Font(
                normalFont.FontFamily, normalFont.Size, System.Drawing.FontStyle.Underline);
            int reducedSubWordStart = 0;
            int reducedSubWordLength = 0;
            foreach (ReductionStep<T> reductionStep in reductionPath)
            {
                string word = reductionStep.word.ToString();
                richTextBox.SelectionLength = 0;
                richTextBox.SelectionFont = normalFont;
                richTextBox.SelectionColor = System.Drawing.Color.Black;
                richTextBox.AppendText(word.Substring(0, reductionStep.subWordStart));
                richTextBox.SelectionLength = 0;
                richTextBox.SelectionFont = underlinedFont;
                richTextBox.SelectionColor = System.Drawing.Color.Black;
                richTextBox.AppendText(word.Substring(
                    reductionStep.subWordStart, reductionStep.subWordLength));
                richTextBox.SelectionLength = 0;
                richTextBox.SelectionFont = normalFont;
                richTextBox.SelectionColor = System.Drawing.Color.Black;
                richTextBox.AppendText(word.Substring(
                    reductionStep.subWordStart + reductionStep.subWordLength));
                if (reducedSubWordLength > 0)
                {
                    richTextBox.SelectionStart = richTextBox.Text.Length -
                        word.Length + reducedSubWordStart;
                    richTextBox.SelectionLength = reducedSubWordLength;
                    richTextBox.SelectionColor = System.Drawing.Color.Blue;
                }
                richTextBox.SelectionStart = richTextBox.Text.Length;
                richTextBox.SelectionLength = 0;
                richTextBox.SelectionFont = normalFont;
                richTextBox.SelectionColor = System.Drawing.Color.Black;
                reducedSubWordStart = reductionStep.reducedSubWordStart;
                reducedSubWordLength = reductionStep.reducedSubWordLength;
                richTextBox.AppendText(" -");
                richTextBox.AppendText(reductionStep.reducingMetainstructionName);
                richTextBox.AppendText("-> ");
            }
            string lastWord = this.toWord.ToString();
            richTextBox.AppendText(lastWord);
            if (reducedSubWordLength > 0)
            {
                richTextBox.SelectionStart = richTextBox.Text.Length -
                    lastWord.Length + reducedSubWordStart;
                richTextBox.SelectionLength = reducedSubWordLength;
                richTextBox.SelectionColor = System.Drawing.Color.Blue;
            }
            richTextBox.SelectionStart = richTextBox.Text.Length;
            richTextBox.SelectionLength = 0;
            richTextBox.SelectionFont = normalFont;
            richTextBox.SelectionColor = System.Drawing.Color.Black;
            richTextBox.AppendText("\r\n");
        }
        #endregion
    }

    /// <summary>
    /// Struct that encapsulates a word with the list of the
    /// names of accepting metainstructions that accept this word.
    /// This word is accepted if and only if this list is not empty.
    /// </summary>
    public struct WordAcceptNames<T> :
        IComparable<WordAcceptNames<T>>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of the WordAcceptNames&lt;T&gt; struct.
        /// </summary>
        public WordAcceptNames(GenericString<T> word, string[] acceptingNames)
        {
            this.word = word;
            this.acceptingNames = acceptingNames;
            this.accept = acceptingNames.Length > 0;
        }
        #endregion

        /// Word.
        public GenericString<T> word;
        /// List of the names of accepting metainstructions that accept the word.
        public string[] acceptingNames;
        /// Acceptance of the word.
        public bool accept;

        #region IComparable Interface
        /// <summary>
        /// Standard enumeration based only on the generic word.
        /// </summary>
        public int CompareTo(WordAcceptNames<T> wordAcceptNames)
        {
            return this.word.CompareTo(wordAcceptNames.word);
        }
        #endregion
    }

    /// <summary>
    /// Class that represents a restarting automaton.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public sealed class RestartingAutomaton<T> :
        IRestartingAutomatonData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of RestartingAutomaton&lt;T&gt; class with given alphabet.
        /// </summary>
        /// <param name="alphabet">Input alphabet.</param>
        public RestartingAutomaton(Alphabet<T> alphabet)
        {
            const int minimalCapacity = 16;
            this.alphabet = alphabet;
            this.acceptingMetainstructionList =
                new List<AcceptingMetainstruction<T>>(minimalCapacity);
            this.reducingMetainstructionList =
                new List<ReducingMetainstruction<T>>(minimalCapacity);
        }

        /// <summary>
        /// Initializes a new instance of the RestartingAutomaton&lt;T&gt; class
        /// based on the restarting automaton data provider which implements
        /// IRestartingAutomatonData&lt;T&gt; interface.
        /// </summary>
        /// <param name="restartingAutomatonData">Restarting automaton data provider.</param>
        public RestartingAutomaton(IRestartingAutomatonData<T> restartingAutomatonData)
        {
            this.alphabet = new Alphabet<T>(restartingAutomatonData);
            IAcceptingMetainstructionData<T>[] acceptingMetainstructionsData =
                restartingAutomatonData.GetAcceptingMetainstructionsData();
            IReducingMetainstructionData<T>[] reducingMetainstructionsData =
                restartingAutomatonData.GetReducingMetainstructionsData();
            this.acceptingMetainstructionList =
                new List<AcceptingMetainstruction<T>>(acceptingMetainstructionsData.Length);
            this.reducingMetainstructionList =
                new List<ReducingMetainstruction<T>>(reducingMetainstructionsData.Length);
            foreach (IAcceptingMetainstructionData<T> acceptingMetainstructionData in acceptingMetainstructionsData)
                this.acceptingMetainstructionList.Add(new AcceptingMetainstruction<T>(acceptingMetainstructionData));
            foreach (IReducingMetainstructionData<T> reducingMetainstructionData in reducingMetainstructionsData)
                this.reducingMetainstructionList.Add(new ReducingMetainstruction<T>(reducingMetainstructionData));
            this.CheckConsistency();
        }
        #endregion

        #region IRestartingAutomatonData Interface
        /// <summary>
        /// Returns data describing the alphabet.
        /// </summary>
        public T[] GetAlphabetData()
        {
            return this.alphabet.GetAlphabetData();
        }

        /// <summary>
        /// Returns an array of IAcceptingMetainstructionData&lt;T&gt; interfaces describing
        /// the set of accepting metainstructions of the restarting automaton.
        /// </summary>
        public IAcceptingMetainstructionData<T>[] GetAcceptingMetainstructionsData()
        {
            List<IAcceptingMetainstructionData<T>> acceptingMetainstructionDataList =
                new List<IAcceptingMetainstructionData<T>>(this.acceptingMetainstructionList.Count);
            foreach (AcceptingMetainstruction<T> acceptingMetainstruction in this.acceptingMetainstructionList)
                acceptingMetainstructionDataList.Add(acceptingMetainstruction);
            return acceptingMetainstructionDataList.ToArray();
        }

        /// <summary>
        /// Returns an array of IReducingMetainstructionData&lt;T&gt; interfaces describing
        /// the set of reducing metainstructions of the restarting automaton.
        /// </summary>
        public IReducingMetainstructionData<T>[] GetReducingMetainstructionsData()
        {
            List<IReducingMetainstructionData<T>> reducingMetainstructionDataList =
                new List<IReducingMetainstructionData<T>>(this.reducingMetainstructionList.Count);
            foreach (ReducingMetainstruction<T> reducingMetainstruction in this.reducingMetainstructionList)
                reducingMetainstructionDataList.Add(reducingMetainstruction);
            return reducingMetainstructionDataList.ToArray();
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets the alphabet of the RestartingAutomaton&lt;T&gt; class.
        /// </summary>
        public Alphabet<T> Alphabet
        {
            get
            {
                return this.alphabet;
            }
        }

        /// <summary>
        /// Gets the read-only collection of accepting metainstructions.
        /// </summary>
        public ReadOnlyCollection<AcceptingMetainstruction<T>> AcceptingMetainstructions
        {
            get
            {
                return new ReadOnlyCollection<AcceptingMetainstruction<T>>(
                    this.acceptingMetainstructionList);
            }
        }

        /// <summary>
        /// Gets the read-only collection of reducing metainstructions.
        /// </summary>
        public ReadOnlyCollection<ReducingMetainstruction<T>> ReducingMetainstructions
        {
            get
            {
                return new ReadOnlyCollection<ReducingMetainstruction<T>>(
                    this.reducingMetainstructionList);
            }
        }
        #endregion

        #region Methods
        /// <summary>
        /// Adds an accepting metainstruction to the restarting automaton.
        /// </summary>
        public void AddAcceptingMetainstruction(AcceptingMetainstruction<T> acceptingMetainstruction)
        {
            this.alphabet.CheckAlphabet(acceptingMetainstruction.Alphabet);
            this.acceptingMetainstructionList.Add(acceptingMetainstruction);
        }

        /// <summary>
        /// Adds a reducing metainstruction to the restarting automaton.
        /// </summary>
        public void AddReducingMetainstruction(ReducingMetainstruction<T> reducingMetainstruction)
        {
            this.alphabet.CheckAlphabet(reducingMetainstruction.Alphabet);
            this.reducingMetainstructionList.Add(reducingMetainstruction);
        }

        /// <summary>
        /// Returns the list of all reducing paths from the fromWord to the toWord.
        /// </summary>
        public List<WordToWordPath<T>> WordToWordPaths(GenericString<T> fromWord, GenericString<T> toWord,
            System.ComponentModel.BackgroundWorker backgroundWorker)
        {
            const int minimalCapacity = 32;
            this.wordToWordPathBuffer = new List<WordToWordPath<T>>(minimalCapacity);
            this.reductionStepBuffer = new List<ReductionStep<T>>(minimalCapacity);
            this.tempFromWord = fromWord;
            this.tempToWord = toWord;
            this.ReduceWordPath(fromWord, backgroundWorker);
            return this.wordToWordPathBuffer;
        }

        /// <summary>
        /// Returns the list of all words that can be reduced from the fromWord.
        /// </summary>
        public List<WordAcceptNames<T>> WordToAll(GenericString<T> fromWord,
            System.ComponentModel.BackgroundWorker backgroundWorker)
        {
            const int minimalCapacity = 32;
            List<WordAcceptNames<T>> wordList =
                new List<WordAcceptNames<T>>(minimalCapacity);
            SortedDictionary<GenericString<T>, bool> wordCache =
                new SortedDictionary<GenericString<T>, bool>();
            string[] fromWordAcceptNames = this.DecideStepNames(fromWord);
            wordList.Add(new WordAcceptNames<T>(fromWord, fromWordAcceptNames));
            wordCache[fromWord] = true;
            for (int wordIndex = 0; wordIndex < wordList.Count; ++wordIndex)
            {
                if (backgroundWorker != null &&
                    backgroundWorker.CancellationPending)
                    break;
                GenericString<T> word = wordList[wordIndex].word;
                List<ReductionStep<T>> reductionStepList = this.ReduceWordStep(word);
                foreach (ReductionStep<T> reductionStep in reductionStepList)
                {
                    GenericString<T> reducedWord = reductionStep.reducedWord;
                    if (!wordCache.ContainsKey(reducedWord))
                    {
                        string[] reducedWordAcceptNames = this.DecideStepNames(reducedWord);
                        wordList.Add(new WordAcceptNames<T>(reducedWord, reducedWordAcceptNames));
                        wordCache[reducedWord] = true;
                    }
                }
            }
            wordList.Sort();
            return wordList;
        }

        /// <summary>
        /// Decides whether the given word is accepted by at least one of the accepting metainstructions.
        /// </summary>
        public bool DecideStep(GenericString<T> word)
        {
            foreach (AcceptingMetainstruction<T> acceptingMetainstruction in this.acceptingMetainstructionList)
                if (acceptingMetainstruction.AcceptingLanguage.Decide(word))
                    return true;
            return false;
        }

        /// <summary>
        /// Returns an array of names of accepting metainstructions that accept the input word.
        /// </summary>
        public string[] DecideStepNames(GenericString<T> word)
        {
            List<string> acceptingNames = new List<string>(this.acceptingMetainstructionList.Count);
            foreach (AcceptingMetainstruction<T> acceptingMetainstruction in this.acceptingMetainstructionList)
                if (acceptingMetainstruction.AcceptingLanguage.Decide(word))
                    acceptingNames.Add(acceptingMetainstruction.Name);
            return acceptingNames.ToArray();
        }

        /// <summary>
        /// Decides whether the given word can be reduced to an accepted word.
        /// </summary>
        public bool Decide(GenericString<T> word)
        {
            List<WordAcceptNames<T>> reducedWordList = this.WordToAll(word, null);
            foreach (WordAcceptNames<T> reducedWord in reducedWordList)
                if (reducedWord.accept) return true;
            return false;
        }
        #endregion

        #region Private Methods
        /// <summary>
        /// Checks the consistency of the private data members.
        /// </summary>
        private void CheckConsistency()
        {
            foreach (AcceptingMetainstruction<T> acceptingMetainstruction in this.acceptingMetainstructionList)
                this.alphabet.CheckAlphabet(acceptingMetainstruction.Alphabet);
            foreach (ReducingMetainstruction<T> reducingMetainstruction in this.reducingMetainstructionList)
                this.alphabet.CheckAlphabet(reducingMetainstruction.Alphabet);
        }

        /// <summary>
        /// Returns the list of words (ReductionStep&lt;T&gt; structures)
        /// that can be reduced from the given word by using exactly one
        /// reducing metainstruction.
        /// </summary>
        private List<ReductionStep<T>> ReduceWordStep(GenericString<T> word)
        {
            const int minimalCapacity = 32;
            List<ReductionStep<T>> reductionStepList =
                new List<ReductionStep<T>>(minimalCapacity);
            for (int reducingMetainstructionIndex = 0;
                reducingMetainstructionIndex < 
                    this.reducingMetainstructionList.Count;
                ++reducingMetainstructionIndex)
            {
                ReducingMetainstruction<T> reducingMetainstruction =
                    this.reducingMetainstructionList[reducingMetainstructionIndex];
                GenericString<T> fromSubWord = reducingMetainstruction.FromWord;
                GenericString<T> toSubWord = reducingMetainstruction.ToWord;
                int wordLength = word.Length;
                int subWordLength = fromSubWord.Length;
                for (int subWordStart = 0; subWordStart < wordLength - subWordLength + 1; ++subWordStart)
                    if (word.EqualsFrom(fromSubWord, subWordStart))
                    {
                        GenericString<T> leftPart = word.SubString(0, subWordStart);
                        GenericString<T> rightPart = word.SubString(subWordStart + subWordLength,
                            wordLength - subWordStart - subWordLength);
                        if (reducingMetainstruction.LeftLanguage.Decide(leftPart) &&
                            reducingMetainstruction.RightLanguage.Decide(rightPart))
                        {
                            GenericString<T> reducedWord = leftPart + toSubWord + rightPart;
                            ReductionStep<T> reductionStep = new ReductionStep<T>();
                            reductionStep.word = word;
                            reductionStep.reducedWord = reducedWord;
                            reductionStep.subWordStart = subWordStart;
                            reductionStep.subWordLength = subWordLength;
                            reductionStep.reducedSubWordStart = leftPart.Length;
                            reductionStep.reducedSubWordLength = toSubWord.Length;
                            reductionStep.reducingMetainstructionIndex = reducingMetainstructionIndex;
                            reductionStep.reducingMetainstructionName = reducingMetainstruction.Name;
                            reductionStepList.Add(reductionStep);
                        }
                    }
            }
            return reductionStepList;
        }

        /// <summary>
        /// Fills the wordToWordPathBuffer that holds all the reduction paths
        /// from the word tempFromWord to the word tempToWord.
        /// </summary>
        private void ReduceWordPath(GenericString<T> word,
            System.ComponentModel.BackgroundWorker backgroundWorker)
        {
            if (backgroundWorker != null &&
                backgroundWorker.CancellationPending)
                return;
            if (word.Equals(tempToWord))
            {
                WordToWordPath<T> wordToWordPath = new WordToWordPath<T>();
                wordToWordPath.fromWord = this.tempFromWord;
                wordToWordPath.toWord = this.tempToWord;
                wordToWordPath.reductionPath = this.reductionStepBuffer.ToArray();
                this.wordToWordPathBuffer.Add(wordToWordPath);
            }
            else
            {
                List<ReductionStep<T>> reductionStepList = this.ReduceWordStep(word);
                foreach (ReductionStep<T> reductionStep in reductionStepList)
                {
                    this.reductionStepBuffer.Add(reductionStep);
                    this.ReduceWordPath(reductionStep.reducedWord, backgroundWorker);
                    this.reductionStepBuffer.RemoveAt(
                        reductionStepBuffer.Count - 1);
                }
            }
        }
        #endregion

        #region Private Data Members
        private Alphabet<T> alphabet;
        private List<AcceptingMetainstruction<T>> acceptingMetainstructionList;
        private List<ReducingMetainstruction<T>> reducingMetainstructionList;
        private List<WordToWordPath<T>> wordToWordPathBuffer;
        private List<ReductionStep<T>> reductionStepBuffer;
        private GenericString<T> tempFromWord;
        private GenericString<T> tempToWord;
        #endregion
    }

    /// <summary>
    /// Class that represents a static language factory.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public static class LanguageFactory<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// <summary>
        /// Returns the instance of the language class that corresponds to the given
        /// ILanguageData&lt;T&gt; interface.
        /// </summary>
        public static ILanguage<T> GetLanguageInstance(
            ILanguageData<T> languageData)
        {
            if (languageData == null)
                return null;
            else if (languageData is ILanguageDFALStarData<T>)
                return new LanguageDFALStar<T>(languageData as ILanguageDFALStarData<T>);
            else if (languageData is ILanguageDFARPNIData<T>)
                return new LanguageDFARPNI<T>(languageData as ILanguageDFARPNIData<T>);
            else if (languageData is ILanguageDFAData<T>)
                return new LanguageDFA<T>(languageData as ILanguageDFAData<T>);
            else if (languageData is ILanguageRegexData<T>)
                return new LanguageRegex<T>(languageData as ILanguageRegexData<T>);
            else if (languageData is ILanguageSLTData<T>)
                return new LanguageSLT<T>(languageData as ILanguageSLTData<T>);
            else throw new InvalidProgramException("LanguageFactory: Unrecognized ILanguageData interface.");
        }
    }
}