using System;

namespace RegularLanguages
{
    /// <summary>
    /// LanguageForm&lt;T&gt; is an interface that should be implemented by every
    /// form that provides a language ILanguage&lt;T&gt;.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface LanguageForm<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// <summary>
        /// Gets the language constructed in the form.
        /// </summary>
        ILanguage<T> Language
        {
            get;
        }
    }

    /// <summary>
    /// Class that represents a static language form factory.
    /// </summary>
    public static class LanguageFormFactory
    {
        /// <summary>
        /// Returns the instance of the language form that corresponds to the given
        /// ILanguageData&lt;char&gt; interface.
        /// </summary>
        public static System.Windows.Forms.Form GetLanguageFormInstance(
            ILanguageData<char> languageData)
        {
            if (languageData is ILanguageDFALStarData<char>)
                return new LStarForm(new LStar<char>(languageData as ILanguageDFALStarData<char>));
            else if (languageData is ILanguageDFARPNIData<char>)
                return new RPNIForm(new RPNI<char>(languageData as ILanguageDFARPNIData<char>));
            else if (languageData is ILanguageDFAData<char>)
                return new DFAModelerForm(new DFA<char>(languageData as ILanguageDFAData<char>));
            else if (languageData is ILanguageRegexData<char>)
                return new RegexForm(new LanguageRegex<char>(languageData as ILanguageRegexData<char>));
            else if (languageData is ILanguageSLTData<char>)
                return new SLTForm(new SLT<char>(languageData as ILanguageSLTData<char>));
            else throw new InvalidProgramException("LanguageFormFactory: Unrecognized ILanguageData interface.");
        }
    }
}