using System;
using System.Text;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Channels;
using System.Runtime.Remoting.Channels.Http;
using System.IO;

namespace RegularLanguages
{
    /// <summary>
    /// Class that represents a service provided for clients.
    /// This class mediates the functionalities of the restarting automaton
    /// stored in the RestartingAutomatonServer static class.
    /// </summary>
    public class RestartingAutomatonService : MarshalByRefObject
    {
        /// <summary>
        /// Returns the alphabet as a comma-separated string of alphabet tokens.
        /// </summary>
        public string GetAlphabet()
        {
            try
            {
                return RestartingAutomatonServer.
                    restartingAutomaton.Alphabet.ToString();
            }
            catch { return null; }
        }

        /// <summary>
        /// Decides whether the given word has all symbols from the alphabet.
        /// </summary>
        public bool IsFromAlphabet(string word)
        {
            try
            {
                return RestartingAutomatonServer.
                    restartingAutomaton.Alphabet.Contains(word);
            }
            catch { return false; }
        }

        /// <summary>
        /// Decides whether the given word is accepted or rejected.
        /// </summary>
        public bool Decide(string word)
        {
            try
            {
                if (RestartingAutomatonServer.
                        restartingAutomaton.Alphabet.Contains(word))
                    return RestartingAutomatonServer.restartingAutomaton.
                        Decide(new GenericString<char>(word));
                else return false;
            }
            catch { return false; }
        }

        /// <summary>
        /// Returns the string of all reducing paths from the fromWord to the toWord.
        /// </summary>
        public string WordToWord(string fromWord, string toWord)
        {
            try
            {
                StringBuilder stringBuilder = new StringBuilder(1024);
                GenericString<char> genericFromWord =
                    new GenericString<char>(fromWord);
                GenericString<char> genericToWord =
                    new GenericString<char>(toWord);
                foreach (WordToWordPath<char> path in
                    RestartingAutomatonServer.
                        restartingAutomaton.WordToWordPaths(
                            genericFromWord, genericToWord, null))
                    stringBuilder.AppendLine(path.ToString());
                return stringBuilder.ToString();
            }
            catch { return null; }
        }

        /// <summary>
        /// Returns the string of all words that can be reduced from the fromWord.
        /// </summary>
        public string WordToAll(string fromWord)
        {
            try
            {
                StringBuilder stringBuilder = new StringBuilder(1024);
                GenericString<char> genericFromWord =
                    new GenericString<char>(fromWord);
                foreach (WordAcceptNames<char> wanames in
                    RestartingAutomatonServer.
                        restartingAutomaton.WordToAll(
                            genericFromWord, null))
                {
                    stringBuilder.Append(wanames.word.ToString());
                    if (wanames.accept)
                    {
                        stringBuilder.Append(" [");
                        stringBuilder.Append(
                            String.Join(", ", wanames.acceptingNames));
                        stringBuilder.Append("]");
                    }
                    stringBuilder.AppendLine();
                }
                return stringBuilder.ToString();
            }
            catch { return null; }
        }
    }

    /// <summary>
    /// Static class that represents a server.
    /// </summary>
    public static class RestartingAutomatonServer
    {
        /// <summary>
        /// Runs the server on the port channelPort.
        /// </summary>
        public static void Run(int channelPort, string serviceName,
            RestartingAutomaton<char> restartingAutomaton,
            System.ComponentModel.BackgroundWorker backgroundWorker)
        {
            HttpChannel channel = new HttpChannel(channelPort);
            RestartingAutomatonServer.restartingAutomaton = restartingAutomaton;
            ChannelServices.RegisterChannel(channel, false);
            RemotingConfiguration.RegisterWellKnownServiceType(
                typeof(RestartingAutomatonService),
                serviceName, WellKnownObjectMode.Singleton);
            while (!backgroundWorker.CancellationPending)
                System.Threading.Thread.Sleep(100);
            ChannelServices.UnregisterChannel(channel);
        }

        public static RestartingAutomaton<char> restartingAutomaton;
    }
}