using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Text;
using System.Xml.Serialization;

namespace RegularLanguages
{
    /// <summary>
    /// The root interface of all the interfaces that provide some data.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface IObjectData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {

    }

    /// <summary>
    /// IAlphabetData&lt;T&gt; interface provides all the neccessary data describing an alphabet.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface IAlphabetData<T> :
        IObjectData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// <summary>
        /// Returns data describing the alphabet.
        /// Always return these data as a clone.
        /// </summary>
        T[] GetAlphabetData();
    }

    /// <summary>
    /// IDFAData&lt;T&gt; interface provides all the neccessary data describing a DFA.
    /// IDFAData&lt;T&gt; interface also extends IAlphabetData&lt;T&gt; interface.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface IDFAData<T> :
        IAlphabetData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// <summary>
        /// Returns the number of states of the DFA.
        /// </summary>
        int GetDFAStateCountData();
        /// <summary>
        /// Returns the string that consist of the indices of the accepted states
        /// separated by a single space.
        /// </summary>
        string GetDFAAcceptStatesData();
        /// <summary>
        /// Returns the content of the transition table as a string of numbers
        /// separated by a single space.
        /// </summary>
        string GetDFATransitionTableData();
        /// <summary>
        /// Returns a comment (a short description) of the dfa.
        /// </summary>
        string GetDFACommentData();
    }

    /// <summary>
    /// Struct that encapsulates a basic piece of information in the knowledge base.
    /// (A word together with the information whether it is accepted or rejected.)
    /// This struct is constructed to be easy serialized and deserialized.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("WordAccept")]
    public struct WordAcceptData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// Word as an array of tokens.
        [XmlArray("Word"), XmlArrayItem("T")]
        public T[] word;
        /// Acceptance of the word.
        [XmlElement("Accept")]
        public bool accept;
        /// <summary>
        /// Returns a new instance of the WordAcceptData&lt;T&gt; struct that is a deep copy of this instance.
        /// </summary>
        public WordAcceptData<T> Clone()
        {
            WordAcceptData<T> wordAcceptData = 
                new WordAcceptData<T>();
            wordAcceptData.word = this.word.Clone() as T[];
            wordAcceptData.accept = this.accept;
            return wordAcceptData;
        }
    }

    /// <summary>
    /// Struct that encapsulates a sample (positive or negative).
    /// This struct is constructed to be easy serialized and deserialized.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("Sample")]
    public struct SampleData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// Word as an array of tokens.
        [XmlArray("Word"), XmlArrayItem("T")]
        public T[] word;
        /// <summary>
        /// Returns a new instance of the SampleData&lt;T&gt; struct that is a deep copy of this instance.
        /// </summary>
        public SampleData<T> Clone()
        {
            SampleData<T> sampleData =
                new SampleData<T>();
            sampleData.word = this.word.Clone() as T[];
            return sampleData;
        }
    }

    /// <summary>
    /// IKnowledgeBaseData&lt;T&gt; interface provides all the neccessary data describing a knowledge base.
    /// IKnowledgeBaseData&lt;T&gt; interface also extends IAlphabetData&lt;T&gt; interface.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface IKnowledgeBaseData<T> :
        IAlphabetData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// <summary>
        /// Returns an array of WordAcceptData&lt;T&gt; structures describing the whole knowledge base.
        /// Always return this array as a deep clone.
        /// </summary>
        WordAcceptData<T>[] GetKnowledgeBaseData();
        /// <summary>
        /// Returns a comment (a short description) of the knowledge base.
        /// </summary>
        string GetKnowledgeBaseCommentData();
    }

    /// <summary>
    /// IPositiveNegativeSamplesData&lt;T&gt; interface provides all the neccessary 
    /// data describing the set of positive and negative examples.
    /// IPositiveNegativeSamplesData&lt;T&gt; interface also extends IAlphabetData&lt;T&gt; interface.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface IPositiveNegativeSamplesData<T> :
        IAlphabetData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// <summary>
        /// Returns special options for positive and negative samples.
        /// </summary>
        PositiveNegativeSamplesOptions GetPNSamplesOptions();
        /// <summary>
        /// Returns an array of SampleData&lt;T&gt; structures describing the set of positive samples.
        /// Always return this array as a deep clone.
        /// </summary>
        SampleData<T>[] GetPositiveSamplesData();
        /// <summary>
        /// Returns an array of SampleData&lt;T&gt; structures describing the set of negative samples.
        /// Always return this array as a deep clone.
        /// </summary>
        SampleData<T>[] GetNegativeSamplesData();
        /// <summary>
        /// Returns a comment (a short description) of the positive and negative samples.
        /// </summary>
        string GetPNSamplesCommentData();
    }

    /// <summary>
    /// ILStarData&lt;T&gt; interface provides all the data necessary
    /// to initialize an instance of LStar class.
    /// ILStarData&lt;T&gt; interface extends IKnowledgeBaseData&lt;T&gt; interface.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface ILStarData<T> :
        IKnowledgeBaseData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {

    }

    /// <summary>
    /// IRPNIData&lt;T&gt; interface provides all data necessary
    /// to initialize an instance of RPNI class.
    /// IRPNIData&lt;T&gt; interface extends IPositiveNegativeSamplesData&lt;T&gt; interface.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface IRPNIData<T> :
        IPositiveNegativeSamplesData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {

    }

    /// <summary>
    /// ISLTData&lt;T&gt; interface provides all data neccessary
    /// to initalize an instance of SLT class.
    /// ISLTData&lt;T&gt; interface extends IPositiveNegativeSamplesData&lt;T&gt; interface.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface ISLTData<T> :
        IPositiveNegativeSamplesData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// <summary>
        /// Returns k for k-SLT languages.
        /// </summary>
        int GetSLTK();
    }

    /// <summary>
    /// The root interface of all the interfaces that provide some data of a language.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface ILanguageData<T> :
        IObjectData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {

    }

    /// <summary>
    /// ILanguageDFAData&lt;T&gt; interface provides all the neccessary data describing
    /// a language based on the DFA.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface ILanguageDFAData<T> :
        ILanguageData<T>,
        IDFAData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {

    }

    /// <summary>
    /// ILanguageDFALStarData&lt;T&gt; interface provides all the neccessary data describing
    /// a language based on the DFA constructed by LStar (Dana Angluin's L*) algorithm.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface ILanguageDFALStarData<T> :
        ILanguageDFAData<T>,
        ILStarData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {

    }

    /// <summary>
    /// ILanguageDFARPNIData&lt;T&gt; interface provides all the neccessary data describing
    /// a language based on the DFA constructed by RPNI algorithm.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface ILanguageDFARPNIData<T> :
        ILanguageDFAData<T>,
        IRPNIData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {

    }

    /// <summary>
    /// ILanguageRegexData&lt;T&gt; interface provides all the neccessary data describing
    /// a language based on the regular expression.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface ILanguageRegexData<T> :
        IAlphabetData<T>,
        ILanguageData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// <summary>
        /// Returns the string pattern of the regular expression.
        /// </summary>
        string GetRegexData();
    }

    /// <summary>
    /// ILanguageSLTData&lt;T&gt; interface provides all the necessary data describing
    /// a language based on the SLT&lt;T&gt; class.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface ILanguageSLTData<T> :
        ISLTData<T>,
        ILanguageData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {

    }

    /// <summary>
    /// IAcceptingMetainstructionData&lt;T&gt; interface provides all the neccessary data
    /// decsribing an accepting metainstruction.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface IAcceptingMetainstructionData<T> :
        IAlphabetData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// <summary>
        /// Returns an interface that provides all the neccessary data
        /// describing the accepting language of the accepting metainstruction.
        /// </summary>
        ILanguageData<T> GetAcceptingLanguageData();
        /// <summary>
        /// Returns the name of the accepting metainstruction.
        /// </summary>
        string GetAcceptingName();
    }

    /// <summary>
    /// IReducingMetainstructionData&lt;T&gt; interface provides all the neccessary data
    /// decsribing a reducing metainstruction.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public interface IReducingMetainstructionData<T> :
        IAlphabetData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// <summary>
        /// Returns an interface that provides all the neccessary data
        /// describing the left language of the reducing metainstruction.
        /// </summary>
        ILanguageData<T> GetLeftLanguageData();
        /// <summary>
        /// Returns an interface that provides all the neccessary data
        /// describing the right language of the reducing metainstruction.
        /// </summary>
        ILanguageData<T> GetRightLanguageData();
        /// <summary>
        /// Returns data describing the from-word of the reducing metainstruction.
        /// Always return these data as a clone.
        /// </summary>
        T[] GetFromWordData();
        /// <summary>
        /// Returns data describing the to-word of the reducing metainstruction.
        /// Always return these data as a clone.
        /// </summary>
        T[] GetToWordData();
        /// <summary>
        /// Returns the name of the reducing metainstruction.
        /// </summary>
        string GetReducingName();
    }

    /// <summary>
    /// IRestartingAutomatonData&lt;T&gt; interface provides all the neccessary data
    /// decsribing a restarting automaton.
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public interface IRestartingAutomatonData<T> :
        IAlphabetData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// <summary>
        /// Returns an array of IAcceptingMetainstructionData&lt;T&gt; interfaces describing
        /// the set of accepting metainstructions of the restarting automaton.
        /// </summary>
        IAcceptingMetainstructionData<T>[] GetAcceptingMetainstructionsData();
        /// <summary>
        /// Returns an array of IReducingMetainstructionData&lt;T&gt; interfaces describing
        /// the set of reducing metainstructions of the restarting automaton.
        /// </summary>
        IReducingMetainstructionData<T>[] GetReducingMetainstructionsData();
    }

    /// <summary>
    /// Class that represents an XML representation of the alphabet.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("Alphabet")]
    public sealed class AlphabetXML<T> :
        IAlphabetData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        public AlphabetXML()
        {
            this.alphabetData = null;
        }

        /// <summary>
        /// Initializes a new instance of the AlphabetXML&lt;T&gt; class
        /// based on the alphabet data provider which implements
        /// IAlphabetData&lt;T&gt; interface.
        /// </summary>
        /// <param name="alphabetData">Alphabet data provider.</param>
        public AlphabetXML(IAlphabetData<T> alphabetData)
        {
            this.alphabetData = alphabetData.GetAlphabetData();
        }
        #endregion

        #region IAlphabetData Interface
        /// <summary>
        /// Returns data describing this alphabet.
        /// </summary>
        public T[] GetAlphabetData() { return this.alphabetData.Clone() as T[]; }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the alphabet tokens.
        /// </summary>
        [XmlArray("Tokens"), XmlArrayItem("T")]
        public T[] Tokens
        {
            get { return this.alphabetData; }
            set { this.alphabetData = value; }
        }
        #endregion

        #region Private Data Members
        private T[] alphabetData;
        #endregion
    }

    /// <summary>
    /// Class that represents an XML representation of the DFA.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("DFA")]
    public sealed class DFAXML<T> :
        IDFAData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        public DFAXML()
        {
            this.alphabetXML = null;
            this.stateCount = 0;
            this.acceptStatesString = null;
            this.transitionTableString = null;
            this.comment = null;
        }

        /// <summary>
        /// Initializes a new instance of the DFAXML&lt;T&gt; class
        /// based on the DFA data provider which implements
        /// IDFAData&lt;T&gt; interface.
        /// </summary>
        /// <param name="dfaData">DFA data provider.</param>
        public DFAXML(IDFAData<T> dfaData)
        {
            this.alphabetXML = new AlphabetXML<T>(dfaData);
            this.stateCount = dfaData.GetDFAStateCountData();
            this.acceptStatesString = dfaData.GetDFAAcceptStatesData();
            this.transitionTableString = dfaData.GetDFATransitionTableData();
            this.comment = dfaData.GetDFACommentData();
        }
        #endregion

        #region IDFAData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.alphabetXML.GetAlphabetData(); }
        /// Returns the number of states of this DFA.
        public int GetDFAStateCountData() { return this.stateCount; }
        /// Returns the string that consist of the indices of the accepted states
        /// separated by a single space.
        public string GetDFAAcceptStatesData() { return this.acceptStatesString; }
        /// Returns the content of the transition table as a string of numbers
        /// separated by a single space.
        public string GetDFATransitionTableData() { return this.transitionTableString; }
        /// Returns a comment (a short description) of the dfa.
        public string GetDFACommentData() { return this.comment; }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the XML representation of the alphabet.
        /// </summary>
        [XmlElement("Alphabet")]
        public AlphabetXML<T> Alphabet
        {
            get { return this.alphabetXML; }
            set { this.alphabetXML = value; }
        }

        /// <summary>
        /// Gets or sets the number of states.
        /// </summary>
        [XmlElement("StateCount")]
        public int StateCount
        {
            get { return this.stateCount; }
            set { this.stateCount = value; }
        }

        /// <summary>
        /// Gets or sets the string of accept states.
        /// </summary>
        [XmlElement("AcceptStates")]
        public string AcceptStates
        {
            get { return this.acceptStatesString; }
            set { this.acceptStatesString = value; }
        }

        /// <summary>
        /// Gets or sets the string of transition table.
        /// </summary>
        [XmlElement("TransitionTable")]
        public string TransitionTable
        {
            get { return this.transitionTableString; }
            set { this.transitionTableString = value; }
        }

        /// <summary>
        /// Gets or sets the comment (a short description) of the knowledge base.
        /// </summary>
        [XmlElement("Comment")]
        public string Comment
        {
            get { return this.comment; }
            set { this.comment = value; }
        }
        #endregion

        #region Private Data Members
        private AlphabetXML<T> alphabetXML;
        private int stateCount;
        private string acceptStatesString;
        private string transitionTableString;
        private string comment;
        #endregion
    }

    /// <summary>
    /// Class that represents an XML representation of the knowledge base.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("KnowledgeBase")]
    public sealed class KnowledgeBaseXML<T> :
        IKnowledgeBaseData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        public KnowledgeBaseXML()
        {
            this.alphabetXML = null;
            this.knowledgeBaseData = null;
            this.comment = null;
        }

        /// <summary>
        /// Initializes a new instance of the KnowledgeBaseXML&lt;T&gt; class
        /// based on the KnowledgeBase data provider which implements
        /// IKnowledgeBaseData&lt;T&gt; interface.
        /// </summary>
        /// <param name="knowledgeBaseData">KnowledgeBase data provider.</param>
        public KnowledgeBaseXML(IKnowledgeBaseData<T> knowledgeBaseData)
        {
            this.alphabetXML = new AlphabetXML<T>(knowledgeBaseData);
            this.knowledgeBaseData = knowledgeBaseData.GetKnowledgeBaseData();
            this.comment = knowledgeBaseData.GetKnowledgeBaseCommentData();
        }
        #endregion

        #region IKnowledgeBaseData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.alphabetXML.GetAlphabetData(); }
        /// Returns an array of WordAcceptData&lt;T&gt; structures describing the whole knowledge base.
        public WordAcceptData<T>[] GetKnowledgeBaseData()
        {
            List<WordAcceptData<T>> wordAcceptDataList =
                new List<WordAcceptData<T>>(knowledgeBaseData.Length);
            foreach (WordAcceptData<T> wordAcceptData in knowledgeBaseData)
                wordAcceptDataList.Add(wordAcceptData.Clone());
            return wordAcceptDataList.ToArray();
        }
        /// Returns a comment (a short description) of the knowledge base.
        public string GetKnowledgeBaseCommentData() { return this.comment; }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the XML representation of the alphabet.
        /// </summary>
        [XmlElement("Alphabet")]
        public AlphabetXML<T> Alphabet
        {
            get { return this.alphabetXML; }
            set { this.alphabetXML = value; }
        }

        /// <summary>
        /// Gets or sets the array of WordAcceptData&lt;T&gt; structures describing the knowledge base.
        /// </summary>
        [XmlArray("Knowledge"), XmlArrayItem("WordAccept")]
        public WordAcceptData<T>[] Knowledge
        {
            get { return this.knowledgeBaseData; }
            set { this.knowledgeBaseData = value; }
        }

        /// <summary>
        /// Gets or sets the comment (a short description) of the knowledge base.
        /// </summary>
        [XmlElement("Comment")]
        public string Comment
        {
            get { return this.comment; }
            set { this.comment = value; }
        }
        #endregion

        #region Private Data Members
        private AlphabetXML<T> alphabetXML;
        private WordAcceptData<T>[] knowledgeBaseData;
        private string comment;
        #endregion
    }

    /// <summary>
    /// Class that represents an XML representation of the positive and negative samples.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("PositiveNegativeSamples")]
    public sealed class PositiveNegativeSamplesXML<T> :
        IPositiveNegativeSamplesData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        public PositiveNegativeSamplesXML()
        {
            this.alphabetXML = null;
            this.options = PositiveNegativeSamplesOptions.AllSamplesEnabled;
            this.positiveSamplesData = null;
            this.negativeSamplesData = null;
            this.comment = null;
        }

        /// <summary>
        /// Initializes a new instance of the PositiveNegativeSamplesXML&lt;T&gt; class
        /// based on the PositiveNegativeSamples data provider which implements
        /// IPositiveNegativeSamplesData&lt;T&gt; interface.
        /// </summary>
        /// <param name="positiveNegativeSamplesData">PositiveNegativeSamples data provider.</param>
        public PositiveNegativeSamplesXML(IPositiveNegativeSamplesData<T> positiveNegativeSamplesData)
        {
            this.alphabetXML = new AlphabetXML<T>(positiveNegativeSamplesData);
            this.options = positiveNegativeSamplesData.GetPNSamplesOptions();
            this.positiveSamplesData = positiveNegativeSamplesData.GetPositiveSamplesData();
            this.negativeSamplesData = positiveNegativeSamplesData.GetNegativeSamplesData();
            this.comment = positiveNegativeSamplesData.GetPNSamplesCommentData();
        }
        #endregion

        #region IPositiveNegativeSamplesData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.alphabetXML.GetAlphabetData(); }
        /// Returns special options for positive and negative samples.
        public PositiveNegativeSamplesOptions GetPNSamplesOptions() { return this.options; }
        /// Returns an array of SampleData&lt;T&gt; structures describing the set of positive samples.
        public SampleData<T>[] GetPositiveSamplesData()
        {
            List<SampleData<T>> positiveSampleDataList =
                new List<SampleData<T>>(this.positiveSamplesData.Length);
            foreach (SampleData<T> sampleData in this.positiveSamplesData)
                positiveSampleDataList.Add(sampleData.Clone());
            return positiveSampleDataList.ToArray();
        }
        /// Returns an array of SampleData&lt;T&gt; structures describing the set of negative samples.
        public SampleData<T>[] GetNegativeSamplesData()
        {
            List<SampleData<T>> negativeSampleDataList =
                new List<SampleData<T>>(this.negativeSamplesData.Length);
            foreach (SampleData<T> sampleData in this.negativeSamplesData)
                negativeSampleDataList.Add(sampleData.Clone());
            return negativeSampleDataList.ToArray();
        }
        /// Returns a comment (a short description) of the positive and negative samples.
        public string GetPNSamplesCommentData() { return this.comment; }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the XML representation of the alphabet.
        /// </summary>
        [XmlElement("Alphabet")]
        public AlphabetXML<T> Alphabet
        {
            get { return this.alphabetXML; }
            set { this.alphabetXML = value; }
        }

        /// <summary>
        /// Gets or sets the XML representation of special options.
        /// </summary>
        [XmlElement("Options")]
        public PositiveNegativeSamplesOptions Options
        {
            get { return this.options; }
            set { this.options = value; }
        }

        /// <summary>
        /// Gets or sets the array of SampleData&lt;T&gt; structures 
        /// describing the set of positive samples.
        /// </summary>
        [XmlArray("PositiveSamples"), XmlArrayItem("Sample")]
        public SampleData<T>[] PositiveSamples
        {
            get { return this.positiveSamplesData; }
            set { this.positiveSamplesData = value; }
        }

        /// <summary>
        /// Gets or sets the array of SampleData&lt;T&gt; structures 
        /// describing the set of negative samples.        
        /// </summary>
        [XmlArray("NegativeSamples"), XmlArrayItem("Sample")]
        public SampleData<T>[] NegativeSamples
        {
            get { return this.negativeSamplesData; }
            set { this.negativeSamplesData = value; }
        }

        /// <summary>
        /// Gets or sets the comment (a short description) of the positive and negative samples.
        /// </summary>
        [XmlElement("Comment")]
        public string Comment
        {
            get { return this.comment; }
            set { this.comment = value; }
        }
        #endregion

        #region Private Data Members
        private AlphabetXML<T> alphabetXML;
        private PositiveNegativeSamplesOptions options;
        private SampleData<T>[] positiveSamplesData;
        private SampleData<T>[] negativeSamplesData;
        private string comment;
        #endregion
    }

    /// <summary>
    /// Class that represents an XML representation of the SLT class.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("SLT")]
    public sealed class SLTXML<T> :
        ISLTData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        public SLTXML()
        {
            this.positiveNegativeSamplesXML = null;
            this.k = 0;
        }

        /// <summary>
        /// Initializes a new instance of the SLTXML&lt;T&gt; class
        /// based on the SLT data provider which implements
        /// ISLTData&lt;T&gt; interface.
        /// </summary>
        /// <param name="sltData">SLT data provider.</param>
        public SLTXML(ISLTData<T> sltData)
        {
            this.positiveNegativeSamplesXML = 
                new PositiveNegativeSamplesXML<T>(sltData);
            this.k = sltData.GetSLTK();
        }
        #endregion

        #region ISLTData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.positiveNegativeSamplesXML.GetAlphabetData(); }
        /// Returns special options for positive and negative samples.
        public PositiveNegativeSamplesOptions GetPNSamplesOptions() { return this.positiveNegativeSamplesXML.GetPNSamplesOptions(); }
        /// Returns data describing the positive samples.
        public SampleData<T>[] GetPositiveSamplesData() { return this.positiveNegativeSamplesXML.GetPositiveSamplesData(); }
        /// Returns data describing the negative samples.
        public SampleData<T>[] GetNegativeSamplesData() { return this.positiveNegativeSamplesXML.GetNegativeSamplesData(); }
        /// Returns a comment (a short description) of the positive and negative samples.
        public string GetPNSamplesCommentData() { return this.positiveNegativeSamplesXML.GetPNSamplesCommentData(); }
        /// Returns k for k-SLT languages.
        public int GetSLTK() { return this.k; }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the xml representation of the positive (and negative) set of samples.
        /// </summary>
        [XmlElement("PositiveNegativeSamples")]
        public PositiveNegativeSamplesXML<T> PositiveNegativeSamples
        {
            get { return this.positiveNegativeSamplesXML; }
            set { this.positiveNegativeSamplesXML = value; }
        }

        /// <summary>
        /// Gets or sets k for k-SLT languages.
        /// </summary>
        [XmlElement("K")]
        public int K
        {
            get { return this.k; }
            set { this.k = value; }
        }
        #endregion

        #region Private Data Members
        private PositiveNegativeSamplesXML<T> positiveNegativeSamplesXML;
        private int k;
        #endregion
    }

    /// <summary>
    /// Root class of every class that is an XML representation of a language.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [Serializable]
    public class ILanguageXML<T> :
        ILanguageData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {

    }

    /// <summary>
    /// Class that represents an XML representation of language based on DFA.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("LanguageDFA")]
    public class LanguageDFAXML<T> :
        ILanguageXML<T>,
        ILanguageDFAData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        public LanguageDFAXML()
        {
            this.dfaXML = null;
        }

        /// <summary>
        /// Initializes a new instance of the LanguageDFAXML&lt;T&gt; class
        /// based on the language data provider which implements
        /// ILanguageDFAData&lt;T&gt; interface.
        /// </summary>
        /// <param name="languageDFAData">Language data provider.</param>
        public LanguageDFAXML(ILanguageDFAData<T> languageDFAData)
        {
            this.dfaXML = new DFAXML<T>(languageDFAData);
        }
        #endregion

        #region ILanguageDFAData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.dfaXML.GetAlphabetData(); }
        /// Returns the number of states of the DFA.
        public int GetDFAStateCountData() { return this.dfaXML.GetDFAStateCountData(); }
        /// Returns the string that consist of the indices of the accepted states
        /// separated by a single space.
        public string GetDFAAcceptStatesData() { return this.dfaXML.GetDFAAcceptStatesData(); }
        /// Returns the content of the transition table as a string of numbers
        /// separated by a single space.
        public string GetDFATransitionTableData() { return this.dfaXML.GetDFATransitionTableData(); }
        /// Returns a comment (a short description) of the dfa.
        public string GetDFACommentData() { return this.dfaXML.GetDFACommentData(); }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the XML representation of the dfa.
        /// </summary>
        [XmlElement("DFA")]
        public DFAXML<T> DFA
        {
            get { return this.dfaXML; }
            set { this.dfaXML = value; }
        }
        #endregion

        #region Private Data Members
        private DFAXML<T> dfaXML;
        #endregion
    }

    /// <summary>
    /// Class that represents an XML representation of language based on DFA 
    /// created by LStar (Dana Angluin's L*) algorithm.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("LanguageDFALStar")]
    public sealed class LanguageDFALStarXML<T> :
        LanguageDFAXML<T>,
        ILanguageDFALStarData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        public LanguageDFALStarXML()
        {
            this.knowledgeBaseXML = null;
        }

        /// <summary>
        /// Initializes a new instance of the LanguageDFALStarXML&lt;T&gt; class
        /// based on the language data provider which implements
        /// ILanguageDFALStarData&lt;T&gt; interface.
        /// </summary>
        /// <param name="languageDFALStarData">Language data provider.</param>
        public LanguageDFALStarXML(ILanguageDFALStarData<T> languageDFALStarData) :
            base(languageDFALStarData)
        {
            this.knowledgeBaseXML = new KnowledgeBaseXML<T>(languageDFALStarData);
        }
        #endregion

        #region ILanguageDFALStarData Interface
        /// Returns an array of WordAcceptData&lt;T&gt; structures describing the whole knowledge base.
        public WordAcceptData<T>[] GetKnowledgeBaseData() { return this.knowledgeBaseXML.GetKnowledgeBaseData(); }
        /// Returns a comment (a short description) of the knowledge base.
        public string GetKnowledgeBaseCommentData() { return this.knowledgeBaseXML.GetKnowledgeBaseCommentData(); }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the xml representation of the knowledge base.
        /// </summary>
        [XmlElement("KnowledgeBase")]
        public KnowledgeBaseXML<T> KnowledgeBase
        {
            get { return this.knowledgeBaseXML; }
            set { this.knowledgeBaseXML = value; }
        }
        #endregion

        #region Private Data Members
        private KnowledgeBaseXML<T> knowledgeBaseXML;
        #endregion
    }

    /// <summary>
    /// Class that represents an XML representation of language based on DFA 
    /// created by RPNI algorithm.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("LanguageDFARPNI")]
    public sealed class LanguageDFARPNIXML<T> :
        LanguageDFAXML<T>,
        ILanguageDFARPNIData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        public LanguageDFARPNIXML()
        {
            this.positiveNegativeSamplesXML = null;
        }

        /// <summary>
        /// Initializes a new instance of the LanguageDFARPNIXML&lt;T&gt; class
        /// based on the language data provider which implements
        /// ILanguageDFARPNIData&lt;T&gt; interface.
        /// </summary>
        /// <param name="languageDFARPNIData">Language data provider.</param>
        public LanguageDFARPNIXML(ILanguageDFARPNIData<T> languageDFARPNIData) :
            base(languageDFARPNIData)
        {
            this.positiveNegativeSamplesXML =
                new PositiveNegativeSamplesXML<T>(languageDFARPNIData);
        }
        #endregion

        #region ILanguageDFARPNIData Interface
        /// Returns special options for positive and negative samples.
        public PositiveNegativeSamplesOptions GetPNSamplesOptions() { return this.positiveNegativeSamplesXML.GetPNSamplesOptions(); }
        /// Returns an array of SampleData&lt;T&gt; structures describing the set of positive samples.
        public SampleData<T>[] GetPositiveSamplesData() { return this.positiveNegativeSamplesXML.GetPositiveSamplesData(); }
        /// Returns an array of SampleData&lt;T&gt; structures describing the set of negative samples.
        public SampleData<T>[] GetNegativeSamplesData() { return this.positiveNegativeSamplesXML.GetNegativeSamplesData(); }
        /// Returns a comment (a short description) of the positive and negative samples.
        public string GetPNSamplesCommentData() { return this.positiveNegativeSamplesXML.GetPNSamplesCommentData(); }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the xml representation of the positive and negative set of samples.
        /// </summary>
        [XmlElement("PositiveNegativeSamples")]
        public PositiveNegativeSamplesXML<T> PositiveNegativeSamples
        {
            get { return this.positiveNegativeSamplesXML; }
            set { this.positiveNegativeSamplesXML = value; }
        }
        #endregion

        #region Private Data Members
        private PositiveNegativeSamplesXML<T> positiveNegativeSamplesXML;
        #endregion
    }

    /// <summary>
    /// Class that represents an XML representation of language based on regular expression.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("LanguageRegex")]
    public class LanguageRegexXML<T> :
        ILanguageXML<T>,
        ILanguageRegexData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        public LanguageRegexXML()
        {
            this.alphabetXML = null;
            this.regexPattern = null;
            this.CheckConsistency();
        }

        /// <summary>
        /// Initializes a new instance of the LanguageRegexXML&lt;T&gt; class
        /// based on the language data provider which implements
        /// ILanguageRegexData&lt;T&gt; interface.
        /// </summary>
        /// <param name="languageRegexData">Language data provider.</param>
        public LanguageRegexXML(ILanguageRegexData<T> languageRegexData)
        {
            this.alphabetXML = new AlphabetXML<T>(languageRegexData);
            this.regexPattern = languageRegexData.GetRegexData();
            this.CheckConsistency();
        }
        #endregion

        #region ILanguageRegexData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.alphabetXML.GetAlphabetData(); }
        /// Returns the string pattern of the regular expression.
        public string GetRegexData() { return this.regexPattern; }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the XML representation of the alphabet.
        /// </summary>
        [XmlElement("Alphabet")]
        public AlphabetXML<T> Alphabet
        {
            get { return this.alphabetXML; }
            set { this.alphabetXML = value; }
        }

        /// <summary>
        /// Gets or sets the XML representation of the regex pattern.
        /// </summary>
        [XmlElement("RegexPattern")]
        public string RegexPattern
        {
            get { return this.regexPattern; }
            set { this.regexPattern = value; }
        }
        #endregion

        #region Private Methods
        private void CheckConsistency()
        {
            if (typeof(T) != typeof(char))
                throw new InvalidProgramException("Only type char is supported.");
        }
        #endregion

        #region Private Data Members
        private AlphabetXML<T> alphabetXML;
        private string regexPattern;
        #endregion
    }

    /// <summary>
    /// Class that represents an XML representation of language based on SLT.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("LanguageSLT")]
    public class LanguageSLTXML<T> :
        ILanguageXML<T>,
        ILanguageSLTData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        public LanguageSLTXML()
        {
            this.sltXML = null;
        }

        /// <summary>
        /// Initializes a new instance of the LanguageSLTXML&lt;T&gt; class
        /// based on the language data provider which implements
        /// ILanguageSLTData&lt;T&gt; interface.
        /// </summary>
        /// <param name="languageSLTData">Language data provider.</param>
        public LanguageSLTXML(ILanguageSLTData<T> languageSLTData)
        {
            this.sltXML = new SLTXML<T>(languageSLTData);
        }
        #endregion

        #region ILanguageSLTData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.sltXML.GetAlphabetData(); }
        /// Returns special options for positive and negative samples.
        public PositiveNegativeSamplesOptions GetPNSamplesOptions() { return this.sltXML.GetPNSamplesOptions(); }
        /// Returns data describing the positive samples.
        public SampleData<T>[] GetPositiveSamplesData() { return this.sltXML.GetPositiveSamplesData(); }
        /// Returns data describing the negative samples.
        public SampleData<T>[] GetNegativeSamplesData() { return this.sltXML.GetNegativeSamplesData(); }
        /// Returns a comment (a short description) of the positive and negative samples.
        public string GetPNSamplesCommentData() { return this.sltXML.GetPNSamplesCommentData(); }
        /// Returns k for k-SLT languages.
        public int GetSLTK() { return this.sltXML.GetSLTK(); }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the XML representation of the SLT.
        /// </summary>
        [XmlElement("SLT")]
        public SLTXML<T> SLT
        {
            get { return this.sltXML; }
            set { this.sltXML = value; }
        }
        #endregion

        #region Private Data Members
        private SLTXML<T> sltXML;
        #endregion
    }

    /// <summary>
    /// Class that represents an XML representation of the accepting metainstruction.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("AcceptingMetainstruction")]
    public sealed class AcceptingMetainstructionXML<T> :
        IAcceptingMetainstructionData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        public AcceptingMetainstructionXML()
        {
            this.alphabetXML = null;
            this.acceptingLanguageXML = null;
            this.name = "";
            this.CheckConsistency();
        }

        /// <summary>
        /// Initializes a new instance of the AcceptingMetainstructionXML&lt;T&gt; class
        /// based on the accepting metainstruction data provider which implements
        /// IAcceptingMetainstructionData&lt;T&gt; interface.
        /// </summary>
        /// <param name="acceptingMetainstructionData">Accepting metainstruction data provider.</param>
        public AcceptingMetainstructionXML(IAcceptingMetainstructionData<T> acceptingMetainstructionData)
        {
            this.alphabetXML = new AlphabetXML<T>(acceptingMetainstructionData);
            this.acceptingLanguageXML = LanguageXMLFactory<T>.GetLanguageXMLInstance(
                acceptingMetainstructionData.GetAcceptingLanguageData());
            this.name = acceptingMetainstructionData.GetAcceptingName();
            this.CheckConsistency();
        }
        #endregion

        #region IAcceptingMetainstructionData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.alphabetXML.GetAlphabetData(); }
        /// Returns an interface that provides all the neccessary data
        /// describing the accepting language of the accepting metainstruction.
        public ILanguageData<T> GetAcceptingLanguageData() { return this.acceptingLanguageXML; }
        /// Returns the name of the accepting metainstruction.
        public string GetAcceptingName() { return this.name; }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the name of the accepting metainstruction.
        /// </summary>
        public string Name
        {
            get { return this.name; }
            set { this.name = value; }
        }

        /// <summary>
        /// Gets or sets the XML representation of the alphabet.
        /// </summary>
        [XmlElement("Alphabet")]
        public AlphabetXML<T> Alphabet
        {
            get { return this.alphabetXML; }
            set { this.alphabetXML = value; }
        }

        /// <summary>
        /// Gets or sets the XML representation of the accepting language.
        /// </summary>
        [XmlElement("AcceptingLanguageDFA", typeof(LanguageDFAXML<char>)),
        XmlElement("AcceptingLanguageDFALStar", typeof(LanguageDFALStarXML<char>)),
        XmlElement("AcceptingLanguageDFARPNI", typeof(LanguageDFARPNIXML<char>)),
        XmlElement("AcceptingLanguageRegex", typeof(LanguageRegexXML<char>)),
        XmlElement("AcceptingLanguageSLT", typeof(LanguageSLTXML<char>))]
        public ILanguageXML<T> AcceptingLanguage
        {
            get { return this.acceptingLanguageXML; }
            set { this.acceptingLanguageXML = value; }
        }
        #endregion

        #region Private Methods
        private void CheckConsistency()
        {
            if (typeof(T) != typeof(char))
                throw new InvalidProgramException("Only type char is supported.");
        }
        #endregion

        #region Private Data Members
        private AlphabetXML<T> alphabetXML;
        private ILanguageXML<T> acceptingLanguageXML;
        string name;
        #endregion
    }

    /// <summary>
    /// Class that represents an XML representation of the reducing metainstruction.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("ReducingMetainstruction")]
    public sealed class ReducingMetainstructionXML<T> :
        IReducingMetainstructionData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        public ReducingMetainstructionXML()
        {
            this.alphabetXML = null;
            this.leftLanguageXML = null;
            this.rightLanguageXML = null;
            this.fromWordData = null;
            this.toWordData = null;
            this.name = "";
            this.CheckConsistency();
        }

        /// <summary>
        /// Initializes a new instance of the ReducingMetainstructionXML&lt;T&gt; class
        /// based on the reducing metainstruction data provider which implements
        /// IReducingMetainstructionData&lt;T&gt; interface.
        /// </summary>
        /// <param name="reducingMetainstructionData">Reducing metainstruction data provider.</param>
        public ReducingMetainstructionXML(IReducingMetainstructionData<T> reducingMetainstructionData)
        {
            this.alphabetXML = new AlphabetXML<T>(reducingMetainstructionData);
            this.leftLanguageXML = LanguageXMLFactory<T>.GetLanguageXMLInstance(
                reducingMetainstructionData.GetLeftLanguageData());
            this.rightLanguageXML = LanguageXMLFactory<T>.GetLanguageXMLInstance(
                reducingMetainstructionData.GetRightLanguageData());
            this.fromWordData = reducingMetainstructionData.GetFromWordData();
            this.toWordData = reducingMetainstructionData.GetToWordData();
            this.name = reducingMetainstructionData.GetReducingName();
            this.CheckConsistency();
        }
        #endregion

        #region IAcceptingMetainstructionData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.alphabetXML.GetAlphabetData(); }
        /// Returns an interface that provides all the neccessary data
        /// describing the left language of the reducing metainstruction.
        public ILanguageData<T> GetLeftLanguageData() { return this.leftLanguageXML; }
        /// Returns an interface that provides all the neccessary data
        /// describing the right language of the reducing metainstruction.
        public ILanguageData<T> GetRightLanguageData() { return this.rightLanguageXML; }
        /// Returns data describing the from-word of the reducing metainstruction.
        public T[] GetFromWordData() { return this.fromWordData.Clone() as T[]; }
        /// Returns data describing the to-word of the reducing metainstruction.
        public T[] GetToWordData() { return this.toWordData.Clone() as T[]; }
        /// Returns the name of the reducing metainstruction.
        public string GetReducingName() { return this.name; }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the name of the reducing metainstruction.
        /// </summary>
        public string Name
        {
            get { return this.name; }
            set { this.name = value; }
        }

        /// <summary>
        /// Gets or sets the XML representation of the alphabet.
        /// </summary>
        [XmlElement("Alphabet")]
        public AlphabetXML<T> Alphabet
        {
            get { return this.alphabetXML; }
            set { this.alphabetXML = value; }
        }

        /// <summary>
        /// Gets or sets the XML representation of the left language.
        /// </summary>
        [XmlElement("LeftLanguageDFA", typeof(LanguageDFAXML<char>)),
        XmlElement("LeftLanguageDFALStar", typeof(LanguageDFALStarXML<char>)),
        XmlElement("LeftLanguageDFARPNI", typeof(LanguageDFARPNIXML<char>)),
        XmlElement("LeftLanguageRegex", typeof(LanguageRegexXML<char>)),
        XmlElement("LeftLanguageSLT", typeof(LanguageSLTXML<char>))]
        public ILanguageXML<T> LeftLanguage
        {
            get { return this.leftLanguageXML; }
            set { this.leftLanguageXML = value; }
        }

        /// <summary>
        /// Gets or sets the XML representation of the right language.
        /// </summary>
        [XmlElement("RightLanguageDFA", typeof(LanguageDFAXML<char>)),
        XmlElement("RightLanguageDFALStar", typeof(LanguageDFALStarXML<char>)),
        XmlElement("RightLanguageDFARPNI", typeof(LanguageDFARPNIXML<char>)),
        XmlElement("RightLanguageRegex", typeof(LanguageRegexXML<char>)),
        XmlElement("RightLanguageSLT", typeof(LanguageSLTXML<char>))]
        public ILanguageXML<T> RightLanguage
        {
            get { return this.rightLanguageXML; }
            set { this.rightLanguageXML = value; }
        }

        /// <summary>
        /// Gets or sets the XML representation of the from-word.
        /// </summary>
        [XmlArray("FromWord"), XmlArrayItem("T")]
        public T[] FromWord
        {
            get { return this.fromWordData; }
            set { this.fromWordData = value; }
        }

        /// <summary>
        /// Gets or sets the XML representation of the to-word.
        /// </summary>
        [XmlArray("ToWord"), XmlArrayItem("T")]
        public T[] ToWord
        {
            get { return this.toWordData; }
            set { this.toWordData = value; }
        }
        #endregion

        #region Private Methods
        private void CheckConsistency()
        {
            if (typeof(T) != typeof(char))
                throw new InvalidProgramException("Only type char is supported.");
        }
        #endregion

        #region Private Data Members
        private AlphabetXML<T> alphabetXML;
        private ILanguageXML<T> leftLanguageXML;
        private ILanguageXML<T> rightLanguageXML;
        private T[] fromWordData;
        private T[] toWordData;
        string name;
        #endregion
    }

    /// <summary>
    /// Class that represents an XML representation of the restarting automaton.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    [XmlRoot("RestartingAutomaton")]
    public sealed class RestartingAutomatonXML<T> :
        IRestartingAutomatonData<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        public RestartingAutomatonXML()
        {
            this.alphabetXML = null;
            this.acceptingMetainstructionsXML = null;
            this.reducingMetainstructionsXML = null;
        }

        /// <summary>
        /// Initializes a new instance of the RestartingAutomatonXML&lt;T&gt; class
        /// based on the restarting automaton data provider which implements
        /// IRestartingAutomatonData&lt;T&gt; interface.
        /// </summary>
        /// <param name="restartingAutomatonData">Restarting automaton data provider.</param>
        public RestartingAutomatonXML(IRestartingAutomatonData<T> restartingAutomatonData)
        {
            this.alphabetXML = new AlphabetXML<T>(restartingAutomatonData);
            IAcceptingMetainstructionData<T>[] acceptingMetainstructionsData =
                restartingAutomatonData.GetAcceptingMetainstructionsData();
            IReducingMetainstructionData<T>[] reducingMetainstructionsData =
                restartingAutomatonData.GetReducingMetainstructionsData();
            List<AcceptingMetainstructionXML<T>> acceptingMetainstructionXMLList =
                new List<AcceptingMetainstructionXML<T>>(acceptingMetainstructionsData.Length);
            List <ReducingMetainstructionXML<T>> reducingMetainstructionXMLList =
                new List<ReducingMetainstructionXML<T>>(reducingMetainstructionsData.Length);
            foreach (IAcceptingMetainstructionData<T> acceptingMetainstructionData in acceptingMetainstructionsData)
                acceptingMetainstructionXMLList.Add(new AcceptingMetainstructionXML<T>(acceptingMetainstructionData));
            foreach (IReducingMetainstructionData<T> reducingMetainstructionData in reducingMetainstructionsData)
                reducingMetainstructionXMLList.Add(new ReducingMetainstructionXML<T>(reducingMetainstructionData));
            this.acceptingMetainstructionsXML = acceptingMetainstructionXMLList.ToArray();
            this.reducingMetainstructionsXML = reducingMetainstructionXMLList.ToArray();
        }
        #endregion

        #region IRestartingAutomatonData Interface
        /// Returns data describing the alphabet.
        public T[] GetAlphabetData() { return this.alphabetXML.GetAlphabetData(); }
        /// Returns an array of IAcceptingMetainstructionData&lt;T&gt; interfaces
        /// describing the set of accepting metainstructions.
        public IAcceptingMetainstructionData<T>[] GetAcceptingMetainstructionsData()
        {
            List<IAcceptingMetainstructionData<T>> acceptingMetainstructionDataList =
                new List<IAcceptingMetainstructionData<T>>(this.acceptingMetainstructionsXML.Length);
            foreach (AcceptingMetainstructionXML<T> acceptingMetainstructionXML in this.acceptingMetainstructionsXML)
                acceptingMetainstructionDataList.Add(acceptingMetainstructionXML);
            return acceptingMetainstructionDataList.ToArray();
        }
        /// Returns an array of IReducingMetainstructionData&lt;T&gt; interfaces
        /// describing the set of reducing metainstructions.
        public IReducingMetainstructionData<T>[] GetReducingMetainstructionsData()
        {
            List<IReducingMetainstructionData<T>> reducingMetainstructionDataList =
                new List<IReducingMetainstructionData<T>>(this.reducingMetainstructionsXML.Length);
            foreach (ReducingMetainstructionXML<T> reducingMetainstructionXML in this.reducingMetainstructionsXML)
                reducingMetainstructionDataList.Add(reducingMetainstructionXML);
            return reducingMetainstructionDataList.ToArray();
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the XML representation of the alphabet.
        /// </summary>
        [XmlElement("Alphabet")]
        public AlphabetXML<T> Alphabet
        {
            get { return this.alphabetXML; }
            set { this.alphabetXML = value; }
        }

        /// <summary>
        /// Gets or sets the XML representation of the accepting metainstructions.
        /// </summary>
        [XmlArray("AcceptingMetainstructions"), XmlArrayItem("AcceptingMetainstruction")]
        public AcceptingMetainstructionXML<T>[] AcceptingMetainstructions
        {
            get { return this.acceptingMetainstructionsXML; }
            set { this.acceptingMetainstructionsXML = value; }
        }

        /// <summary>
        /// Gets or sets the XML representation of the reducing metainstructions.
        /// </summary>
        [XmlArray("ReducingMetainstructions"), XmlArrayItem("ReducingMetainstruction")]
        public ReducingMetainstructionXML<T>[] ReducingMetainstructions
        {
            get { return this.reducingMetainstructionsXML; }
            set { this.reducingMetainstructionsXML = value; }
        }
        #endregion

        #region Private Data Members
        private AlphabetXML<T> alphabetXML;
        private AcceptingMetainstructionXML<T>[] acceptingMetainstructionsXML;
        private ReducingMetainstructionXML<T>[] reducingMetainstructionsXML;
        #endregion
    }

    /// <summary>
    /// Class that represents a static XML language factory.
    /// </summary>
    /// <typeparam name="T">Type that represents a token (for instance a char or int).</typeparam>
    public static class LanguageXMLFactory<T>
        where T : struct, IComparable<T>, IEquatable<T>
    {
        /// <summary>
        /// Returns the instance of the XML language class that corresponds to the given
        /// ILanguageData&lt;T&gt; interface.
        /// </summary>
        public static ILanguageXML<T> GetLanguageXMLInstance(
            ILanguageData<T> languageData)
        {
            if (languageData == null)
                return null;
            else if (languageData is ILanguageDFALStarData<T>)
                return new LanguageDFALStarXML<T>(languageData as ILanguageDFALStarData<T>);
            else if (languageData is ILanguageDFARPNIData<T>)
                return new LanguageDFARPNIXML<T>(languageData as ILanguageDFARPNIData<T>);
            else if (languageData is ILanguageDFAData<T>)
                return new LanguageDFAXML<T>(languageData as ILanguageDFAData<T>);
            else if (languageData is ILanguageRegexData<T>)
                return new LanguageRegexXML<T>(languageData as ILanguageRegexData<T>);
            else if (languageData is ILanguageSLTData<T>)
                return new LanguageSLTXML<T>(languageData as ILanguageSLTData<T>);
            else throw new InvalidProgramException("LanguageXMLFactory: Unrecognized ILanguageData interface.");
        }
    }
}