using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Xml.Serialization;
using System.IO;
using SimpleTextParser;

namespace RegularLanguages
{
    public partial class SLTForm : Form, LanguageForm<char>
    {
        public SLTForm()
        {
            InitializeComponent();
            this.lockAlphabet = false;
            this.alphabet = null;
            this.slt = null;
            this.UpdateUI();
        }

        public SLTForm(Alphabet<char> alphabet, int k)
        {
            InitializeComponent();
            this.lockAlphabet = true;
            this.alphabet = alphabet;
            this.slt = new SLT<char>(alphabet, k);
            this.UpdateUI();
        }

        public SLTForm(SLT<char> slt)
        {
            InitializeComponent();
            this.lockAlphabet = true;
            this.alphabet = slt.Alphabet;
            this.slt = slt;
            this.DataToUI();
            this.UpdateUI();
        }

        public ILanguage<char> Language
        {
            get
            {
                if (this.UIToData())
                    return new LanguageSLT<char>(this.slt);
                else return null;
            }
        }

        private string RegularToSLT(string regularWord)
        {
            if (regularWord == "") return "''";
            else return regularWord;
        }

        private void ClearUI()
        {
            this.acceptTextBox.Clear();
            this.commentTextBox.Clear();
        }

        private void UpdateUI()
        {
            if (this.lockAlphabet)
                this.alphabetButton.Enabled = false;
            else
            {
                this.alphabetButton.Enabled = true;
                this.okButton.Enabled = false;
                this.cancelButton.Enabled = false;
            }
            if (alphabet == null)
            {
                this.clearButton.Enabled = false;
                this.testButton.Enabled = false;
                this.acceptTextBox.Clear();
                this.acceptTextBox.Enabled = false;
                this.saveButton.Enabled = false;
                this.commentTextBox.Enabled = false;
                this.copyButton.Enabled = false;
                this.pasteButton.Enabled = true;
                this.viewButton.Enabled = false;
            }
            else
            {
                this.clearButton.Enabled = true;
                this.testButton.Enabled = true;
                this.acceptTextBox.Enabled = true;
                this.saveButton.Enabled = true;
                this.commentTextBox.Enabled = true;
                this.copyButton.Enabled = true;
                this.pasteButton.Enabled = true;
                this.viewButton.Enabled = true;
            }
        }

        private bool lockAlphabet;
        private Alphabet<char> alphabet;
        private SLT<char> slt;
        private TextParser textParser = new TextParser();
        private bool canClose = false;

        private bool LoadAlphabet(IAlphabetData<char> alphabetData)
        {
            if (this.alphabet != null)
            {
                Alphabet<char> alphabet_cmp =
                    new Alphabet<char>(alphabetData);
                if (!this.alphabet.Equals(alphabet_cmp))
                {
                    MessageBox.Show("The SLT has an inconsistent alphabet.", "Error");
                    return false;
                }
            }
            else
            {
                this.alphabet = new Alphabet<char>(alphabetData);
            }
            return true;
        }

        private void LoadSLTData(ISLTData<char> sltData)
        {
            if (!LoadAlphabet(sltData)) return;
            this.slt = new SLT<char>(sltData);
            this.DataToUI();
            this.UpdateUI();
        }

        private void LoadPositiveNegativeSamplesData(
            IPositiveNegativeSamplesData<char> positiveNegativeSamplesData)
        {
            if (!LoadAlphabet(positiveNegativeSamplesData)) return;
            if (this.slt == null) this.slt = new SLT<char>(this.alphabet, 1);
            this.slt.Samples = new PositiveNegativeSamples<char>(
                positiveNegativeSamplesData, PositiveNegativeSamplesOptions.PositiveSamplesEnabled);
            this.DataToUI();
            this.UpdateUI();
        }

        private void LoadKnowledgeBaseData(
            IKnowledgeBaseData<char> knowledgeBaseData)
        {
            if (!LoadAlphabet(knowledgeBaseData)) return;
            if (this.slt == null) this.slt = new SLT<char>(this.alphabet, 1);
            this.slt.Samples = new PositiveNegativeSamples<char>(
                knowledgeBaseData, PositiveNegativeSamplesOptions.PositiveSamplesEnabled);
            this.DataToUI();
            this.UpdateUI();
        }

        private void DataToUI()
        {
            this.kTextBox.Text = this.slt.K.ToString();
            this.acceptTextBox.Clear();
            List<string> positiveSampleList = new List<string>();
            foreach (GenericString<char> positiveSample in this.slt.Samples.PositiveSamples)
                positiveSampleList.Add(RegularToSLT(positiveSample.ToString()));
            this.acceptTextBox.Text = String.Join(" ", positiveSampleList.ToArray());
            this.commentTextBox.Text = this.slt.Samples.Comment;
        }

        private bool UIToData()
        {
            if (kTextBox.IntValue <= 0)
            {
                MessageBox.Show("Please set the k value to a positive number.", "Warning");
                return false;
            }
            this.slt.Samples.Clear();
            this.slt.K = kTextBox.IntValue;
            TextParserStruct[] acceptStructs =
                this.textParser.Parse(acceptTextBox.Text);
            int acceptCount = 0;
            if (acceptStructs != null)
                acceptCount = acceptStructs.Length;
            if (acceptStructs != null)
                foreach (TextParserStruct acceptStruct in acceptStructs)
                {
                    if (!this.slt.Alphabet.Contains(acceptStruct.word))
                    {
                        MessageBox.Show("All the letters of the word must be from the alphabet.", "Warning");
                        return false;
                    }
                    slt.Samples.AddPositiveSample(new GenericString<char>(acceptStruct.word));
                }
            this.slt.Samples.Comment = this.commentTextBox.Text;
            return true;
        }

        private void alphabetButton_Click(object sender, EventArgs e)
        {
            if (kTextBox.IntValue <= 0)
            {
                MessageBox.Show("Please set the k value to a positive number.", "Warning");
                return;
            }
            AlphabetForm alphabetForm = new AlphabetForm();
            DialogResult dialogResult = alphabetForm.ShowDialog();
            if (dialogResult == DialogResult.OK)
            {
                if (alphabetForm.Alphabet == "")
                {
                    this.alphabet = null;
                    this.slt = null;
                }
                else
                {
                    this.alphabet = new Alphabet<char>(alphabetForm.Alphabet);
                    this.slt = new SLT<char>(alphabet, kTextBox.IntValue);
                }
                this.ClearUI();
                this.UpdateUI();
            }
        }

        private void clearButton_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show("Do you really want to clear the SLT?",
                "Confirmation", MessageBoxButtons.YesNo) == DialogResult.No)
                return;
            this.ClearUI();
            this.UpdateUI();
        }

        private void testButton_Click(object sender, EventArgs e)
        {
            if (this.UIToData())
            {
                SLTPreviewForm sltPreviewForm = new SLTPreviewForm(this.slt);
                sltPreviewForm.ShowDialog();
            }
        }

        private void saveButton_Click(object sender, EventArgs e)
        {
            if (this.UIToData())
            {
                SaveFileDialog saveFileDialog = new SaveFileDialog();
                saveFileDialog.Filter = "XML files (*.xml)|*.xml|All files (*.*)|*.*";
                DialogResult saveFileDialogResult = saveFileDialog.ShowDialog();
                if (saveFileDialogResult == DialogResult.OK)
                {
                    XmlSerializer sltSerializer =
                        new XmlSerializer(typeof(SLTXML<char>));
                    FileStream fileSaveStream =
                        new FileStream(saveFileDialog.FileName, FileMode.Create);
                    SLTXML<char> sltXML =
                        new SLTXML<char>(this.slt);
                    sltSerializer.Serialize(
                        fileSaveStream, sltXML);
                    fileSaveStream.Close();
                }
            }
        }

        private void loadButton_Click(object sender, EventArgs e)
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();
            openFileDialog.Filter = "XML files (*.xml)|*.xml|All files (*.*)|*.*";
            DialogResult openFileDialogResult = openFileDialog.ShowDialog();
            if (openFileDialogResult == DialogResult.OK)
            {
                // SLT
                FileStream fileLoadStream = new FileStream(openFileDialog.FileName, FileMode.Open);
                try
                {
                    XmlSerializer sltSerializer =
                        new XmlSerializer(typeof(SLTXML<char>));
                    SLTXML<char> sltXML =
                        sltSerializer.Deserialize(fileLoadStream) as SLTXML<char>;
                    this.LoadSLTData(sltXML);
                    return;
                }
                catch { }
                finally { fileLoadStream.Close(); }
                // Positive and negative samples
                fileLoadStream = new FileStream(openFileDialog.FileName, FileMode.Open);
                try
                {
                    XmlSerializer positiveNegativeSamplesSerializer =
                        new XmlSerializer(typeof(PositiveNegativeSamplesXML<char>));
                    PositiveNegativeSamplesXML<char> positiveNegativeSamplesXML =
                        positiveNegativeSamplesSerializer.Deserialize(fileLoadStream)
                            as PositiveNegativeSamplesXML<char>;
                    this.LoadPositiveNegativeSamplesData(positiveNegativeSamplesXML);
                    return;
                }
                catch { }
                finally { fileLoadStream.Close(); }
                // Knowledge base
                fileLoadStream = new FileStream(openFileDialog.FileName, FileMode.Open);
                try
                {
                    XmlSerializer knowledgeBaseSerializer =
                        new XmlSerializer(typeof(KnowledgeBaseXML<char>));
                    KnowledgeBaseXML<char> knowledgeBaseXML =
                        knowledgeBaseSerializer.Deserialize(fileLoadStream) as KnowledgeBaseXML<char>;
                    this.LoadKnowledgeBaseData(knowledgeBaseXML);
                    return;
                }
                catch { }
                finally { fileLoadStream.Close(); }
                MessageBox.Show("Unable to deserialize the SLT.", "Error");
            }
        }

        private void copyButton_Click(object sender, EventArgs e)
        {
            if (this.UIToData())
            {
                XmlSerializer sltSerializer = 
                    new XmlSerializer(typeof(SLTXML<char>));
                StringWriter stringWriter = new StringWriter();
                SLTXML<char> sltXML = new SLTXML<char>(this.slt);
                sltSerializer.Serialize(stringWriter, sltXML);
                Clipboard.SetText(stringWriter.ToString());
                stringWriter.Close();
            }

        }

        private void pasteButton_Click(object sender, EventArgs e)
        {
            if (Clipboard.ContainsText())
            {
                string confirmation = "Do you really want to replace the current SLT?";
                // SLT
                StringReader stringReader = new StringReader(Clipboard.GetText()); 
                try
                {
                    XmlSerializer sltSerializer =
                        new XmlSerializer(typeof(SLTXML<char>));
                    SLTXML<char> sltXML =
                        sltSerializer.Deserialize(stringReader) as SLTXML<char>;
                    if (MessageBox.Show(confirmation, "Confirmation",
                            MessageBoxButtons.YesNo) == DialogResult.No)
                        return;
                    this.LoadSLTData(sltXML);
                    return;
                }
                catch { }
                finally { stringReader.Close(); }
                // Positive and negative samples
                stringReader = new StringReader(Clipboard.GetText());
                try
                {
                    XmlSerializer positiveNegativeSamplesSerializer =
                        new XmlSerializer(typeof(PositiveNegativeSamplesXML<char>));
                    PositiveNegativeSamplesXML<char> positiveNegativeSamplesXML =
                        positiveNegativeSamplesSerializer.Deserialize(stringReader)
                            as PositiveNegativeSamplesXML<char>;
                    if (MessageBox.Show(confirmation, "Confirmation",
                            MessageBoxButtons.YesNo) == DialogResult.No)
                        return;
                    this.LoadPositiveNegativeSamplesData(positiveNegativeSamplesXML);
                    return;
                }
                catch { }
                finally { stringReader.Close(); }
                // Knowledge base
                stringReader = new StringReader(Clipboard.GetText());
                try
                {
                    XmlSerializer knowledgeBaseSerializer =
                        new XmlSerializer(typeof(KnowledgeBaseXML<char>));
                    KnowledgeBaseXML<char> knowledgeBaseXML =
                        knowledgeBaseSerializer.Deserialize(stringReader) as KnowledgeBaseXML<char>;
                    if (MessageBox.Show(confirmation, "Confirmation",
                            MessageBoxButtons.YesNo) == DialogResult.No)
                        return;
                    this.LoadKnowledgeBaseData(knowledgeBaseXML);
                    return;
                }
                catch { }
                finally { stringReader.Close(); }
                MessageBox.Show("Unable to deserialize the SLT.", "Error");
            }
        }

        private void viewButton_Click(object sender, EventArgs e)
        {
            XmlSerializer sltSerializer =
                new XmlSerializer(typeof(SLTXML<char>));
            StringWriter stringWriter = new StringWriter();
            SLTXML<char> sltXML = new SLTXML<char>(this.slt);
            sltSerializer.Serialize(stringWriter, sltXML);
            XMLPreviewForm xmlPreviewForm = new XMLPreviewForm(stringWriter.ToString());
            stringWriter.Close();
            xmlPreviewForm.ShowDialog();
        }

        private void okButton_Click(object sender, EventArgs e)
        {
            this.canClose = this.UIToData();
        }

        private void cancelButton_Click(object sender, EventArgs e)
        {
            this.canClose = true;
        }

        private void SLTForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (e.CloseReason == CloseReason.None)
                e.Cancel = !this.canClose;
            else e.Cancel = false;
        }
    }
}