using System;
using System.Collections.Generic;
using System.Text;

namespace SimpleTextParser
{
    /// <summary>
    /// Basic structure for parsing the text. It holds all
    /// the necessary information about one parsed word.
    /// </summary>
    struct TextParserStruct
    {
        public string word;         /// Parsed word.
        public int begin;           /// Starting index of the parsed word in the text.
        public int end;             /// Index of the first char behind the parsed word.
        public bool singleQuoted;   /// Indicates whether the parsed word was enclosed in the single quotes ' ... '.
        public bool doubleQuoted;   /// Indicates whether the parsed word was enclosed in the double quotes " ... ".
        public bool quoted;         /// Indicates whether the parsed word was enclosed in the quotes.
    }

    /// <summary>
    /// Class TextParser encapsulates the parsing algorithm.
    /// </summary>
    class TextParser
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of the TextParser class.
        /// </summary>
        public TextParser()
        {
            includeSingleQuotes = false;
            includeDoubleQuotes = false;
        }
        #endregion

        #region Methods
        /// <summary>
        /// Returns the array of TextParserStruct structures
        /// that represent the parsed text.
        /// </summary>
        /// <param name="text">Input text to be parsed.</param>
        public TextParserStruct[] Parse(string text)
        {
            if (text == null)
                return null;
            length = text.Length;
            index = 0;
            if (length == 0)
                return null;
            currentText = text;
            currentChar = text[0];
            acc_word = new StringBuilder(minimal_capacity);
            acc_begin = 0;
            acc_end = 0;
            singleQuoted = false;
            doubleQuoted = false;
            list = new List<TextParserStruct>();
            state = ParseState.Start;
            while (index < length)
            {
                switch (state)
                {
                    case ParseState.Start:
                        if (IsIdentifierLetter(currentChar))
                        {
                            StartToken(ParseState.Text);
                        }
                        else if (IsSingleQuote(currentChar))
                        {
                            StartQuoteToken(ParseState.SingleQuote);
                        }
                        else if (IsDoubleQuote(currentChar))
                        {
                            StartQuoteToken(ParseState.DoubleQuote);
                        }
                        else if (IsWhiteSpace(currentChar))
                        {
                            StartBlankToken(ParseState.WhiteSpace);
                        }
                        else
                        {
                            StartToken(ParseState.SpecialSymbol);
                        }
                        break;
                    case ParseState.WhiteSpace:
                        if (IsIdentifierLetter(currentChar))
                        {
                            StartToken(ParseState.Text);
                        }
                        else if (IsSingleQuote(currentChar))
                        {
                            StartQuoteToken(ParseState.SingleQuote);
                        }
                        else if (IsDoubleQuote(currentChar))
                        {
                            StartQuoteToken(ParseState.DoubleQuote);
                        }
                        else if (IsWhiteSpace(currentChar))
                        {
                            break;
                        }
                        else
                        {
                            StartToken(ParseState.SpecialSymbol);
                        }
                        break;
                    case ParseState.Text:
                        if (IsIdentifierLetter(currentChar))
                        {
                            acc_word.Append(currentChar);
                        }
                        else if (IsSingleQuote(currentChar))
                        {
                            acc_word.Append('\'');
                        }
                        else if (IsDoubleQuote(currentChar))
                        {
                            acc_word.Append('"');
                        }
                        else if (IsWhiteSpace(currentChar))
                        {
                            AddTokenToList();
                            StartBlankToken(ParseState.WhiteSpace);
                        }
                        else
                        {
                            AddTokenToList();
                            StartToken(ParseState.SpecialSymbol);
                        }
                        break;
                    case ParseState.SingleQuote:
                        if (IsSingleQuote(currentChar))
                        {
                            if (index + 1 < length &&
                                IsSingleQuote(text[index + 1]))
                            {
                                acc_word.Append('\'');
                                MoveIndex();
                            }
                            else
                            {
                                if (includeSingleQuotes)
                                    acc_word.Append('\'');
                                AddTokenToList_CurrentIncluded();
                                StartBlankToken(ParseState.Start);
                            }
                        }
                        else
                        {
                            acc_word.Append(currentChar);
                        }
                        break;
                    case ParseState.DoubleQuote:
                        if (IsDoubleQuote(currentChar))
                        {
                            if (index + 1 < length &&
                                IsDoubleQuote(text[index + 1]))
                            {
                                acc_word.Append('"');
                                MoveIndex();
                            }
                            else
                            {
                                if (includeDoubleQuotes)
                                    acc_word.Append('"');
                                AddTokenToList_CurrentIncluded();
                                StartBlankToken(ParseState.Start);
                            }
                        }
                        else
                        {
                            acc_word.Append(currentChar);
                        }
                        break;
                    case ParseState.SpecialSymbol:
                        if (IsIdentifierLetter(currentChar))
                        {
                            AddTokenToList();
                            StartToken(ParseState.Text);
                        }
                        else if (IsSingleQuote(currentChar))
                        {
                            AddTokenToList();
                            StartQuoteToken(ParseState.SingleQuote);
                        }
                        else if (IsDoubleQuote(currentChar))
                        {
                            AddTokenToList();
                            StartQuoteToken(ParseState.DoubleQuote);
                        }
                        else if (IsWhiteSpace(currentChar))
                        {
                            AddTokenToList();
                            StartBlankToken(ParseState.WhiteSpace);
                        }
                        else
                        {
                            string acc_word_string = acc_word.ToString();
                            if ((acc_word_string == "/" && currentChar == '/') ||
                                (acc_word_string == "/" && currentChar == '*') ||
                                (acc_word_string == "*" && currentChar == '/') ||
                                (acc_word_string == "*" && currentChar == '*') ||
                                (acc_word_string == "+" && currentChar == '+') ||
                                (acc_word_string == "-" && currentChar == '-') ||
                                (acc_word_string == ":" && currentChar == ':') ||
                                (acc_word_string == ":" && currentChar == '=') ||
                                (acc_word_string == "-" && currentChar == '>') ||
                                (acc_word_string == ">" && currentChar == '=') ||
                                (acc_word_string == "<" && currentChar == '=') ||
                                (acc_word_string == "=" && currentChar == '=') ||
                                (acc_word_string == "<" && currentChar == '>') ||
                                (acc_word_string == "=" && currentChar == '>') ||
                                (acc_word_string == "=" && currentChar == '<') ||
                                (acc_word_string == "!" && currentChar == '=') ||
                                (acc_word_string == "+" && currentChar == '=') ||
                                (acc_word_string == "-" && currentChar == '=') ||
                                (acc_word_string == "*" && currentChar == '=') ||
                                (acc_word_string == "/" && currentChar == '=') ||
                                (acc_word_string == "%" && currentChar == '=') ||
                                (acc_word_string == "&" && currentChar == '=') ||
                                (acc_word_string == "|" && currentChar == '=') ||
                                (acc_word_string == "|" && currentChar == '|') ||
                                (acc_word_string == "&" && currentChar == '&') ||
                                (acc_word_string == "." && currentChar == '.'))
                                acc_word.Append(currentChar);
                            else
                            {
                                AddTokenToList();
                                StartToken(ParseState.SpecialSymbol);
                            }
                        }
                        break;
                    default:
                        break;
                }
                MoveIndex();
            }
            switch (state)
            {
                case ParseState.Start:
                case ParseState.WhiteSpace:
                    return list.ToArray();
                case ParseState.Text:
                case ParseState.SpecialSymbol:
                    AddTokenToList();
                    return list.ToArray();
                case ParseState.SingleQuote:
                    if (includeSingleQuotes)
                        acc_word.Append('\'');
                    AddTokenToList();
                    return list.ToArray();
                case ParseState.DoubleQuote:
                    if (includeDoubleQuotes)
                        acc_word.Append('"');
                    AddTokenToList();
                    return list.ToArray();
                default:
                    return null;
            }
        }
        #endregion

        #region Public Data Members
        public bool includeSingleQuotes; /// Indicates whether the single quotes should be included in the parsed word.
        public bool includeDoubleQuotes; /// Indicates whether the double quotes should be included in the parsed word.
        #endregion

        #region Private Data Members and Methods
        private enum ParseState
        {
            Start,
            WhiteSpace,
            Text,
            SingleQuote,
            DoubleQuote,
            SpecialSymbol
        }

        private int length;
        private int index;
        private string currentText;
        private char currentChar;
        private const int minimal_capacity = 50;
        private StringBuilder acc_word;
        private int acc_begin;
        private int acc_end;
        private bool singleQuoted;
        private bool doubleQuoted;
        private List<TextParserStruct> list;
        private ParseState state;

        private bool IsIdentifierLetter(char c)
        {
            return Char.IsLetterOrDigit(c) || c == '_';
        }

        private bool IsWhiteSpace(char c)
        {
            return Char.IsWhiteSpace(c);
        }

        private bool IsSingleQuote(char c)
        {
            return c == '\'';
        }

        private bool IsDoubleQuote(char c)
        {
            return c == '"';
        }

        private void MoveIndex()
        {
            ++index;
            if (index < length)
                currentChar = currentText[index];
            else currentChar = '\n';
        }

        private void StartToken(ParseState newstate)
        {
            acc_word = new StringBuilder(currentChar.ToString(), minimal_capacity);
            acc_begin = index;
            state = newstate;
        }

        private void StartBlankToken(ParseState newState)
        {
            acc_word = new StringBuilder(minimal_capacity);
            acc_begin = index;
            singleQuoted = false;
            doubleQuoted = false;
            state = newState;
        }

        private void StartQuoteToken(ParseState newState)
        {
            if (newState == ParseState.SingleQuote)
            {
                if (includeSingleQuotes)
                    StartToken(ParseState.SingleQuote);
                else
                    StartBlankToken(ParseState.SingleQuote);
                singleQuoted = true;
            }
            else if (newState == ParseState.DoubleQuote)
            {
                if (includeDoubleQuotes)
                    StartToken(ParseState.DoubleQuote);
                else
                    StartBlankToken(ParseState.DoubleQuote);
                doubleQuoted = true;
            }
        }

        private void AddTokenToList()
        {
            acc_end = index;
            TextParserStruct textParserStruct = new TextParserStruct();
            textParserStruct.word = acc_word.ToString();
            textParserStruct.begin = acc_begin;
            textParserStruct.end = acc_end;
            textParserStruct.singleQuoted = singleQuoted;
            textParserStruct.doubleQuoted = doubleQuoted;
            textParserStruct.quoted = singleQuoted || doubleQuoted;
            list.Add(textParserStruct);
            acc_word = new StringBuilder(minimal_capacity); ;
            acc_begin = acc_end;
        }

        private void AddTokenToList_CurrentIncluded()
        {
            ++index;
            AddTokenToList();
            --index;
        }
        #endregion
    }
}
