using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Text.RegularExpressions;
using System.IO;
using System.Xml.Serialization;
using System.Windows.Forms;
using RegularLanguages;

namespace RestartingAutomaton
{
    public partial class WordToAllFilterForm : Form
    {
        public WordToAllFilterForm()
        {
            InitializeComponent();
            this.restartingAutomaton = null;
            this.fromWord = null;
            this.toWord = null;
            this.wordAcceptNamesList = null;
            this.UpdateUI();
        }

        public WordToAllFilterForm(
            RestartingAutomaton<char> restartingAutomaton,
            GenericString<char> fromWord,
            List<WordAcceptNames<char>> wordAcceptNamesList)
        {
            InitializeComponent();
            this.Text += " [" + fromWord.ToString() + "]";
            this.restartingAutomaton = restartingAutomaton;
            this.fromWord = fromWord;
            this.toWord = null;
            this.wordAcceptNamesList = wordAcceptNamesList;
            bool accept = false;
            foreach (WordAcceptNames<char> wordAcceptNames in wordAcceptNamesList)
                if (wordAcceptNames.accept)
                {
                    accept = true;
                    break;
                }
            if (accept)
                acceptLabel.Text = "The input word is accepted.";
            else acceptLabel.Text = "The input word is rejected.";
            this.UpdateUI();
        }

        private RestartingAutomaton<char> restartingAutomaton;
        private GenericString<char> fromWord;
        private GenericString<char> toWord;
        private List<WordAcceptNames<char>> wordAcceptNamesList;

        private bool UpdateUI()
        {
            if (this.wordAcceptNamesList != null)
            {
                this.filterTextBox.Enabled = true;
                this.filterButton.Enabled = true;
                this.filterListView.Enabled = true;
                this.bothRadioButton.Enabled = true;
                this.acceptedRadioButton.Enabled = true;
                this.rejectedRadioButton.Enabled = true;
                Regex filterRegex = null;
                try
                {
                    filterRegex = new Regex(this.filterTextBox.Text);
                }
                catch
                {
                    MessageBox.Show("Invalid regex pattern.", "Warning");
                    return false;
                }
                this.filterListView.SuspendLayout();
                this.filterListView.Items.Clear();
                foreach (WordAcceptNames<char> wordAcceptNames in 
                    this.wordAcceptNamesList)
                {
                    string word = wordAcceptNames.word.ToString();
                    string accept = "accept";
                    if (!wordAcceptNames.accept)
                        accept = "reject";
                    if (((this.bothRadioButton.Checked) ||
                         (this.acceptedRadioButton.Checked && wordAcceptNames.accept) ||
                         (this.rejectedRadioButton.Checked && !wordAcceptNames.accept)) &&
                        filterRegex.IsMatch(word))
                    {
                        ListViewItem listViewItem = this.filterListView.Items.Add(word);
                        listViewItem.SubItems.Add(accept);
                        listViewItem.SubItems.Add(String.Join(" ", wordAcceptNames.acceptingNames));
                        listViewItem.Tag = wordAcceptNames;
                    }
                }
                this.filterListView.ResumeLayout(true);
                this.cancelButton.Enabled = false;
                this.okButton.Enabled = true;
                return true;
            }
            else
            {
                this.filterTextBox.Enabled = false;
                this.filterButton.Enabled = false;
                this.filterListView.Enabled = false;
                this.bothRadioButton.Enabled = false;
                this.acceptedRadioButton.Enabled = false;
                this.rejectedRadioButton.Enabled = false;
                this.cancelButton.Enabled = false;
                this.okButton.Enabled = true;
                return false;
            }
        }

        private void StartWorkerUI()
        {
            this.filterTextBox.ReadOnly = true;
            this.filterButton.Enabled = false;
            this.filterListView.Enabled = false;
            this.bothRadioButton.Enabled = false;
            this.acceptedRadioButton.Enabled = false;
            this.rejectedRadioButton.Enabled = false;
            this.cancelButton.Enabled = true;
            this.okButton.Enabled = false;
        }

        private void StopWorkerUI()
        {
            this.filterTextBox.ReadOnly = false;
            this.filterButton.Enabled = true;
            this.filterListView.Enabled = true;
            this.bothRadioButton.Enabled = true;
            this.acceptedRadioButton.Enabled = true;
            this.rejectedRadioButton.Enabled = true;
            this.cancelButton.Enabled = false;
            this.okButton.Enabled = true;
        }

        private void filterTextBox_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
                this.UpdateUI();
        }

        private void filterButton_Click(object sender, EventArgs e)
        {
            this.UpdateUI();
        }

        private void filterListView_DoubleClick(object sender, EventArgs e)
        {
            if (this.restartingAutomaton != null &&
                this.filterListView.SelectedItems.Count == 1)
            {
                ListViewItem listViewItem = this.filterListView.SelectedItems[0];
                if (listViewItem.Tag != null && listViewItem.Tag is WordAcceptNames<char>)
                {
                    WordAcceptNames<char> wordAcceptNames =
                        (WordAcceptNames<char>)listViewItem.Tag;
                    this.toWord = wordAcceptNames.word;
                    if (this.wordToWordBackgroundWorker.IsBusy)
                    {
                        MessageBox.Show("Background worker is running an asynchronous operation.", "Warning");
                        return;
                    }
                    this.StartWorkerUI();
                    this.wordToWordBackgroundWorker.RunWorkerAsync();
                }
            }
        }

        private void wordToWordBackgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            e.Result = this.restartingAutomaton.WordToWordPaths(
                this.fromWord, this.toWord, this.wordToWordBackgroundWorker);
        }

        private void wordToWordBackgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            this.StopWorkerUI();
            if (e.Result is List<WordToWordPath<char>>)
            {
                List<WordToWordPath<char>> wordToWordPathList =
                    e.Result as List<WordToWordPath<char>>;
                WordToWordFilterForm filterForm = new WordToWordFilterForm(
                    this.fromWord, this.toWord, wordToWordPathList);
                filterForm.Show();
            }
        }

        private void cancelButton_Click(object sender, EventArgs e)
        {
            this.wordToWordBackgroundWorker.CancelAsync();
            this.StopWorkerUI();
        }

        private void okButton_Click(object sender, EventArgs e)
        {
            this.Close();
        }
    }
}